
package tip.util;

/**
 * {@code EqualityElement} encapsulates elements of type {@code T} with an
 * {@link Equalitor} of the same type. The class is used with {@link EqualitySet}
 * and {@link EqualityMap}.
 *  
 * @author Johnni Winther, jw@brics.dk
 *
 * @param <T> the element type
 */
public class EqualityElement<T> {
	private T element;				// the encapsulated element
	private Equalitor<T> equalitor;	// the equalitor for the element
	
	/**
	 * Creates a new {@code EqualityElement} encapsulating {@code element} with
	 * the {@link Equalitor} {@code equalitor}.
	 * @param element	the encapsulated element
	 * @param equalitor	the equalitor for the element
	 */
	public EqualityElement(T element, Equalitor<T> equalitor) {
		this.element = element;
		this.equalitor = equalitor;
	}
	
	/**
	 * Returns the encapsulated element.
	 * @return the encapsulated element
	 */
	public T getElement() {
		return element;
	}
	
	/**
	 * Returns the equalitor for the element.
	 * @return the equalitor for the element
	 */
	public Equalitor<T> getEqualitor() {
		return equalitor;
	}
	
	/**
	 * Returns {@code false} if {@code o} is not an {@code EqualityElement},
	 * otherwise {@link #equals(EqualityElement)}.
	 * @return {@code true} if {@code o} an equal {@code EqualityElement}
	 */
	@SuppressWarnings("unchecked")
	public boolean equals(Object o) {
		if (o instanceof EqualityElement) return equals((EqualityElement<T>)o);
		return false;
	}
	
	/**
	 * Returns {link {@link Equalitor#equals(Object, Object)} from the equalitor
	 * on the encapsulated element and {@code element}.
	 * @param element the compared element
	 * @return equals according to the equalitor
	 */
	public boolean equals(EqualityElement<T> element) {
		return equalitor.equals(this.element, element.element);
	}
	
	/**
	 * Returns the hashcode from equalitor on the encapsulated element.
	 * @return the hashcode from equalitor on the encapsulated element
	 */
	public int hashCode() {
		return equalitor.hashCode(element);
	}
}
