
package tip.util;

import java.util.*;

/**
 * {@code EqualityMap} implements a {@link Map} where equality of key elements
 * can be abstracted using an {@link Equalitor}, i.e. key elements are in the 
 * set are regarded as equal based on {@link Equalitor#equals(Object, Object)}.
 * 
 * @author Johnni Winther, jw@brics.dk
 * 
 * @param <K> the type of the keys
 * @param <V> the type of the values
 */
public class EqualityMap<K,V> implements Map<K,V> {
	// the equalitor used for equality of key elements
	protected Equalitor<K> equalitor;
	// the map holding the keys and values
	protected HashMap<EqualityElement<K>,V> map =
		new HashMap<EqualityElement<K>,V>();
	
	/**
	 * Creates a new empty map using {@link IdentityEqualitor} as equalitor
	 * on key elements.
	 */
	public EqualityMap() {
		equalitor = new IdentityEqualitor<K>();
	}
	
	/**
	 * Create a new empty map using {@code equalitor} as equalitor on key
	 * elements.
	 * @param equalitor the {@link Equalitor} used for equality on key elements
	 * @throws {@link IllegalArgumentException} if {@code equalitor} is 
	 * {@code null}
	 */
	public EqualityMap(Equalitor<K> equalitor) {
		if (equalitor == null)
			throw new IllegalArgumentException("No Equalitor provided.");
		this.equalitor = equalitor;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void clear() {
		map.clear();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public EqualityMap<K,V> clone() {
		EqualityMap<K,V> clone = new EqualityMap<K,V>(equalitor);
		for (Map.Entry<EqualityElement<K>,V> entry : map.entrySet()) {
			clone.map.put(entry.getKey(), entry.getValue());
		}
		return clone;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public boolean containsKey(Object key) {
		return map.containsKey(new EqualityElement(key, equalitor));
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked") 
	public boolean containsValue(Object value) {
		return map.containsValue(value);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public Set<Map.Entry<K,V>> entrySet() {
		Set<Map.Entry<K,V>> set = new HashSet<Map.Entry<K,V>>();
		for (Map.Entry<EqualityElement<K>,V> entry : map.entrySet()) {
			set.add(new Entry<K,V>(entry));
		}
		return set;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean equals(Object o) {
		if (!(o instanceof EqualityMap)) return false;
		if (!getClass().equals(o.getClass())) return false;
		
		@SuppressWarnings("unchecked")
		EqualityMap<K,V> equalityMap = (EqualityMap<K,V>)o;
		
		return equalitor.equals(equalityMap.equalitor)
			&& map.equals(equalityMap.map);
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public V get(Object key) {
		return map.get(new EqualityElement(key, equalitor));
	}
	
	/**
	 * Returns the {@link Equalitor} used for equality on key elements.
	 * @return the {@link Equalitor} used for equality on key elements
	 */
	public Equalitor<K> getEqualitor() {
		return equalitor;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public int hashCode() {
		return equalitor.hashCode()*13 + map.hashCode()*17;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isEmpty() {
		return map.isEmpty();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public EqualitySet<K> keySet() {
		return new EqualitySet<K>(map.keySet(), equalitor);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public V put(K key, V value) {
		return map.put(new EqualityElement<K>(key, equalitor), value);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void putAll(Map<? extends K, ? extends V> t) {
		for (Map.Entry<? extends K,? extends V> entry : t.entrySet()) {
			map.put(new EqualityElement<K>(entry.getKey(), equalitor), entry.getValue());
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public V remove(Object key) {
		return map.remove(new EqualityElement(key, equalitor));
	}
	
	/**
	 * {@inheritDoc}
	 */
	public int size() {
		return map.size();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return toString(new DefaultStringifier<K>(), new DefaultStringifier<V>());
	}
	
	/**
	 * Returns a text representation of the map using the {@code keyStringifier}
	 * and {@code valueStringifier}.
	 * @param keyStringifier the {@link Stringifier} used for key elements
	 * @param valueStringifier the {@link Stringifier} used for values
	 * @return a text representation of the set using the stringifiers
	 */
	public String toString(
			Stringifier<K> keyStringifier, 
			Stringifier<V> valueStringifier
		)
	{
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		Iterator<Map.Entry<K,V>> iterator = entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<K,V> entry = iterator.next();
			sb.append(keyStringifier.toString(entry.getKey()));
			sb.append("=");
			sb.append(valueStringifier.toString(entry.getValue()));
			if (iterator.hasNext()) {
				sb.append(",");
			}
		}
		sb.append("]");
		return sb.toString();
	}
	
	/**
	 * Returns a text representation of a subset of the keys in the map 
	 * using the {@code keyStringifier} and {@code valueStringifier}.
	 * @param keyStringifier the {@link Stringifier} used for key elements
	 * @param valueStringifier the {@link Stringifier} used for values
	 * @return a text representation of the set using the stringifiers
	 */
	public String toString(
			Collection<K> keySet, 
			Stringifier<K> keyStringifier, 
			Stringifier<V> valueStringifier
		)
	{
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		Iterator<K> iterator = keySet.iterator();
		while (iterator.hasNext()) {
			K key = iterator.next();
			if (containsKey(key)) {
				sb.append(keyStringifier.toString(key));
				sb.append("=");
				sb.append(valueStringifier.toString(get(key)));
				if (iterator.hasNext()) {
					sb.append(",");
				}
			}
		}
		sb.append("]");
		return sb.toString();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public Collection<V> values() {
		return map.values();
	}
	
	private static class Entry<K,V> implements Map.Entry<K,V> {
		private Map.Entry<EqualityElement<K>, V> entry;

		public Entry(Map.Entry<EqualityElement<K>,V> entry) {
			this.entry = entry;
		}

		public K getKey() {
			return entry.getKey().getElement();
		}

		public V getValue() {
			return entry.getValue();
		}

		public final V setValue(V newValue) {
			V oldValue = entry.getValue();
			entry.setValue(newValue);
			return oldValue;
		}

		public final boolean equals(Object o) {
			if (!(o instanceof Map.Entry))
				return false;
			Map.Entry e = (Map.Entry)o;
			Object k1 = getKey();
			Object k2 = e.getKey();
			if (k1 == k2 || (k1 != null && k1.equals(k2))) {
				Object v1 = getValue();
				Object v2 = e.getValue();
				if (v1 == v2 || (v1 != null && v1.equals(v2)))
					return true;
			}
			return false;
		}

		public int hashCode() {
			return (entry.getKey().hashCode()) ^
			(entry.getValue()==null ? 0 : entry.getValue().hashCode());
		}

		public String toString() {
			return getKey() + "=" + getValue();
		}
	}
}
