
package tip.util;

import java.util.*;

/**
 * {@code EqualitySet} implements a {@link Set} where equality of elements can
 * be abstracted using an {@link Equalitor}, i.e. elements in the set are
 * regarded as equal based on {@link Equalitor#equals(Object, Object)}.
 * 
 * @author Johnni Winther, jw@brics.dk
 *
 * @param <E> the type of the elements in the set
 */
public class EqualitySet<E> implements Set<E>, Stringifiable<E> {
	// the equalitor used for equality on elements
	protected Equalitor<E> equalitor;
	
	//	the set holding the elements
	protected HashSet<EqualityElement<E>> set = 
		new HashSet<EqualityElement<E>>();
	
	/**
	 * Creates a new empty {@code EqualitySet} using the 
	 * {@link IdentityEqualitor} as equalitor.
	 */
	public EqualitySet() {
		equalitor = new IdentityEqualitor<E>();
	}
	
	/**
	 * Create a new {@code EqualitySet} using the {@link IdentityEqualitor} as
	 * equalitor, initially containing the elements in {@code c}.
	 * @param c	the initial set
	 */
	public EqualitySet(Collection<E> c) {
		equalitor = new IdentityEqualitor<E>();
		addAll(c);
	}
	
	/**
	 * Creates a new empty {@code EqualitySet} using {@code equalitor} as the
	 * equalitor.
	 * @param equalitor the equalitor used for element equality
	 * @throws IllegalArgumentException if no {@link Equalitor} was provided
	 */
	public EqualitySet(Equalitor<E> equalitor) {
		if (equalitor == null) throw new IllegalArgumentException("No Equalitor provided.");
		this.equalitor = equalitor;
	}
	
	/**
	 * Creates a new {@code EqualitySet} using {@code equalitor} as the
	 * equalitor, initially containing the elements in {@code c}.
	 * @param equalitor the equalitor used for element equality
	 * @param c	the initial set
	 * @throws IllegalArgumentException if no {@link Equalitor} was provided
	 */
	public EqualitySet(Collection<E> c, Equalitor<E> equalitor) {
		if (equalitor == null) throw new IllegalArgumentException("No Equalitor provided.");
		this.equalitor = equalitor;
		addAll(c);
	}
	
	
	/**
	 * Create an {@code EqualitySet} containing the elements of {@code set} with
	 * {@code equalitor} as the equalitor.
	 * @param set	the initial set of elements 
	 * @param equalitor	the equalitor used for element equality
	 * @throws IllegalArgumentException if no {@link Equalitor} was provided
	 */
	public EqualitySet(Set<EqualityElement<E>> set, Equalitor<E> equalitor) {
		if (equalitor == null) throw new IllegalArgumentException("No Equalitor provided.");
		this.equalitor = equalitor;
		if (set != null)
			this.set.addAll(set);
	}
	
	/**
	 * Adds the element {@code o} to this set.
	 * @param o the element added to this set
	 * @return {@code true} is the element {@code o} was added to this set
	 */
	public boolean add(E o) {
		return set.add(new EqualityElement<E>(o, equalitor));
	}
	
	/**
	 * Adds the element held by the {@link EqualityElement} {@code e} to this
	 * set. If the equalitor of {@code e} is not the same as the equalitor used
	 * in the set, an {@link IllegalArgumentException} is thrown. 
	 * @param e	the element added to this set
	 * @return {@code true} if this set was changed
	 * @throws IllegalArgumentException if the equalitors mismatch
	 */
	public boolean add(EqualityElement<E> e) {
		if (!e.getEqualitor().equals(equalitor))
			throw new IllegalArgumentException("Mismatching Equalitor in EqualityElement.");
		return set.add(e);
	}
	
	/**
	 * Adds all elements of the collection {@code c} to this set
	 * @param c the collection in question
	 * @return {@code true} if this set was changed
	 */
	public boolean addAll(Collection<? extends E> c) {
		boolean change = false;
		for (E e : c) {
			change |= set.add(new EqualityElement<E>(e, equalitor));
		}
		return change;
	}
	
	/**
	 * Adds all elements in {@code c} to the set. If the equalitor of {@code e} 
	 * is not the same as the equalitor used in the set, an
	 * {@link IllegalArgumentException} is thrown. 
	 * @param c the set of elements added to this set
	 * @return {@code true} if this set was changed
	 * @throws IllegalArgumentException if the equalitors mismatch
	 */
	public boolean addAll(EqualitySet<E> c) {
		if (!c.getEqualitor().equals(equalitor))
			throw new IllegalArgumentException("Mismatching Equalitor in EqualitySet.");
		return set.addAll(c.set);
	}

	/**
	 * {@inheritDoc}
	 */
	public void clear() {
		set.clear();
	}
	
	/**
	 * Returns a shallow clone of this set. 
	 * @return a shallow clone of this set
	 */
	public EqualitySet<E> clone() {
		return new EqualitySet<E>(set, equalitor);
	}
	
	/**
	 * Returns {@code true} if this set contains the element {@code o}.
	 * @param o the element in question
	 * @return {@code true} if this set contains the element {@code o}.
	 */
	@SuppressWarnings("unchecked")
	public boolean contains(Object o) {
		return set.contains(new EqualityElement(o, equalitor));
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean containsAll(Collection<?> c) {
		for (Object o : c) {
			if (!contains(o)) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean equals(Object o) {
		if (o == this)
			return true;

		if (!(o instanceof Set))
			return false;
		Collection c = (Collection) o;
		if (c.size() != size())
			return false;
		try {
			return containsAll(c);
		} catch (ClassCastException unused)   {
			return false;
		} catch (NullPointerException unused) {
			return false;
		}
	}
	
	/**
	 * Returns the {@link Equalitor} used for element equality.
	 * @return the {@link Equalitor} used for element equality
	 */
	public Equalitor<E> getEqualitor() {
		return equalitor;
	}
	
	/**
	 * Returns the hashcode for this {@link EqualitySet}.
	 * @return the hashcode for this {@link EqualitySet}
	 */
	public int hashCode() {
		return equalitor.hashCode() ^ set.hashCode();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isEmpty() {
		return set.isEmpty();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public Iterator<E> iterator() {
		return new EqualitySetIterator<E>(set.iterator());
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public boolean remove(Object o) {
		return set.remove(new EqualityElement(o, equalitor));
	}
	
	/**
	 * Removes all element in the collection {@code c} from this set.
	 * @param c the collection in question
	 * @return {@code true} if this set was changed
	 */
	public boolean removeAll(Collection<?> c) {
		boolean change = false;
		for (Object o : c) {
			change |= remove(o);
		}
		return change;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean retainAll(Collection<?> c) {
		EqualitySet<E> clone = clone();
		clear();
		for (EqualityElement<E> e : clone.set) {
			if (c.contains(e.getElement())) {
				set.add(e);
			}
		}
		return clone.size() != size();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public int size() {
		return set.size();
	}
	
	/**
	 * {@inheritDoc}
	 */
    public Object[] toArray() {
        // Estimate size of array; be prepared to see more or fewer elements
    	Object[] r = new Object[size()];
        Iterator<E> it = iterator();
        for (int i = 0; i < r.length; i++) {
        	if (!it.hasNext())	// fewer elements than expected
        		return copyOf(r, i);
        	r[i] = it.next();
        }
        return it.hasNext() ? finishToArray(r, it) : r;
    }
 	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
    public <T> T[] toArray(T[] a) {
    	// Estimate size of array; be prepared to see more or fewer elements
    	int size = size();
    	T[] r = a.length >= size ? a :
    		(T[])java.lang.reflect.Array
    		.newInstance(a.getClass().getComponentType(), size);
    	Iterator<E> it = iterator();

    	for (int i = 0; i < r.length; i++) {
    		if (! it.hasNext()) { // fewer elements than expected
    			if (a != r)
    				return copyOf(r, i);
    			r[i] = null; // null-terminate
    			return r;
    		}
    		r[i] = (T)it.next();
    	}
    	return it.hasNext() ? finishToArray(r, it) : r;
    }
	
	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return toString(new DefaultStringifier<E>());
	}
	
	/**
	 * Returns a text representation of the set using the {@code stringifier}.
	 * @param stringifier the {@link Stringifier} used for elements
	 * @return a text representation of the set using the {@code stringifier}
	 */
	public String toString(Stringifier<E> stringifier) {
		StringBuffer sb = new StringBuffer();
		sb.append("{");
		Iterator<E> iterator = iterator();
		while (iterator.hasNext()) {
			sb.append(stringifier.toString(iterator.next()));
			if (iterator.hasNext()) {
				sb.append(",");
			}
		}
		sb.append("}");
		return sb.toString();
	}
 	
	@SuppressWarnings("unchecked")
    private static <T> T[] finishToArray(T[] r, Iterator<?> it) {
    	int i = r.length;
    	while (it.hasNext()) {
    		int cap = r.length;
    		if (i == cap) {
    			int newCap = ((cap / 2) + 1) * 3;
    			if (newCap <= cap) { // integer overflow
    				if (cap == Integer.MAX_VALUE)
    					throw new OutOfMemoryError
    					("Required array size too large");
    				newCap = Integer.MAX_VALUE;
    			}
    			r = copyOf(r, newCap);
    		}
    		r[i++] = (T)it.next();
    	}
    	// trim if overallocated
    	return (i == r.length) ? r : copyOf(r, i);
    }
	
	private static <T> T[] copyOf(T[] a, int size) {
		@SuppressWarnings("unchecked")
		T[] r = (T[])java.lang.reflect.Array
    		.newInstance(a.getClass().getComponentType(), size);
		for (int i = 0; i < size; i++) {
			r[i] = a[i];
		}
		return r;
	}
     	
	private static class EqualitySetIterator<E> implements Iterator<E> {
		private Iterator<EqualityElement<E>> iterator;
			
		public EqualitySetIterator(Iterator<EqualityElement<E>> iterator) {
			this.iterator = iterator;
		}
		
		public boolean hasNext() {
			return iterator.hasNext();
		}
		
		public E next() {
			EqualityElement<E> e = iterator.next();
			return e.getElement();
		}
		
		public void remove() {
			iterator.remove();
		}
	}
}
