package tip;

import tip.display.*;
import tip.display.ASTDisplay.Phase;

import tip.vm.*;

import tip.phases.*;
import tip.node.*;
import tip.lexer.*;
import tip.parser.*;

import tip.cfg.*;
import tip.cfg.cfa.*;
import tip.cfg.dfa.*;

import java.io.*;
import java.util.*;

/**
 * Commandline interface for the compiler
 */
public class Main {
	private static boolean run = false;

	private static boolean runFunction = false;

	private static boolean runShape = false;

	private static boolean runAndersen = false;

	private static boolean runSteensgaard = false;

	private static boolean runOneLevelFlow = false;

	private static int makeInterprocedural = -1;

	private static boolean showAST = false;

	private static boolean showAsm = false;

	private static boolean showCFG = false;

	private static boolean showPointsTo = false;

	private static boolean saveBinary = false;

	private static boolean LA = false;

	private static boolean AE = false;

	private static boolean VBE = false;

	private static boolean RD = false;

	private static boolean IVA = false;

	private static boolean SA = false;

	private static boolean ESA = false;

	private static boolean CP = false;

	private static boolean IA = false;

	private static boolean NP = false;

	private static boolean performNormalization = false;

	private static int vicinity = -1;

	public static PointsToMap pointsToMap = null;

	private static void showUsage() {
		System.out.println("Usage: tip [options] source");
		System.out.println("where options include");
		System.out.println("   -asm        : show compiled code");
		System.out.println("   -cfg        : show CFG (including analyses results)");
		System.out.println("   -pt         : show calculated points-to-maps in CFG display");
		System.out.println("   -run        : run program");
		System.out.println("   -bin        : save compiled (binary) program");
		System.out.println("   -Parser        : Show AST after the Parser phase");
		System.out.println("   -Normalization : Show AST after the Normalization phase");
		System.out.println("   -Environments     : Show AST after the Environments phase");
		System.out.println("   -NameLinking     : Show AST after the NameLinking phase");
		System.out.println("   -TypeChecking     : Show AST after the TypeChecking phase");
		System.out.println("   -Resources     : Show AST after the Resources phase");
		System.out.println("   -CodeGeneration     : Show AST after the CodeGeneration phase");
		System.out.println("   -a [A]      : perform the given analysis where [A] is one of");
		System.out.println("        LA : Liveness Analysis");
		System.out.println("        AE : Available Expressions Analysis");
		System.out.println("        VBE: Very Busy Expressions Analysis");
		System.out.println("        RD : Reaching Definitions Analysis");
		System.out.println("        IVA: Initialized Variables Analysis");
		System.out.println("        SA : Sign Analysis");
		System.out.println("        ESA: Extended Sign Analysis");
		System.out.println("        CP : Constant Propagation Analysis");
		System.out.println("        IA : Interval Analysis");
		System.out.println("        NP : Null Pointer Analysis");
		System.out.println("   -function   : compute function call \"points-to map\"");
		System.out.println("   -shape      : compute points-to map using Shape Analysis");
		System.out.println("   -andersen   : compute points-to map using Andersen's Algorithm");
		System.out.println("   -steensgaard: compute points-to map using Steensgaard's Algorithm");
		System.out.println("   -onelevel   : compute points-to map using One Level Flow");
		System.out.println("   -inter<n>   : make a mono/polyvariant interproduceral CFG");
		System.out.println("                  with at most <n> copies of for each call site");
		System.out.println("   -v<n>       : displays analysis information for a vicinity of <n> points");
	}

	private static void setAnalysesOption(String option) {
		if (option.equals("LA")) {
			LA = true;
		} else if (option.equals("AE")) {
			AE = true;
		} else if (option.equals("VBE")) {
			VBE = true;
		} else if (option.equals("RD")) {
			RD = true;
		} else if (option.equals("IVA")) {
			IVA = true;
		} else if (option.equals("SA")) {
			SA = true;
		} else if (option.equals("ESA")) {
			ESA = true;
		} else if (option.equals("CP")) {
			CP = true;
		} else if (option.equals("IA")) {
			IA = true;
		} else if (option.equals("NP")) {
			NP = true;
		} else {
			System.err.print("Unknown analysis : " + option);
		}
	}

	/**
	 * Main entry point for the TIP compiler.
	 * @param args
	 * 		command line arguments.
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		try {
			boolean showUsage = false;
			String filename = "";

			if (args.length == 0) {
				showUsage();
				return;
			}

			for (int i = 0; i < args.length; i++) {
				if (args[i].equals("-cfg")) {
					showCFG = true;
				} else if (args[i].equals("-pt")) {
					showPointsTo = true;
				} else if (args[i].equals("-shape")) {
					runShape = true;
				} else if (args[i].equals("-function")) {
					runFunction = true;
				} else if (args[i].equals("-andersen")) {
					runAndersen = true;
				} else if (args[i].equals("-steensgaard")) {
					runSteensgaard = true;
				} else if (args[i].equals("-onelevel")) {
					runOneLevelFlow = true;
				} else if (args[i].startsWith("-inter")) {
					try {
						makeInterprocedural = Integer.valueOf(args[i].substring(6));
					} catch (NumberFormatException e) {
						System.err.println("'" + args[i].substring(6)
								+ "' must an integer in the " + args[i]
								+ " option");
						showUsage = true;
					}
				} else if (args[i].equals("-run")) {
					run = true;
				} else if (args[i].equals("-ast")) {
					showAST = true;
				} else if (args[i].equals("-asm")) {
					showAsm = true;
				} else if (args[i].equals("-bin")) {
					saveBinary = true;
				} else if (args[i].equals("-a")) {
					if ((i + 1) >= args.length) {
						Errors.errorMessage("Missing argument for option " + args[i]);
						showUsage = true;
					} else {
						setAnalysesOption(args[i + 1]);
					}
					i++;
				} else if (args[i].startsWith("-v")) {
					try {
						vicinity = Integer.valueOf(args[i].substring(2));
					} catch (NumberFormatException e) {
						System.err.println("'" + args[i].substring(2)
								+ "' must an integer in the " + args[i]
								+ " option");
					}
				} else if (args[i].startsWith("-")) {
					boolean recognized = false;
					for (ASTDisplay.Phase phase : ASTDisplay.Phase.values()) {
						if (args[i].equals(phase.commandline_option)) {
							ASTDisplay.addDisplayPhase(phase);
							recognized = true;
							break;
						}
					}
					if (!recognized) {
						Errors.errorMessage("Invalid option: " + args[i]);
						showUsage = true;
					}
				} else {
					if (i != args.length - 1) {
						showUsage();
						return;
					}
					filename = args[i];
				}
			}

			if (filename.equals("")) {
				showUsage = true;
			}
			
			if (showUsage) {
				showUsage();
				return;
			}

			//
			// Determine if normalization is required
			//
			if (NP || runFunction || runShape || runAndersen || runSteensgaard
					|| runOneLevelFlow || makeInterprocedural >= 0) {
				performNormalization = true;
			}

			AProgram program = parseProgram(filename);
			ASTDisplay.showPhaseDisplay(program, Phase.PARSER);
			Errors.check();

			if (performNormalization) {
				program.apply(new Normalization());
				ASTDisplay.showPhaseDisplay(program, Phase.NORMALIZATION);
			}

			program.apply(new Environments());
			ASTDisplay.showPhaseDisplay(program, Phase.ENVIRONMENTS);
			Errors.check();

			program.apply(new NameLinking());
			ASTDisplay.showPhaseDisplay(program, Phase.NAME_LINKING);
			Errors.check();

			program.apply(new TypeChecking());
			ASTDisplay.showPhaseDisplay(program, Phase.TYPE_CHECKING);
			Errors.check();

			NodeFilter<PStm> filter = null;
			if (false) { // if true, only use statements as program points
				filter = new SubTypeFilter<PStm>(PStm.class);
			}
			List<FixedPointAnalysis<?>> analysesList = new LinkedList<FixedPointAnalysis<?>>();
			List<PointsToMap> pointsToMapList = new LinkedList<PointsToMap>();
			ICFGCollection icfgList = CFGBuilder.createCFGCollection(program,
					filter);

			if (runFunction
					|| performNormalization
					&& !(runShape || runAndersen || runSteensgaard || runOneLevelFlow)) {
				FunctionPointers functionPointers = new FunctionPointers();
				program.apply(functionPointers);
				pointsToMapList.add(functionPointers);
				pointsToMap = functionPointers;
			}

			if (runAndersen) {
				Andersen andersen = new Andersen();
				program.apply(andersen);
				pointsToMap = andersen;
				pointsToMapList.add(andersen);
			}

			if (runSteensgaard) {
				Steensgaard steensgaard = new Steensgaard();
				program.apply(steensgaard);
				pointsToMap = steensgaard;
				pointsToMapList.add(steensgaard);
			}

			if (runOneLevelFlow) {
				OneLevelFlow onelevelflow = new OneLevelFlow();
				program.apply(onelevelflow);
				pointsToMap = onelevelflow;
				pointsToMapList.add(onelevelflow);
			}

			if (makeInterprocedural >= 0) {
				icfgList = Interprocedural.makeInterprocedural(icfgList,
						makeInterprocedural);
			}

			if (runShape) {
				ShapeAnalysis shape = new ShapeAnalysis();
				WorkList.apply(shape, icfgList);
				pointsToMap = shape;
				pointsToMapList.add(shape);
				analysesList.add(shape);
			}

			if (LA) {
				LivenessAnalysis la = new LivenessAnalysis();
				analysesList.add(la);
				WorkList.apply(la, icfgList);
			}
			if (AE) {
				AvailableExpressionsAnalysis ae = new AvailableExpressionsAnalysis();
				analysesList.add(ae);
				WorkList.apply(ae, icfgList);
			}
			if (VBE) {
				VeryBusyExpressionsAnalysis vbe = new VeryBusyExpressionsAnalysis();
				analysesList.add(vbe);
				WorkList.apply(vbe, icfgList);
			}
			if (RD) {
				ReachingDefinitionsAnalysis rd = new ReachingDefinitionsAnalysis();
				analysesList.add(rd);
				WorkList.apply(rd, icfgList);
				rd.showDefUseGraph(icfgList);
			}
			if (IVA) {
				System.err.println("Initialized Variables Analysis has not been implemented.");
			}
			if (SA) {
				SignAnalysis sa = new SignAnalysis();
				analysesList.add(sa);
				WorkList.apply(sa, icfgList);
			}
			if (ESA) {
				System.err.println("Extended Sign Analysis has not been implemented.");
			}
			if (CP) {
				ConstantPropagationAnalysis cp = new ConstantPropagationAnalysis();
				analysesList.add(cp);
				WorkList.apply(cp, icfgList);
			}
			if (IA) {
				IntervalAnalysis ia = new IntervalAnalysis();
				analysesList.add(ia);
				WorkList.apply(ia, icfgList);
			}
			if (NP) {
				NullPointerAnalysis np = new NullPointerAnalysis();
				analysesList.add(np);
				WorkList.apply(np, icfgList);
			}

			program.apply(new Resources());
			ASTDisplay.showPhaseDisplay(program, Phase.RESOURCES);
			Errors.check();

			program.apply(new CodeGeneration());
			ASTDisplay.showPhaseDisplay(program, Phase.CODE_GENERATION);
			Errors.check();

			program.apply(new CodePrepare());
			Errors.check();

			program.apply(new CodeEmission());
			Errors.check();

			// Save to file
			if (saveBinary) {
				java.io.FileOutputStream out = new java.io.FileOutputStream(
						"file.tipobj");
				int numWords = program.getCode()[0];
				for (int i = 0; i < numWords; i++) {
					int word = program.getCode()[i];
					out.write(word & 0xff);
					out.write(word >> 8 & 0xff);
					out.write(word >> 16 & 0xff);
					out.write(word >> 24 & 0xff);
				}
			}

			// Run program
			if (run) {
				TipVm vm = new TipVm(program.getCode());
				vm.runProgram();
			}

			// Show asm
			if (showAST) {
				program.apply(new TypeCheckingDisplay());
			}

			// Show asm
			if (showAsm) {
				program.apply(new ShowAsm());
			}

			// Display CFG
			if (showCFG) {
				if (showPointsTo) {
					CFGDisplay.displayCFGs(icfgList, new NodeText<Node>(),
							analysesList, pointsToMapList, vicinity);
				} else {
					CFGDisplay.displayCFGs(icfgList, new NodeText<Node>(),
							analysesList, null, vicinity);
				}
			}

		} catch (SourceError e) {
			System.err.println("Errors during compilation.");
			System.exit(5);
		} finally {
			ASTDisplay.awaitDisplays();
		}
		System.exit(0);
	}

	static AProgram parseProgram(String filename) throws IOException {
		AProgram program = null;
		try {
			Parser parser = new Parser(new Lexer(new PushbackReader(
					new InputStreamReader(new FileInputStream(filename)))));

			Start startsym = parser.parse();
			program = startsym.getProgram();
		} catch (FileNotFoundException e) {
			Errors.errorMessage("File `" + filename + "' not found");
		} catch (LexerException e) {
			Errors.errorMessage("Error during lexing: " + e);
		} catch (ParserException e) {
			Errors.errorMessage("Error during parsing: " + e);
		}
		return program;
	}
}
