package tip.cfg;

import java.util.*;

/**
 * {@code Cubic} is an implementation of the cubic algorithm. The algorithm is
 * performed as constraints are added.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <Token>
 *            the type of the tokens
 * @param <Variable>
 *            the type of the variables
 */
public class Cubic<Token, Variable> {
	// if true, constraints and result are printed
	private static final boolean DEBUG = false;

	// the list of tokens
	private List<Token> tokens;

	// the map from variables to nodes in the graph of the algorithm
	private Map<Variable, Node> variables;

	// the list of nodes in the graph of the algorithm
	private List<Node> nodes;

	// the number of tokens
	private int numTokens;

	private class Node {
		public BitSet bitVector;

		public List<Variable[]>[] bitVectorList;

		public List<Node> adjacencyList;

		@SuppressWarnings("unchecked")
		public Node() {
			bitVector = new BitSet(numTokens);
			bitVectorList = new List[numTokens];
			adjacencyList = new LinkedList<Node>();
		}
	}

	/**
	 * Creates the {@code Cubic} object holding information for the algorithm.
	 * 
	 * @param tokens
	 *            the set of tokens for the algorithm
	 * @param variables
	 *            the set of variables for the algorithm
	 */
	public Cubic(Set<Token> tokens, Set<Variable> variables) {
		this.tokens = new ArrayList<Token>(tokens);
		numTokens = this.tokens.size();
		this.variables = new HashMap<Variable, Node>();
		nodes = new LinkedList<Node>();
		for (Variable v : variables) {
			Node n = new Node();
			this.variables.put(v, n);
			nodes.add(n);
		}
	}

	/**
	 * Returns the collection of tokens used in the algorithm.
	 * 
	 * @return the collection of tokens used in the algorithm
	 */
	public Collection<Token> getTokens() {
		return tokens;
	}

	/**
	 * Add a constraint of type {<i>t</i>} &#8838; <i>x</i>
	 * 
	 * @param t
	 *            the token <i>t</i> representing the singleton set {<i>i</i>}.
	 * @param x
	 *            the variable <i>x</i> which must contain the token {@code t}
	 */

	public void addConstraint1(Token t, Variable x) {
		if (DEBUG) {
			System.out.println("{" + t + "} \\subseteq [[" + x + "]]");
		}
		Node n = variables.get(x);
		int index = tokens.indexOf(t);
		setTokenTransitively(n, index);
	}

	private void setTokenTransitively(Node n, int index) {
		if (!n.bitVector.get(index)) { // not set
			n.bitVector.set(index);
			if (n.bitVectorList[index] != null) {
				for (Variable[] pair : n.bitVectorList[index]) {
					addConstraint2(pair[0], pair[1]);
				}
				n.bitVectorList[index] = null;
			}
			for (Node p : n.adjacencyList) {
				setTokenTransitively(p, index);
			}
		}
	}

	/**
	 * Add a constraint of type <i>x</i> &#8838; <i>y</i>
	 * 
	 * @param x
	 *            the variable <i>x</i> in question
	 * @param y
	 *            the variable <i>y</i> which must contain <i>x</i>
	 */
	public void addConstraint2(Variable x, Variable y) {
		if (DEBUG) {
			System.out.println("[[" + x + "]] \\subseteq [[" + y + "]]");
		}
		Node p = variables.get(x);
		Node q = variables.get(y);
		if (p == null || q == null) {
			throw new IllegalArgumentException(
					"addConstraint2: illegal argument : " + x + ", " + y);
		}
		if (p != q && !p.adjacencyList.contains(q)) {
			p.adjacencyList.add(q);
			Stack<Node> stack = new Stack<Node>();
			Set<Node> cycleSet = new HashSet<Node>();
			detectCycle(p, p, stack, cycleSet);
			if (cycleSet.size() != 0) { // a cycle is found => merge nodes
				Node n = new Node();
				for (Node r : cycleSet) { // merge node in cycle
					n.bitVector.or(r.bitVector); // union bit vectors
					for (int i = 0; i < numTokens; i++) { // concatenate bit
															// vector lists
						if (r.bitVectorList[i] != null) {
							if (n.bitVectorList[i] != null) {
								n.bitVectorList[i] = new LinkedList<Variable[]>();
							}
							n.bitVectorList[i].addAll(r.bitVectorList[i]);
						}
					}
					for (Node s : r.adjacencyList) { // merge edges
						if (!cycleSet.contains(s)
								&& !n.adjacencyList.contains(s)) {
							n.adjacencyList.add(s);
						}
					}
				}
				for (Variable v : variables.keySet()) { // update
														// variable-to-node map
					if (cycleSet.contains(variables.get(v))) {
						variables.put(v, n);
					}
				}
				for (int i = 0; i < numTokens; i++) { // 
					if (n.bitVector.get(i) && n.bitVectorList[i] != null) {
						for (Variable[] pair : n.bitVectorList[i]) {
							addConstraint2(pair[0], pair[1]);
						}
						n.bitVectorList[i] = null;
					}
				}
			} else { // no cycle found => set tokens transitively
				for (int i = 0; i < numTokens; i++) {
					for (Node r : p.adjacencyList) {
						if (p.bitVector.get(i)) {
							setTokenTransitively(r, i);
						}
					}
				}
			}
		}
	}

	private void detectCycle(Node n, Node origin, Stack<Node> stack,
			Set<Node> cycleSet) {
		stack.push(n);
		for (Node p : n.adjacencyList) {
			if (p == origin) {
				cycleSet.addAll(stack);
			} else {
				detectCycle(p, origin, stack, cycleSet);
			}
		}
		stack.pop();
	}

	/**
	 * Add a conditional constraint of type <i>t</i> &#8712; <i>x</i> &#8658;
	 * <i>y</i> &#8838; <i>z</i>
	 * 
	 * @param t
	 *            the variable <i>t</i> in question
	 * @param x
	 *            the variable <i>x</i> which may contain <i>t</i>
	 * @param y
	 *            the variable <i>y</i> in question
	 * @param z
	 *            the variable <i>z</i> which must contain <i>y</i> if <i>x</i>
	 *            contains <i>t</i>
	 */
	@SuppressWarnings("unchecked")
	public void addConstraint3(Token t, Variable x, Variable y, Variable z) {
		if (DEBUG) {
			System.out.println(t + "\\in [[" + x + "]] => [[" + y
					+ "]] \\subseteq [[" + z + "]]");
		}
		int index = tokens.indexOf(t);
		Node p = variables.get(x);
		if (p.bitVector.get(index)) {
			addConstraint2(y, z);
		} else {
			if (p.bitVectorList[index] == null) {
				p.bitVectorList[index] = new LinkedList<Variable[]>();
			}
			p.bitVectorList[index].add((Variable[]) new Object[] { y, z });
		}
	}

	/**
	 * Returns the set of tokens in the variable {@code v} as computed by the
	 * algorithm from the added constraints
	 * 
	 * @param v
	 *            the variable in question
	 * @return the set of tokens in the variable {@code v}
	 */
	public Set<Token> getTokenSet(Variable v) {
		Node n = variables.get(v);
		Set<Token> tokenSet = new HashSet<Token>();
		for (int i = 0; i < numTokens; i++) {
			if (n.bitVector.get(i)) {
				tokenSet.add(tokens.get(i));
			}
		}
		return tokenSet;
	}
}
