package tip.cfg;

import java.util.Set;

/**
 * {@code FixedPointAnalysis} defines the interface for general fixed-point
 * analyses for which {@link WorkList} can be used.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <VariableType>
 *            the type of the variables of the analysis
 */
public interface FixedPointAnalysis<VariableType> extends ICFGAnalysis {
	/**
	 * Returns the name of the analysis.
	 * 
	 * @return the name of the analysis
	 */
	public String getName();

	/**
	 * {@code start} is called before the fixed point analysis of the control
	 * flow graph {@code cfg} is started.
	 * 
	 * @param cfg
	 *            the control flow graph on which the analysis is made
	 */
	public void start(ICFG cfg);

	/**
	 * {@code end} is called after the fixed point analysis of the control flow
	 * graph {@code cfg} has ended.
	 * 
	 * @param cfg
	 *            the control flow graph on which the analysis has been made
	 */
	public void end(ICFG cfg);

	/**
	 * Returns the set of {@link IProgramPoint}s that depend upon {@code point}.
	 * 
	 * @param point
	 *            the {@link IProgramPoint} for which the dependency set is
	 *            returned
	 * @return the set of {@link IProgramPoint}s that depend upon {@code point}
	 */
	public Set<IProgramPoint> getDependencySet(IProgramPoint point);

	/**
	 * Returns the (current) variable associated with the {@code point}.
	 * 
	 * @param point
	 *            the {@link IProgramPoint} for which the variable is returned
	 * @return the (current) variable associated with the {@code point}
	 */
	public VariableType getVariable(IProgramPoint point);

	/**
	 * Returns a textual representation of the variable {@code variable}.
	 * 
	 * @param variable
	 *            the variable for which a text representation is returned
	 * @return a textual representation of the variable {@code variable}
	 */
	public String getVariableString(VariableType variable);

	/**
	 * Returns a textual representation of the variable for {@code point}.
	 * 
	 * @param point
	 *            the point for which the variable string is returned
	 * @return a textual representation of the variable for {@code point}
	 */
	public String getVariableStringForPoint(IProgramPoint point);

	/**
	 * Returns a textual representation of the variable for {@code point} only
	 * including information relevant for points in the {@code distance}
	 * vicinity of {@code point}.
	 * 
	 * @see IProgramPoint#getVicinity(int)
	 * @param point
	 *            the point for which the variable string is returned
	 * @param distance
	 *            the vicinity distance taken into account
	 * @return a textual representation of the variable for {@code point}
	 */
	public String getVariableStringForPoint(IProgramPoint point, int distance);
}
