package tip.cfg;

import java.util.*;
import tip.node.*;
import tip.util.EqualityMap;
import tip.*;

/**
 * {@code FlowAnalysis} implements the basic structure for a flow analysis
 * implementing the {@link FixedPointAnalysis} interface. To implement a flow
 * analysis, the following must be provided:
 * <ul>
 * <li>A {@link DependencyStrategy} must be provided to the constructor,
 * determining the direction of dependency in the analysis.</li>
 * <li>The {@link #createLattice(ICFG)} must be implemented to provide the
 * analysis with a {@link Lattice} for the current control flow graph.</li>
 * <li>The {@link FixedPointAnalysis#getVariableString(Object)} must be
 * implemented to provide a textual representation of the variables.</li>
 * <li>The necessary {@code case}-methods of {@link ICFGAnalysis} must be
 * implemented for the types of {@link IProgramPoint}s for which the variable is
 * not a simple join (see {@link #join(IProgramPoint)}. The
 * {@link FlowAnalysis#defaultNode(Node,IProgramPoint)} method has been
 * implemented to provide default for propagation of variables.</li>
 * </ul>
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <VariableType>
 *            the type of the variables associated with the program points
 */
public abstract class FlowAnalysis<VariableType> extends CFGAnalysisAdapter
		implements FixedPointAnalysis<VariableType> {
	// the strategy used for the dependency relation
	private DependencyStrategy strategy;

	// the name of the analysis
	private String name;

	// the map of program points to variables
	private EqualityMap<IProgramPoint, VariableType> variableMap = new EqualityMap<IProgramPoint, VariableType>();

	// the lattice for the variables
	private Lattice<VariableType> lattice;

	// current control flow graph
	private ICFG currentCFG;

	/**
	 * Creates a new flow analysis with name {@code name} using the {@code
	 * strategy} for the dependency relation.
	 * 
	 * @param strategy
	 *            the {@link DependencyStrategy} used for the dependency
	 *            relation
	 * @param name
	 *            the name of the analysis
	 */
	protected FlowAnalysis(DependencyStrategy strategy, String name) {
		if (strategy == null) {
			throw new IllegalArgumentException(
					"No DependencyStrategy provided.");
		}
		this.strategy = strategy;
		this.name = name;
	}

	/**
	 * Returns the name of the analysis.
	 * 
	 * @return the name of the analysis
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the {@link Lattice} used an analysis in progress. If no analysis
	 * is in progress {@code null} is returned.
	 * 
	 * @return the {@link Lattice} used an analysis in progress
	 */
	public Lattice<VariableType> getLattice() {
		return lattice;
	}

	/**
	 * {@code start} is called before the fixed point analysis of the control
	 * flow graph {@code cfg} is started. {@link #createLattice(ICFG)} is called
	 * to initial the lattice for the control flow graph {@code cfg}.
	 * 
	 * @param cfg
	 *            the control flow graph on which the analysis is made
	 */
	public void start(ICFG cfg) {
		currentCFG = cfg;
		lattice = createLattice(cfg);
		if (lattice == null) {
			throw new InternalCompilerError("No Lattice provided.");
		}
	}

	/**
	 * The method must be implemented to provided a {@link Lattice} for the
	 * analysis of the control flow graph {@code cfg}.
	 * 
	 * @param cfg
	 *            the control flow graph on which the lattice is used
	 * @return the lattice to use on the control flow graph
	 */
	protected abstract Lattice<VariableType> createLattice(ICFG cfg);

	/**
	 * {@code end} is called after the fixed point analysis of the control flow
	 * graph {@code cfg} has ended. This resets the lattice information.
	 * 
	 * @param cfg
	 *            the control flow graph on which the analysis has been made
	 */
	public void end(ICFG cfg) {
		lattice = null;
		currentCFG = null;
	}

	/**
	 * Returns the (current) variable associated the {@code point}.
	 * 
	 * @param point
	 *            the {@link IProgramPoint} for which the variable is returned
	 * @return the (current) variable associated the {@code point}
	 */
	public VariableType getVariable(IProgramPoint point) {
		return variableMap.get(point);
	}

	/**
	 * Sets {@code newVariable} as the variable for {@code point}. The method
	 * returns {@code true} if the new variable was not equal to the existing
	 * variable, thus the analysis has not reached a fixedpoint.
	 * 
	 * @param point
	 *            the program point to which the variable is associated
	 * @param newVariable
	 *            the new variable for the program point
	 * @return {@code true} if the variable of the program point was changed
	 */
	public boolean setVariable(IProgramPoint point, VariableType newVariable) {
		VariableType oldVariable = variableMap.put(point, newVariable);
		return !lattice.equals(oldVariable, newVariable);
	}

	/**
	 * Returns the set of {@link IProgramPoint}s that depend upon {@code point}.
	 * 
	 * @param point
	 *            the {@link IProgramPoint} for which the dependency set is
	 *            returned
	 * @return the set of {@link IProgramPoint}s that depend upon {@code point}
	 */
	public Set<IProgramPoint> getDependencySet(IProgramPoint point) {
		return strategy.getDependencySet(point);
	}

	/**
	 * Returns the join variable for {@code point}, i.e. the least upper bound
	 * of all variables for the program points that {@code point} depend upon.
	 * 
	 * @param point
	 *            the program point for which the join variable is returned
	 * @return the join variable for {@code point}
	 */
	protected VariableType join(IProgramPoint point) {
		VariableType val = lattice.getSmallestElement();
		for (IProgramPoint joinPoint : strategy.getJoinSet(point)) {
			assert currentCFG.containsProgramPoint(joinPoint) : "Join set referred to a program point not within the control flow graph";
			val = lattice.getLeastUpperBound(val, getVariable(joinPoint));
		}
		return val;
	}

	/**
	 * Default implementation for program points: The variable for {@code point}
	 * is updated to be the join of the points in its dependency set.
	 * 
	 * @param node
	 *            the node associated with {@code point}
	 * @param point
	 *            the handled point
	 * @return {@code true} is the variable chaned
	 * @see #join(IProgramPoint)
	 */
	public @Override
	boolean defaultNode(Node node, IProgramPoint point) {
		return setVariable(point, join(point));
	}

	/**
	 * Returns a textual representation of the variable for {@code point}.
	 * 
	 * @param point
	 *            the point for which the variable string is returned
	 * @return a textual representation of the variable for {@code point}
	 */
	public String getVariableStringForPoint(IProgramPoint point) {
		return getVariableString(getVariable(point));
	}

	/**
	 * Returns a textual representation of the variable for {@code point} only
	 * included information relevant for points in the {@code distance} vicinity
	 * of {@code point}.
	 * 
	 * @param point
	 *            the point for which the variable string is returned
	 * @param distance
	 *            the vicinity distance taken into account
	 * @return a textual representation of the varaible for {@code point}
	 */
	public String getVariableStringForPoint(IProgramPoint point, int distance) {
		return getVariableString(getVariable(point));
	}
}
