package tip.cfg;

import java.util.*;

import tip.util.Equalitor;
import tip.util.EqualitySet;
import tip.util.IdentityEqualitor;

/**
 * {@code IntersectionSetLattice} defines a set lattice with elements of type
 * {@code L} and with intersection as least-upper-bound operator. This is used
 * for must-analyses over sets.
 * <p>
 * {@code UnionSetLattice} defines a lattice whos elements are
 * {@link EqualitySet} sets with {@code L} as the type of the elements.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <L>
 *            the type of the elements in the set
 */
public class IntersectionSetLattice<L> implements Lattice<EqualitySet<L>> {
	protected EqualitySet<L> bottomElement; // the least element

	/**
	 * Creates a new {@code IntersectionSetLattice} using the
	 * {@link IdentityEqualitor} as equalitor used for element equality. The
	 * least element of the lattice is {@code set}.
	 * 
	 * @param set
	 *            the least element set
	 * @throws IllegalArgumentException
	 *             if no set is provided
	 */
	public IntersectionSetLattice(Collection<L> set) {
		if (set == null) {
			throw new IllegalArgumentException("No least element set provided.");
		}
		bottomElement = new EqualitySet<L>(set);
	}

	/**
	 * Creates a new {@code IntersectionSetLattice} using the {@code equalitor}
	 * as {@link Equalitor} used for element equality. The least element of the
	 * lattice is {@code set}.
	 * 
	 * @param set
	 *            the least element set
	 * @param equalitor
	 *            the {@link Equalitor} used for element equality
	 * @throws IllegalArgumentException
	 *             if no set or equalitor is provided
	 */
	public IntersectionSetLattice(Collection<L> set, Equalitor<L> equalitor) {
		if (set == null) {
			throw new IllegalArgumentException("No least element set provided.");
		}
		if (equalitor == null) {
			throw new IllegalArgumentException("No Equalitor provided.");
		}
		bottomElement = new EqualitySet<L>(set, equalitor);
	}

	/**
	 * Returns {@code true} if the sets {@code a} and {@code b} represent the
	 * same element in this set lattice.
	 * 
	 * @param a
	 *            the first set
	 * @param b
	 *            the second set
	 * @return {@code true} if the sets {@code a} and {@code b} represent the
	 *         same element in this set lattice
	 */
	public boolean equals(EqualitySet<L> a, EqualitySet<L> b) {
		return a == null ? b == null : a.equals(b);
	}

	/**
	 * Returns the hashcode of the set {@code a} according to this set lattice.
	 * 
	 * @param a
	 *            the set in question
	 * @return the hashcode of the set {@code a} according to this set lattice
	 */
	public int hashCode(EqualitySet<L> a) {
		return a == null ? 0 : a.hashCode();
	}

	/**
	 * Returns the smallest element of this set lattice.
	 * 
	 * @return the smallest element of this set lattice
	 */
	public EqualitySet<L> getSmallestElement() {
		return bottomElement.clone();
	}

	/**
	 * Returns the least upper bound of the sets {@code a} and {@code b}
	 * according to this set lattice. The least upper bound is the intersection
	 * of the sets.
	 * 
	 * @param a
	 *            the first set
	 * @param b
	 *            the second set
	 * @return the least upper bound of the sets {@code a} and {@code b}
	 *         according to this set lattice
	 */
	public EqualitySet<L> getLeastUpperBound(EqualitySet<L> a, EqualitySet<L> b) {
		if (a == null && b == null) {
			return bottomElement;
		} else if (a == null) {
			return b.clone();
		} else if (b == null) {
			return a.clone();
		} else {
			assert (bottomElement.containsAll(a));
			assert (bottomElement.containsAll(b));
			EqualitySet<L> lub = a.clone();
			lub.retainAll(b);
			return lub;
		}
	}
}
