package tip.cfg;

import java.util.*;

import tip.util.Equalitor;
import tip.util.EqualityMap;
import tip.util.EqualitySet;

/**
 * {@code MapLattice} defines a map lattice, <i>A</i> &#8594; <i>L</i>, from a
 * fixed set, <i>A</i>, containing the elements of the domain, and a lattice,
 * <i>L</i>, for the range elements.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <A>
 *            the type of the elements in the domain set
 * @param <L>
 *            the type of the elements in the range lattice
 */
public class MapLattice<A, L> implements Lattice<EqualityMap<A, L>> {
	// the lattice of the range elements
	protected Lattice<L> rangeLattice;

	// the smallest element of this map lattice
	protected EqualityMap<A, L> smallestElement;

	/**
	 * Creates a new map lattice using {@code rangeLattice} as the lattice for
	 * range elements and {@code domainSet} as the domain set. The equalitor for
	 * the domain elements will be the {@link tip.util.IdentityEqualitor}.
	 * 
	 * @param domainSet
	 *            the set of elements in the domain
	 * @param rangeLattice
	 *            the lattice for the range element
	 */
	public MapLattice(Collection<A> domainSet, Lattice<L> rangeLattice) {
		if (rangeLattice == null) {
			throw new IllegalArgumentException("No Lattice provided.");
		}
		this.rangeLattice = rangeLattice;
		this.smallestElement = new EqualityMap<A, L>();
		for (A a : domainSet) {
			smallestElement.put(a, rangeLattice.getSmallestElement());
		}
	}

	/**
	 * Creates a new map lattice using {@code rangeLattice} as the lattice for
	 * range elements and {@code domainSet} as the domain set. The equalitor for
	 * the domain elements will be the equalitor from {@code domainSet}.
	 * 
	 * @param domainSet
	 *            the set of elements in the domain
	 * @param rangeLattice
	 *            the lattice for the range element
	 */
	public MapLattice(EqualitySet<A> domainSet, Lattice<L> rangeLattice) {
		if (rangeLattice == null) {
			throw new IllegalArgumentException("No Lattice provided.");
		}
		this.rangeLattice = rangeLattice;
		this.smallestElement = new EqualityMap<A, L>(domainSet.getEqualitor());
		for (A a : domainSet) {
			smallestElement.put(a, rangeLattice.getSmallestElement());
		}
	}

	/**
	 * Returns the {@link Equalitor} used for the domain elements.
	 * 
	 * @return the {@link Equalitor} used for the domain elements
	 */
	public Equalitor<A> getDomainEqualitor() {
		return smallestElement.getEqualitor();
	}

	/**
	 * Returns the {@link Lattice} for the range elements.
	 * 
	 * @return the {@link Lattice} for for the range elements.
	 */
	public Lattice<L> getRangeLattice() {
		return rangeLattice;
	}

	/**
	 * Returns the smallest element of this map lattice.
	 * 
	 * @return the smallest element of this map lattice
	 */
	public EqualityMap<A, L> getSmallestElement() {
		return smallestElement.clone();
	}

	/**
	 * Returns the least upper bound of the maps {@code a} and {@code b}
	 * according to this map lattice.
	 * 
	 * @param a
	 *            the first map
	 * @param b
	 *            the second map
	 * @return the least upper bound of the maps {@code a} and {@code b}
	 *         according to this map lattice
	 */
	public EqualityMap<A, L> getLeastUpperBound(EqualityMap<A, L> a,
			EqualityMap<A, L> b) {
		if (a == null && b == null) {
			return getSmallestElement();
		}
		if (a == null) {
			return b.clone();
		}
		if (b == null) {
			return a.clone();
		}
		EqualityMap<A, L> lub = getSmallestElement();
		for (A key : lub.keySet()) {
			assert lub.keySet().equals(a.keySet()) : "Different range sets used in the same MapLattice";
			assert a.keySet().equals(b.keySet()) : "Different range sets used in the same MapLattice";
			lub.put(key,
					rangeLattice.getLeastUpperBound(a.get(key), b.get(key)));
		}
		return lub;
	}

	/**
	 * Returns {@code true} if the maps {@code a} and {@code b} represent the
	 * same element in this map lattice.
	 * 
	 * @param a
	 *            the first map
	 * @param b
	 *            the second map
	 * @return {@code true} if the maps {@code a} and {@code b} represent the
	 *         same element in this map lattice
	 */
	public boolean equals(EqualityMap<A, L> a, EqualityMap<A, L> b) {
		if (a == null) {
			return b == null;
		}
		for (A key : a.keySet()) {
			if (!rangeLattice.equals(a.get(key), b.get(key))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns the hashcode for the map {@code a} according to this map lattice.
	 * 
	 * @param a
	 *            the map in question
	 * @return the hashcode for the map {@code a} according to this map lattice
	 */
	public int hashCode(EqualityMap<A, L> a) {
		return a == null ? 0 : a.hashCode();
	}
}
