package tip.cfg;

import tip.util.Equalitor;
import tip.util.EqualitySet;
import tip.util.IdentityEqualitor;

/**
 * {@code UnionSetLattice} defines a set lattice with elements of type {@code L}
 * and with union as least-upper-bound operator. This is used for may-analyses
 * over sets.
 * <p>
 * {@code UnionSetLattice} defines a lattice whos elements are
 * {@link EqualitySet} sets with {@code L} as the type of the elements.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <L>
 *            the type of the elements in the set
 */
public class UnionSetLattice<L> implements Lattice<EqualitySet<L>> {
	// the least element
	protected EqualitySet<L> smallestElement;

	/**
	 * Creates a new {@code UnionSetLattice} using the {@link IdentityEqualitor}
	 * as equalitor used for element equality.
	 */
	public UnionSetLattice() {
		smallestElement = new EqualitySet<L>();
	}

	/**
	 * Creates a new {@code UnionSetLattice} using the {@code equalitor} as
	 * {@link Equalitor} used for element equality.
	 * 
	 * @param equalitor
	 *            the {@link Equalitor} used for element equality
	 * @throws IllegalArgumentException
	 *             if no equalitor is provided
	 */
	public UnionSetLattice(Equalitor<L> equalitor) {
		if (equalitor == null) {
			throw new IllegalArgumentException("No Equalitor provided.");
		}
		smallestElement = new EqualitySet<L>(equalitor);
	}

	/**
	 * Returns {@code true} if the sets {@code a} and {@code b} represent the
	 * same element in this set lattice.
	 * 
	 * @param a
	 *            the first set
	 * @param b
	 *            the second set
	 * @return {@code true} if the sets {@code a} and {@code b} represent the
	 *         same element in this set lattice
	 */
	public boolean equals(EqualitySet<L> a, EqualitySet<L> b) {
		return a == null ? b == null : a.equals(b);
	}

	/**
	 * Returns the hashcode of the set {@code a} according to this set lattice.
	 * 
	 * @param a
	 *            the set in question
	 * @return the hashcode of the set {@code a} according to this set lattice
	 */
	public int hashCode(EqualitySet<L> a) {
		return a == null ? 0 : a.hashCode();
	}

	/**
	 * Returns the smallest element of this set lattice.
	 * 
	 * @return the smallest element of this set lattice
	 */
	public EqualitySet<L> getSmallestElement() {
		return smallestElement.clone();
	}

	/**
	 * Returns the least upper bound of the sets {@code a} and {@code b}
	 * according to this set lattice. The least upper bound is the union of the
	 * sets.
	 * 
	 * @param a
	 *            the first set
	 * @param b
	 *            the second set
	 * @return the least upper bound of the sets {@code a} and {@code b}
	 *         according to this set lattice
	 */
	public EqualitySet<L> getLeastUpperBound(EqualitySet<L> a, EqualitySet<L> b) {
		if (a == null && b == null) {
			return smallestElement;
		} else if (a == null) {
			return b.clone();
		} else if (b == null) {
			return a.clone();
		} else {
			EqualitySet<L> lub = a.clone();
			lub.addAll(b);
			return lub;
		}
	}
}
