package tip.cfg;

import java.util.*;

/**
 * {@code WorkList} implements the work list algorithm for
 * {@link FixedPointAnalysis} and {@link WideningNarrowingAnalysis}.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class WorkList {
	/**
	 * Performes the {@link FixedPointAnalysis} {@code fixedPointAnalysis} on
	 * the control flow graph in {@code cfgCollection}.
	 * 
	 * @param fixedPointAnalysis
	 *            the analysis applied
	 * @param cfgCollection
	 *            the collection of control flow graphs
	 * @return the analysis provided by {@code fixedPointAnalysis}
	 */
	public static <V> FixedPointAnalysis<V> apply(
			FixedPointAnalysis<V> fixedPointAnalysis, ICFGCollection cfgCollection) {
		for (ICFG cfg : cfgCollection.getCFGs()) {
			processCFG(fixedPointAnalysis, cfg);
		}
		return fixedPointAnalysis;
	}

	/**
	 * Performes the {@link WideningNarrowingAnalysis} {@code
	 * wideningNarrowingAnalysis} on the control flow graph in {@code
	 * cfgCollection}.
	 * 
	 * @param wideningNarrowingAnalysis
	 *            the analysis applied
	 * @param cfgCollection
	 *            the collection of control flow graphs
	 * @return the analysis provided by {@code wideningNarrowingAnalysis}
	 */
	public static <V> WideningNarrowingAnalysis<V> apply(
			WideningNarrowingAnalysis<V> wideningNarrowingAnalysis,
			ICFGCollection cfgCollection) {
		for (ICFG cfg : cfgCollection.getCFGs()) {
			wideningNarrowingAnalysis.setWidening();
			processCFG(wideningNarrowingAnalysis, cfg);
			wideningNarrowingAnalysis.setNarrowing();
			processCFG(wideningNarrowingAnalysis, cfg);
		}
		return wideningNarrowingAnalysis;
	}

	private static <V> void processCFG(FixedPointAnalysis<V> fixedPointAnalysis,
			ICFG cfg) {
		fixedPointAnalysis.start(cfg);
		Set<IProgramPoint> workList = new HashSet<IProgramPoint>(
				cfg.getProgramPoints());
		while (!workList.isEmpty()) {
			IProgramPoint programPoint = workList.iterator().next();
			workList.remove(programPoint);
			if (programPoint.apply(fixedPointAnalysis)) {
				workList.addAll(fixedPointAnalysis.getDependencySet(programPoint));
			}
		}
		fixedPointAnalysis.end(cfg);
	}
}
