package tip.cfg.cfa;

import java.util.*;

import tip.analysis.*;
import tip.cfg.*;
import tip.node.*;
import tip.util.*;

/**
 * {@code FunctionPointers} is an implemention of a simple control flow analysis
 * for functions pointers.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class FunctionPointers extends DepthFirstAdapter implements PointsToMap {
	// if true, contraints and results are printed
	private static final boolean DEBUG = false;

	// the set variables of the algorithm
	private EqualitySet<PDecl> variables = new EqualitySet<PDecl>();

	// the set of possible targets of the variables
	private EqualitySet<Target> targets = new EqualitySet<Target>();

	// the map of variables to sets of targets as produced by the algorithm
	private EqualityMap<PDecl, Set<Target>> pointsToMap = new EqualityMap<PDecl, Set<Target>>();

	// the cubic algorithm used
	protected Cubic<Target, PDecl> cubic;

	private static class DeclStringifier implements Stringifier<PDecl> {
		public String toString(PDecl decl) {
			return decl.getName().getText();
		}
	}

	public @Override
	void inAProgram(AProgram node) {
		node.getDescendants(variables, new SubTypeFilter<PDecl>(PDecl.class));
		for (PDecl decl : variables) {
			targets.add(new DeclarationTarget(decl));
		}

		cubic = new Cubic<Target, PDecl>(targets, variables);
		if (DEBUG) {
			System.out.println("Variables: "
					+ variables.toString(new DeclStringifier()));
			System.out.println("Targets  : " + targets);
		}
	}

	public @Override @SuppressWarnings("unused")
	void outAProgram(AProgram node) {
		for (PDecl var : variables) {
			pointsToMap.put(var, cubic.getTokenSet(var));
			if (DEBUG) {
				System.out.print("[[" + var.getName().getText() + "]] = ");
				System.out.println(pointsToMap.get(var));
			}
		}
	}

	public @Override
	void inAAssignVarStm(AAssignVarStm node) {
		PDecl id1 = node.getIdentifier().getDecl();
		PExp exp = node.getExp();
		if (exp instanceof AVarExp) {
			/**
			 * id1 = id2 : id2 \in vars => [[id2]] \subseteq [[id1]] id1 = id2 :
			 * id2 \in funcs => {&id2} \subseteq [[id1]]
			 */
			PDecl decl = ((AVarExp) exp).getIdentifier().getDecl();
			if (decl instanceof ALocalDecl) {
				cubic.addConstraint2(decl, id1);
			} else {
				cubic.addConstraint1(new DeclarationTarget(decl), id1);
			}
		} else if (exp instanceof AFunctionPtrCallExp) {
			/**
			 * id1 = (id2)(a1,...,an) => &f \in [[id2]] : f(x1,...,xn) { ...
			 * return id; } => [[a1]] \subseteq [[x1]] ... [[an]] \subseteq
			 * [[xn]] [[id]] \subseteq [[id1]]
			 */
			AFunctionPtrCallExp funcExp = (AFunctionPtrCallExp) exp;
			LinkedList<PExp> arguments = funcExp.getArgs();
			// this assumes normalization of pointer expressions
			PDecl id2 = ((AVarExp) funcExp.getExp()).getIdentifier().getDecl();
			for (Target t : cubic.getTokens()) {
				if (t instanceof DeclarationTarget) {
					PDecl decl = ((DeclarationTarget) t).getTarget();
					if (decl instanceof AFuncDecl) {
						AFuncDecl funcDecl = (AFuncDecl) decl;
						List<ALocalDecl> formals = funcDecl.getFormals();
						if (arguments.size() == formals.size()) {
							for (int i = 0; i < arguments.size(); i++) {
								cubic.addConstraint3(
										t,
										id2,
										((AVarExp) arguments.get(i)).getIdentifier().getDecl(),
										formals.get(i));
							}
							List<PStm> statements = ((funcDecl.getBody()).getBlock()).getStatements();
							AReturnStm returnStm = (AReturnStm) statements.get(statements.size() - 1);
							cubic.addConstraint3(
									t,
									id2,
									((AVarExp) returnStm.getExp()).getIdentifier().getDecl(),
									id1);
						}
					}
				}
			}
		}
	}

	public @Override
	void inAFuncDecl(AFuncDecl node) {
		/**
		 * id(...) => {&id} \subseteq [[id]]
		 */
		cubic.addConstraint1(new DeclarationTarget(node), node);
	}

	/**
	 * Returns the name of this points-to map.
	 * 
	 * @return the name of this points-to map
	 */
	public String getName() {
		return "function";
	}

	/**
	 * Returns the pointer targets for the variable {@code decl} at the program
	 * point {@code point} as computed by the algorithm.
	 * 
	 * @param point
	 *            the program point for which the pointer targets are returned
	 * @param decl
	 *            the variable for which the pointer targets are returned
	 * @return the pointer target for the {@code decl} at {@code point}
	 */
	public Set<Target> pt(IProgramPoint point, ALocalDecl decl) {
		Set<Target> pt = pointsToMap.get(decl);
		if (pt == null) {
			pt = new HashSet<Target>();
			pointsToMap.put(decl, pt);
		}
		return pt;
	}

	public boolean isFlowSensitive() {
		return false;
	}

	private Map<Target, Set<Target>> ptMap;

	@SuppressWarnings("unused")
	public Map<Target, Set<Target>> pt(IProgramPoint point) {
		if (ptMap == null) {
			ptMap = new HashMap<Target, Set<Target>>();
			for (PDecl decl : pointsToMap.keySet()) {
				Set<Target> targetSet = pointsToMap.get(decl);
				if (targetSet != null && !targetSet.isEmpty()) {
					ptMap.put(new DeclarationTarget(decl), targetSet);
				}
			}
		}
		return ptMap;
	}
}
