package tip.cfg.cfa;

import tip.*;
import tip.cfg.*;
import tip.node.*;
import tip.analysis.*;
import java.util.*;

/**
 * {@code Interprocedural} converts a collection of intraprocedural control flow
 * graphs into a single interprocedural control flow graph. The entry point of
 * the new control flow graph is the entry point of the {@code "main"} method.
 * <p>
 * The maximal number of copies of a control flow graph for each call-site can
 * be specified in order make both mono- and polyvariant analyses. If the
 * maximal number of copies is 0, the interproduceral control flow graph is
 * monovariant, and the maximal number of copies is greater than 0, the
 * interprocedural control flow graph is polyvariant.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class Interprocedural extends
		CFGQuestionAnswerAdapter<Map<String, ICFG>, ICFG> {
	public static final boolean DEBUG = false;

	/**
	 * Creates a new {@link ICFGCollection} containing a single interprocedural
	 * control flow graph created from {@code oldCollection}. {@code maxCopies}
	 * specifies the degree of mono-/polyvariance.
	 * 
	 * @param oldCollection
	 *            the collection from which the control flow graph is created
	 * @param maxCopies
	 *            the maximal number of copies of a control flow graph for each
	 *            call-site.
	 * @return a singleton collection containing an interprocedural control flow
	 *         graph
	 */
	public static ICFGCollection makeInterprocedural(
			ICFGCollection oldCollection, int maxCopies) {
		Interprocedural monoVariant = new Interprocedural(oldCollection,
				maxCopies);
		return monoVariant.collection;
	}

	private CCFGCollection collection = new CCFGCollection();

	private CCFGCollection subCollection = new CCFGCollection();

	private Map<ICFG, Map<String, ALocalDecl>> cfgLocalDeclMap = new HashMap<ICFG, Map<String, ALocalDecl>>();

	private Map<ICFG, Map<String, ALocalDecl>> cfgFormalDeclMap = new HashMap<ICFG, Map<String, ALocalDecl>>();

	private Map<String, ALocalDecl> newLocalDeclMap = new HashMap<String, ALocalDecl>();

	private int callIndexCounter = 1;

	private Map<Node, Map<DeclarationTarget, Integer>> callSiteToTargetMap = new HashMap<Node, Map<DeclarationTarget, Integer>>();

	private int maxCopies = 0;

	private Interprocedural(ICFGCollection oldCollection, int maxCopies) {
		this.maxCopies = maxCopies;
		for (ICFG cfg : oldCollection) {
			ICFG subCFG = makeSubCFG(cfg);
			subCollection.addCFG(subCFG);
			collectFormalsAndLocals(subCFG);
		}
		ICFG main = subCollection.getCFG("main");
		if (main == null) {
			throw new SourceError("No main method");
		}

		while (makeMonoVariant(main)) {
		}

		main.clean();
		collection.addCFG(main);
	}

	private ALocalDecl getLocalDecl(String name) {
		ALocalDecl localDecl = newLocalDeclMap.get(name);
		if (localDecl == null) {
			TIdentifier id = new TIdentifier(name);
			localDecl = new ALocalDecl(id);
			id.setDecl(localDecl);
			newLocalDeclMap.put(name, localDecl);
		}
		return localDecl;
	}

	private TIdentifier getIdentifier(String name) {
		ALocalDecl localDecl = getLocalDecl(name);
		return localDecl.getName().clone();
	}

	private TIdentifier makeId(ALocalDecl localDecl) {
		TIdentifier id = localDecl.getName().clone();
		id.setDecl(localDecl);
		return id;
	}

	private AVarExp makeVarExp(ALocalDecl localDecl) {
		TIdentifier id = localDecl.getName().clone();
		id.setDecl(localDecl);
		return new AVarExp(id);
	}

	private void showId(TIdentifier id) {
		System.out.println(id.getText() + ":" + id.getDecl().hashCode());
	}

	private ICFG makeSubCFG(ICFG old_cfg) {
		ICFG cfg = old_cfg.clone();
		String ret = "$ret-" + cfg.getName();
		getLocalDecl(ret);

		ReturnToAssign rta = new ReturnToAssign(cfg, ret);
		for (IProgramPoint point : new HashSet<IProgramPoint>(
				cfg.getProgramPoints())) {
			point.apply(rta);
		}

		return cfg;
	}

	private class ReturnToAssign extends CFGAnalysisAdapter {
		private ICFG cfg;

		private String name;

		public ReturnToAssign(ICFG cfg, String name) {
			this.cfg = cfg;
			this.name = name;
		}

		public @Override
		boolean caseCReturnStmPoint(AReturnStm node, CReturnStmPoint point) {
			AAssignVarStm assignVarStm = new AAssignVarStm(new TAssign(),
					getIdentifier(name), node.getExp().clone());
			if (DEBUG) {
				showId(assignVarStm.getIdentifier());
			}
			cfg.replacePoint(point, new CAssignVarStmPoint(assignVarStm));
			return false;
		}
	}

	private void collectFormalsAndLocals(ICFG cfg) {
		Map<String, ALocalDecl> localDeclSet = new HashMap<String, ALocalDecl>();
		Map<String, ALocalDecl> formalDeclSet = new HashMap<String, ALocalDecl>();
		cfg.getEntry().getNode().apply(
				new LocalsCollector(localDeclSet, formalDeclSet));
		cfgLocalDeclMap.put(cfg, localDeclSet);
		cfgFormalDeclMap.put(cfg, formalDeclSet);
	}

	private class LocalsCollector extends DepthFirstAdapter {
		private Map<String, ALocalDecl> localDeclSet;

		private Map<String, ALocalDecl> formalDeclSet;

		public LocalsCollector(Map<String, ALocalDecl> localDeclSet,
				Map<String, ALocalDecl> formalDeclSet) {
			this.localDeclSet = localDeclSet;
			this.formalDeclSet = formalDeclSet;
		}

		public @Override
		void defaultInPBlock(PBlock block) {
			localDeclSet.putAll(block.getLocalEnv());
		}

		public @Override
		void inAFuncDecl(AFuncDecl node) {
			for (PDecl decl : node.getFormals()) {
				ALocalDecl formal = (ALocalDecl) decl;
				formalDeclSet.put(formal.getName().getText(), formal);
			}
		}
	}

	private boolean makeMonoVariant(ICFG cfg) {
		FunctionCallConverter fcc = new FunctionCallConverter(cfg);
		boolean change = false;
		for (IProgramPoint point : new HashSet<IProgramPoint>(
				cfg.getProgramPoints())) {
			change |= point.apply(fcc);
		}
		return change;
	}

	private class FunctionCallConverter extends CFGAnalysisAdapter {
		private ICFG cfg;

		public FunctionCallConverter(ICFG cfg) {
			this.cfg = cfg;
		}

		private ICFG linkMethods(TIdentifier assignId, List<PExp> args,
				Set<ICFG> calledCFGs) {
			int callIndex = callIndexCounter++;

			// make wrapper cfg
			List<IProgramPoint> wrapperPoints = new LinkedList<IProgramPoint>();
			wrapperPoints.add(new EntryPoint());

			// save context
			// save-i-b_j = b_j
			// save-i-x_j = x_j
			for (Map.Entry<String, ALocalDecl> entry : cfgLocalDeclMap.get(cfg).entrySet()) {
				String name = "$save-" + callIndex + "-" + entry.getKey();
				TIdentifier id = getIdentifier(name);
				AAssignVarStm assignVarStm = new AAssignVarStm(new TAssign(),
						id, makeVarExp(entry.getValue()));
				if (DEBUG) {
					showId(assignVarStm.getIdentifier());
				}
				wrapperPoints.add(new CAssignVarStmPoint(assignVarStm));
			}
			// save args
			// temp-i-a_j = E_j
			int argIndex = 0;
			for (PExp arg : args) {
				String name = "$temp-" + callIndex + "-" + argIndex;
				TIdentifier id = getIdentifier(name);
				AAssignVarStm assignVarStm = new AAssignVarStm(new TAssign(),
						id, arg.clone());
				if (DEBUG) {
					showId(assignVarStm.getIdentifier());
				}
				wrapperPoints.add(new CAssignVarStmPoint(assignVarStm));
				argIndex++;
			}

			// point for loop calls
			IProgramPoint callsPoint = new EntryPoint();
			wrapperPoints.add(callsPoint);

			// create call loop for all called methods
			Set<ICFG> callCFGs = new HashSet<ICFG>();
			for (ICFG calledCFG : calledCFGs) {
				List<IProgramPoint> callPoints = new LinkedList<IProgramPoint>();
				callPoints.add(new EntryPoint());

				AFuncDecl callFunc = (AFuncDecl) calledCFG.getEntry().getNode();
				String callName = calledCFG.getName();
				// assign args
				// a_j = temp-i-a_j
				argIndex = 0;
				for (PDecl formalDecl : callFunc.getFormals()) {
					String name = "$temp-" + callIndex + "-" + argIndex;
					TIdentifier id = getIdentifier(name);
					AAssignVarStm assignVarStm = new AAssignVarStm(
							new TAssign(), makeId((ALocalDecl) formalDecl),
							new AVarExp(id));
					if (DEBUG) {
						showId(((AVarExp) assignVarStm.getExp()).getIdentifier());
					}
					callPoints.add(new CAssignVarStmPoint(assignVarStm));
					argIndex++;
				}

				// point for call cfg
				IProgramPoint callCFGPoint = new EntryPoint();
				callPoints.add(callCFGPoint);

				// save return value
				// call-i = ref-f
				String call = "$call-" + callIndex;
				TIdentifier callId = getIdentifier(call);
				AAssignVarStm callAssignVarStm = new AAssignVarStm(
						new TAssign(), callId, new AVarExp(
								getIdentifier("$ret-" + callName)));
				if (DEBUG) {
					showId(callAssignVarStm.getIdentifier());
				}
				if (DEBUG) {
					showId(((AVarExp) callAssignVarStm.getExp()).getIdentifier());
				}
				callPoints.add(new CAssignVarStmPoint(callAssignVarStm));

				callPoints.add(new ExitPoint());

				CCFG callCFG = new CCFG(callName, callPoints);
				callCFG.replaceByCFG(callCFGPoint, calledCFG);
				callCFGs.add(callCFG);
			}
			CCFG callsCFG = new CCFG("calls", callCFGs);

			// restore context
			// b_j = save-i-b_j
			// x_j = save-i-x_j
			for (Map.Entry<String, ALocalDecl> entry : cfgLocalDeclMap.get(cfg).entrySet()) {
				String name = "$save-" + callIndex + "-" + entry.getKey();
				TIdentifier id = getIdentifier(name);
				AAssignVarStm assignVarStm = new AAssignVarStm(new TAssign(),
						makeId(entry.getValue()), new AVarExp(id));
				if (DEBUG) {
					showId(((AVarExp) assignVarStm.getExp()).getIdentifier());
				}
				wrapperPoints.add(new CAssignVarStmPoint(assignVarStm));
			}

			// save return value
			// x = call-i
			String call = "$call-" + callIndex;
			TIdentifier callId = getIdentifier(call);
			AAssignVarStm xAssignVarStm = new AAssignVarStm(new TAssign(),
					assignId.clone(), new AVarExp(callId.clone()));
			if (DEBUG) {
				showId(((AVarExp) xAssignVarStm.getExp()).getIdentifier());
			}
			wrapperPoints.add(new CAssignVarStmPoint(xAssignVarStm));

			wrapperPoints.add(new ExitPoint());

			CCFG wrapperCFG = new CCFG("wrapper", wrapperPoints);
			wrapperCFG.replaceByCFG(callsPoint, callsCFG);
			return wrapperCFG;
		}

		public @Override
		boolean caseCAssignVarStmPoint(AAssignVarStm node,
				CAssignVarStmPoint point) {
			if (node.getExp() instanceof AFunctionPtrCallExp) {
				AFunctionPtrCallExp callExp = (AFunctionPtrCallExp) node.getExp();
				ALocalDecl decl = (ALocalDecl) ((AVarExp) callExp.getExp()).getIdentifier().getDecl();
				Set<Target> targets = Main.pointsToMap.pt(point, decl);
				Set<ICFG> callCFGs = new HashSet<ICFG>();
				Map<DeclarationTarget, Integer> targetMap = callSiteToTargetMap.get(node);
				if (targetMap == null) {
					targetMap = new HashMap<DeclarationTarget, Integer>();
					callSiteToTargetMap.put(node, targetMap);
				}
				for (Target target : targets) {
					if (target instanceof DeclarationTarget) {
						PDecl targetDecl = ((DeclarationTarget) target).getTarget();
						Integer callSiteToTargetCount = targetMap.get(targetDecl);
						if (callSiteToTargetCount == null) {
							callSiteToTargetCount = 0;
						}
						callSiteToTargetCount++;
						targetMap.put((DeclarationTarget) target,
								callSiteToTargetCount);

						String callName = targetDecl.getName().getText();
						ICFG callCFG = subCollection.getCFG(callName);
						if (callCFG != null) {
							if (callSiteToTargetCount > maxCopies) {
								callCFGs.add(callCFG);
							} else {
								callCFGs.add(callCFG.clone());
							}
						} else {
							throw new SourceError("Unknown method '" + callName
									+ "'.");
						}
					}
				}
				cfg.replaceByCFG(point, linkMethods(node.getIdentifier(),
						callExp.getArgs(), callCFGs));
				return true;
			}
			return false;
		}
	}
}
