package tip.cfg.cfa;

import tip.analysis.*;
import tip.node.*;
import java.util.*;

/** 
 * 	Phase to normalize the AST such that 
 * 	<ul>
 * 	<li>all function calls performed in connection with assignments:
 *  	<pre>id_v = (id_f)(id_1,...,id_n);</pre>
 *	</li> 
 * 	<li>every pointer manipulation is one of the six kinds
 * 		<ol>
 * 		<li><code>id = malloc</code></li>
 * 		<li><code>id1 = &id2</code></li>
 * 		<li><code>id1 = id2</code></li>
 * 		<li><code>id1 = *id2</code></li>
 * 		<li><code>*id1 = id2</code></li>
 * 		<li><code>id = null</code></li>
 * 		</ol>
 *	</li>
 *	</ul>
 *
 *	@author Johnni Winther, jw@brics.dk
 */
/**
 * @author Johnni
 */
public class Normalization extends DepthFirstAdapter {

	// holds the current statement
	private PStm currentStm;

	private void setCurrentStm(PStm stm) {
		currentStm = stm;
	}

	// counter used to added pseudo-variables named $identifierCounter
	private int identifierCounter = 0;

	// list of all identifiers added in the current method
	private LinkedList<ALocalDecl> newIdentifiers = new LinkedList<ALocalDecl>();

	/**
	 * Add {@code stm} before the current statement. Special action has to be
	 * taking if the current is not in a block or if the current statement is a
	 * while statement.
	 * 
	 * @param stm
	 *            the statement to be added
	 */
	@SuppressWarnings({"unchecked","null"})
	private void addStmBefore(PStm stm) {
		Token token = currentStm.getToken();
		ABlock currentBlock = null;
		if (currentStm.parent() instanceof ABlock) { // we have a block already
			currentBlock = (ABlock) currentStm.parent();
		} else { // need to create a block
			List<PStm> statements = new LinkedList<PStm>();
			statements.add(currentStm);
			ABlock block = new ABlock(statements);
			ABlockStm blockStm = new ABlockStm(new TLBrace(token.getLine(),
					token.getPos()), block);
			currentStm.replaceBy(blockStm);
		}

		List<PStm> statements = (List<PStm>) currentBlock.getStatements().clone();
		int index = statements.indexOf(currentStm);

		// default insertion
		statements.add(index, stm);
		currentBlock.setStatements(statements);

		// not enough for AWhileStm - here stm should also be added to
		// end of the body block
		if (currentStm instanceof AWhileStm) {
			AWhileStm whileStm = (AWhileStm) currentStm;
			PStm whileBody = whileStm.getBody();
			if (whileBody instanceof ABlockStm) { // we have a block
				ABlock whileBlock = ((ABlockStm) whileBody).getBlock();
				List<PStm> whileStatements = (List<PStm>) whileBlock.getStatements().clone();
				whileStatements.add(stm.clone());
				whileBlock.setStatements(whileStatements);
			} else { // we need to create a block
				token = whileBody.getToken();
				List<PStm> whileStatements = new LinkedList<PStm>();
				whileStatements.add(whileBody);
				whileStatements.add(stm.clone());
				ABlock whileBlock = new ABlock(whileStatements);
				whileBody = new ABlockStm(new TLBrace(token.getLine(),
						token.getPos()), whileBlock);
				whileStm.setBody(whileBody);
			}

		}
	}

	/**
	 * Create a new linked {@link ALocalDecl} and {@link TIdentifier} and add
	 * the new {@link ALocalDecl} to {@code newIdentifiers} and returns the new
	 * {@link TIdentifier}
	 * 
	 * @return a new identifier
	 */
	private TIdentifier newIdentifier() {
		TIdentifier identifier = new TIdentifier("$" + (identifierCounter++));
		ALocalDecl localDecl = new ALocalDecl(identifier);
		identifier.setDecl(localDecl);
		newIdentifiers.add(localDecl);
		return identifier.clone();
	}

	/**
	 * Handles blocks of statements.
	 * 
	 * @param node
	 *            the current block
	 */
	@SuppressWarnings("unchecked")
	public @Override
	void caseABlock(ABlock node) {
		for (PStm stm : (List<PStm>) node.getStatements().clone()) {
			setCurrentStm(stm);
			stm.apply(this);
		}
	}

	/**
	 * Clears the list of new identifiers for this body.
	 * 
	 * @param node
	 *            the body node
	 */
	public @Override
	void inABody(ABody node) {
		newIdentifiers.clear();
	}

	/**
	 * Add all new identifiers to a ADeclStm in the start of the body.
	 * 
	 * @param node
	 *            the current body node
	 */
	@SuppressWarnings("unchecked")
	public @Override
	void outABody(ABody node) {
		if (newIdentifiers.size() != 0) {
			ADeclStm declStm = new ADeclStm();
			declStm.setToken(new TVar());
			declStm.setDecl((List<ALocalDecl>) newIdentifiers.clone());
			ABlock block = node.getBlock();
			List<PStm> statements = (List<PStm>) block.getStatements().clone();
			statements.add(0, declStm);
			block.setStatements(statements);
			newIdentifiers.clear();
		}
	}

	/**
	 * Normalizes function call expressions:
	 * 
	 * <pre>
	 * E -&gt; f(E1,...,En) =&gt;
	 * 		x1 = E1;
	 * 		...
	 * 		xn = En
	 * 		xf = f;
	 * 		r = (xf)(x1,...,xn)
	 * 		E -&gt; r
	 * </pre>
	 * 
	 * @param node
	 *            the current function call expression
	 */

	public @Override
	void outAFunctionCallExp(AFunctionCallExp node) {
		List<PExp> arguments = new LinkedList<PExp>();
		for (PExp E : node.getArgs()) {
			if (!(E instanceof AVarExp)) {
				TIdentifier x = newIdentifier();
				AVarExp x_exp = new AVarExp(x.clone());
				AAssignVarStm x_assign = new AAssignVarStm(new TAssign(), x,
						E.clone());
				addStmBefore(x_assign);
				arguments.add(x_exp);
			} else {
				arguments.add(E);
			}
		}

		TIdentifier f = newIdentifier();
		AVarExp f_exp = new AVarExp(f.clone());
		AAssignVarStm f_assign = new AAssignVarStm(new TAssign(), f,
				new AVarExp(node.getIdentifier().clone()));
		addStmBefore(f_assign);

		AFunctionPtrCallExp f_call = new AFunctionPtrCallExp(f_exp, arguments);

		if (!(node.parent() instanceof AAssignVarStm)) {
			TIdentifier r = newIdentifier();
			AVarExp r_exp = new AVarExp(r.clone());
			AAssignVarStm r_assign = new AAssignVarStm(new TAssign(), r, f_call);
			addStmBefore(r_assign);
			node.replaceBy(r_exp);
		} else {
			node.replaceBy(f_call);
		}
	}

	/**
	 * Normalizes pointer expressions:
	 * 
	 * <pre>
	 * E -&gt; &amp;id : E \notin S -&gt; id1 = &amp;id2 =&gt;
	 * 
	 * 		S -&gt; x = &amp;id2
	 * 		E -&gt; x
	 * </pre>
	 * 
	 * @param node
	 *            the current pointer expression
	 */

	public @Override
	void outAPointerExp(APointerExp node) {
		if (!(node.parent() instanceof AAssignVarStm)) {
			TIdentifier id = newIdentifier();
			AVarExp varExp = new AVarExp(id.clone());
			AAssignVarStm varAssignStm = new AAssignVarStm(new TAssign(), id,
					node.clone());
			node.replaceBy(varExp);
			addStmBefore(varAssignStm);
		}
	}

	/**
	 * Normalizes deref expressions:
	 * 
	 * <pre>
	 * 	E -&gt; *E1 : E \notin S -&gt; id = E =&gt;
	 * 
	 * 		S -&gt; x = *E1
	 * 		E -&gt; x
	 * </pre>
	 * 
	 * @param node
	 *            the current deref expression
	 */

	public @Override
	void outADerefExp(ADerefExp node) {
		if (!(node.parent() instanceof AAssignVarStm)) {
			TIdentifier id = newIdentifier();
			AVarExp varExp = new AVarExp(id.clone());
			AAssignVarStm varAssignStm = new AAssignVarStm(new TAssign(), id,
					node.clone());
			node.replaceBy(varExp);
			addStmBefore(varAssignStm);
		}
	}

	/**
	 * Normalizes malloc expressions:
	 * 
	 * <pre>
	 * E -&gt; malloc : E \notin S -&gt; id = E =&gt;
	 * 		
	 * 		S -&gt; x = malloc
	 * 		E -&gt; x
	 * </pre>
	 * 
	 * @param node
	 *            the current malloc expression
	 */

	public @Override
	void outAMallocExp(AMallocExp node) {
		if (!(node.parent() instanceof AAssignVarStm)) {
			TIdentifier id = newIdentifier();
			AVarExp varExp = new AVarExp(id.clone());
			AAssignVarStm varAssignStm = new AAssignVarStm(new TAssign(), id,
					node.clone());
			node.replaceBy(varExp);
			addStmBefore(varAssignStm);
		}
	}

	/**
	 * Normalizes pointer assignments:
	 * 
	 * <pre>
	 * 	S -&gt; *E1 = E2 : E1 != id =&gt;
	 * 
	 * 		S1 -&gt; x = E2
	 * 		S2 -&gt; y = E1
	 * 		S -&gt; *y = x
	 * 
	 * 	S -&gt; *E1 = E2 : E1 = id, E2 != id =&gt;
	 * 
	 * 		S1 -&gt; x = E2
	 * 		S -&gt; *E1 = x
	 * </pre>
	 * 
	 * @param node
	 *            the current pointer assignment
	 */

	public @Override
	void inAAssignPtrStm(AAssignPtrStm node) {
		PExp E1 = node.getLeft();
		PExp E2 = node.getRight();
		if (!(E1 instanceof AVarExp)) {
			TIdentifier x = newIdentifier();
			AVarExp x_exp = new AVarExp(x.clone());
			AAssignVarStm S1 = new AAssignVarStm(new TAssign(), x, E2.clone());
			TIdentifier y = newIdentifier();
			AVarExp y_exp = new AVarExp(y.clone());
			AAssignVarStm S2 = new AAssignVarStm(new TAssign(), y, E1.clone());
			node.setLeft(y_exp);
			node.setRight(x_exp);
			addStmBefore(S1);
			addStmBefore(S2);

			setCurrentStm(S1);
			S1.apply(this);
			setCurrentStm(S2);
			S2.apply(this);
			setCurrentStm(node);
		} else if (!(E2 instanceof AVarExp)) {
			TIdentifier x = newIdentifier();
			AVarExp x_exp = new AVarExp(x.clone());
			AAssignVarStm S1 = new AAssignVarStm(new TAssign(), x, E2.clone());
			node.setRight(x_exp);
			addStmBefore(S1);

			setCurrentStm(S1);
			S1.apply(this);
			setCurrentStm(node);
		}
	}

	/**
	 * Normalizes return statements:
	 * 
	 * <pre>
	 * S -&gt; return E : E != id =&gt;
	 * 	
	 * 		S1 -&gt; x = E
	 * 		S -&gt; return x
	 * </pre>
	 * 
	 * @param node
	 *            the current return statement
	 */
	public @Override
	void inAReturnStm(AReturnStm node) {
		PExp E = node.getExp();
		if (!(E instanceof AVarExp)) {
			TIdentifier x = newIdentifier();
			AVarExp x_exp = new AVarExp(x.clone());
			AAssignVarStm S1 = new AAssignVarStm(new TAssign(), x, E.clone());
			node.setExp(x_exp);
			addStmBefore(S1);

			setCurrentStm(S1);
			S1.apply(this);
			setCurrentStm(node);
		}
	}

	/**
	 * Handles if-then statements. Need to manual set current statement for the
	 * body statement - it might not be a ABlockStm.
	 * 
	 * @param node
	 *            the current if-then statement
	 */
	public @Override
	void caseAIfThenStm(AIfThenStm node) {
		setCurrentStm(node);
		node.getCondition().apply(this);
		setCurrentStm(node.getBody());
		node.getBody().apply(this);
	}

	/**
	 * Handles if-then-else statements. Need to manually set current statement
	 * for the body statements - it might not be a ABlockStm.
	 * 
	 * @param node
	 *            the current if-then-else statement
	 */
	public @Override
	void caseAIfThenElseStm(AIfThenElseStm node) {
		setCurrentStm(node);
		node.getCondition().apply(this);
		setCurrentStm(node.getThenBody());
		node.getThenBody().apply(this);
		setCurrentStm(node.getElseBody());
		node.getElseBody().apply(this);
	}

	/**
	 * Handles while statements. Need to manually set current statement for the
	 * body statement - it might not be a ABlockStm.
	 * 
	 * @param node
	 *            the current while statement
	 */
	public @Override
	void caseAWhileStm(AWhileStm node) {
		setCurrentStm(node);
		node.getCondition().apply(this);
		setCurrentStm(node.getBody());
		node.getBody().apply(this);
	}
}
