package tip.cfg.cfa;

import java.util.*;

import tip.analysis.*;
import tip.node.*;
import tip.cfg.*;
import tip.display.*;

public class Steensgaard extends DepthFirstAdapter implements PointsToMap {
	public static final boolean DEBUG = false;

	private Map<PDecl, Location> declToLocationMap = new HashMap<PDecl, Location>();

	private Map<ALocalDecl, Set<Target>> pointsToMap = new HashMap<ALocalDecl, Set<Target>>();

	private Location getLocation(PDecl decl) {
		Location tau = declToLocationMap.get(decl);
		if (tau == null) {
			tau = new Location(new DeclarationTarget(decl));
			declToLocationMap.put(decl, tau);
		}
		return tau.getRoot();
	}

	private static class Location extends UnionFind {
		private Set<Target> phi = new HashSet<Target>();

		private Location alpha = null; // alpha = bot

		private Set<Location> pending = new HashSet<Location>();

		public Location() {

		}

		public Location(Target target) {
			phi.add(target);
		}

		public @Override
		Location getRoot() {
			return (Location) super.getRoot();
		}

		public @Override
		Location union(UnionFind e) {
			return (Location) super.union(e);
		}

		public Set<Target> getIds() {
			return phi;
		}

		public Set<Target> getPtrIds() {
			if (alpha == null) {
				return Collections.emptySet();
			} else {
				return alpha.getIds();
			}
		}

		public Location getValue() {
			Location root = getRoot();
			return root.alpha == null ? null : root.alpha.getRoot();
		}

		public void setValue(Location alpha) {
			assert alpha != null : "setValue(null)";
			alpha = alpha.getRoot();
			Location root = getRoot();
			if (root.alpha == null) {
				root.alpha = alpha;
				if (pending != null) {
					for (Location x : pending) {
						root.union(x);
					}
					pending = null;
				}
			} else {
				root.alpha = root.alpha.union(alpha);
			}
		}

		protected @Override
		void unify(UnionFind _root) {
			Location root = (Location) _root;
			root.phi.addAll(phi);
			phi = null;
			if (root.alpha == null) { // use alpha
				if (alpha == null) { // join the pending lists
					root.pending.addAll(pending);
					pending = null;
				} else { // point to alpha
					root.alpha = alpha;
				}
			} else { // use root.alpha
				if (alpha == null) { // resolve the pending list
					for (Location x : pending) {
						root.alpha.union(x);
					}
					pending = null;
				} else { // join the two locations
					root.alpha = root.alpha.union(alpha);
				}
			}
			alpha = null;
		}

		@Override
		public String rootToString() {
			StringBuffer sb = new StringBuffer();
			sb.append(locationName());
			if (pending != null && pending.size() > 0) {
				String prefix = "";
				sb.append("(");
				for (Location x : pending) {
					sb.append(prefix);
					sb.append(x.locationName());
					prefix = ",";
				}
				sb.append(")");
			}
			if (alpha != null) {
				sb.append("->");
				sb.append(alpha);
			}
			return sb.toString();
		}

		public String locationName() {
			Location root = getRoot();
			StringBuffer sb = new StringBuffer();
			String prefix = "";
			sb.append("[");
			for (Target t : root.phi) {
				sb.append(prefix);
				if (t instanceof DeclarationTarget) {
					sb.append(((DeclarationTarget) t).getTarget().getName().getText());
				} else {
					sb.append(t);
				}
				prefix = ",";
			}
			sb.append("]");
			return sb.toString();
		}

		/**
		 * ptr(t) <= this.alpha
		 */
		public void setPtr(Location t) {
			Location x = getRoot();
			t = t.getRoot();
			if (DEBUG) {
				System.out.println("ptr(" + t.locationName() + ") <= "
						+ x.locationName() + ".alpha");
			}
			setValue(t);
			if (DEBUG) {
				System.out.println(x);
				System.out.println(t);
			}
		}

		/**
		 * y.alpha <= this.alpha
		 */
		public void setLe(Location y) {
			Location x = getRoot();
			y = y.getRoot();
			if (DEBUG) {
				System.out.println(y.locationName() + ".alpha <= "
						+ x.locationName() + ".alpha");
			}
			if (y.alpha == null) {
				y.pending.add(this);
			} else {
				x.setValue(x.getValue().union(y.alpha));
			}
			if (DEBUG) {
				System.out.println(x);
				System.out.println(y);
			}
		}

		/**
		 * y.ptr.alpha <= this.alpha
		 */
		public void setYptrLe(Location y) {
			Location x = getRoot();
			y = y.getRoot();
			if (DEBUG) {
				System.out.println(y.locationName() + ".ptr.alpha <= "
						+ x.locationName() + ".alpha");
			}
			Location y_ptr = y.getValue();
			if (y_ptr == null) {
				y_ptr = new Location(); // dummy location
				y.setPtr(y_ptr);
			}
			setLe(y_ptr);
			if (DEBUG) {
				System.out.println(x);
				System.out.println(y);
				System.out.println(y_ptr);
			}
		}

		/**
		 * y.alpha <= this.ptr.alpha
		 */
		public void setXptrLe(Location y) {
			Location x = getRoot();
			y = y.getRoot();
			if (DEBUG) {
				System.out.println(y.locationName() + ".alpha <= "
						+ x.locationName() + ".ptr.alpha");
			}
			Location x_ptr = x.getValue();
			if (x_ptr == null) {
				x_ptr = new Location(); // dummy location
				setPtr(x_ptr);
			}
			x_ptr.setLe(y);
			if (DEBUG) {
				System.out.println(x);
				System.out.println(y);
				System.out.println(x_ptr);
			}
		}
	}

	public @Override
	void inAAssignVarStm(AAssignVarStm node) {
		PExp exp = node.getExp();
		if (exp instanceof AMallocExp) {
			if (DEBUG) {
				System.out.println(node.apply(new NodeText<Node>()));
			}
			// x = malloc
			PDecl decl = node.getIdentifier().getDecl();
			Location x = getLocation(decl);
			Location malloc = new Location(new MallocTarget((AMallocExp) exp));
			x.setPtr(malloc);
		} else if (exp instanceof APointerExp) {
			if (DEBUG) {
				System.out.println(node.apply(new NodeText<Node>()));
			}
			// x = &y
			PDecl decl1 = node.getIdentifier().getDecl();
			PDecl decl2 = ((APointerExp) exp).getIdentifier().getDecl();
			Location x = getLocation(decl1);
			Location y = getLocation(decl2);
			x.setPtr(y);
		} else if (exp instanceof AVarExp) {
			if (DEBUG) {
				System.out.println(node.apply(new NodeText<Node>()));
			}
			// x = y
			PDecl decl1 = node.getIdentifier().getDecl();
			PDecl decl2 = ((AVarExp) exp).getIdentifier().getDecl();
			Location x = getLocation(decl1);
			Location y = getLocation(decl2);
			x.setLe(y);
		} else if (exp instanceof ADerefExp) {
			if (DEBUG) {
				System.out.println(node.apply(new NodeText<Node>()));
			}
			// x = *y
			PDecl decl1 = node.getIdentifier().getDecl();
			PDecl decl2 = ((AVarExp) ((ADerefExp) exp).getExp()).getIdentifier().getDecl();
			Location x = getLocation(decl1);
			Location y = getLocation(decl2);
			x.setYptrLe(y);
		}
	}

	public @Override
	void inAAssignPtrStm(AAssignPtrStm node) {
		if (DEBUG) {
			System.out.println(node.apply(new NodeText<Node>()));
		}
		// *x = y
		AVarExp leftExp = (AVarExp) node.getLeft();
		AVarExp rightExp = (AVarExp) node.getRight();
		PDecl decl1 = leftExp.getIdentifier().getDecl();
		PDecl decl2 = rightExp.getIdentifier().getDecl();
		Location x = getLocation(decl1);
		Location y = getLocation(decl2);
		x.setXptrLe(y);
	}

	public @Override @SuppressWarnings("unused")
	void outAProgram(AProgram node) {
		if (DEBUG) {
			System.out.println(declToLocationMap);
		}
	}

	public String getName() {
		return "Steensgaard";
	}

	/**
	 * Returns the pointer targets for the variable {@code decl} at the program
	 * point {@code point} as computed by Steensgaard's algorithm.
	 * 
	 * @param point
	 *            the program point for which the pointer targets are returned
	 * @param decl
	 *            the variable for which the pointer targets are returned
	 * @return the pointer target for the {@code decl} at {@code point}
	 */
	public Set<Target> pt(IProgramPoint point, ALocalDecl decl) {
		Set<Target> targetSet = pointsToMap.get(decl);
		if (targetSet == null) {
			targetSet = getLocation(decl).getPtrIds();
			pointsToMap.put(decl, targetSet);
		}
		return targetSet;
	}

	public boolean isFlowSensitive() {
		return false;
	}

	private Map<Target, Set<Target>> ptMap;

	@SuppressWarnings("unused")
	public Map<Target, Set<Target>> pt(IProgramPoint point) {
		if (ptMap == null) {
			ptMap = new HashMap<Target, Set<Target>>();
			if (pointsToMap.isEmpty()) {
				for (PDecl decl : declToLocationMap.keySet()) {
					if (decl instanceof ALocalDecl) {
						pt(null, (ALocalDecl) decl);
					}
				}
			}
			for (PDecl decl : pointsToMap.keySet()) {
				Set<Target> targets = pointsToMap.get(decl);
				if (targets != null && !targets.isEmpty()) {
					ptMap.put(new DeclarationTarget(decl), targets);
				}
			}
		}
		return ptMap;
	}
}
