package tip.cfg.dfa;

import tip.cfg.*;
import tip.node.*;
import tip.util.*;
import tip.display.*;

/**
 * {@code AvailableExpressionsAnalysis} implements the Available Expressions
 * <em>forwards-must</em> analysis. It utilizes the {@link ExpressionEqualitor}
 * for definining equality on expression.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class AvailableExpressionsAnalysis extends
		FlowAnalysis<EqualitySet<ABinopExp>> {
	/**
	 * Creates a new {@code AvailableExpressionsAnalysis} analysis.
	 */
	public AvailableExpressionsAnalysis() {
		// set the dependency strategy to the forwards strategy as this
		// analysis is a forwards analysis
		super(new ForwardsStrategy(), "Available Expressions");
	}

	/**
	 * The {@link NodeFilter} used for retrieving {@link ABinopExp} children of
	 * a node.
	 */
	private NodeFilter<ABinopExp> filter = new SubTypeFilter<ABinopExp>(
			ABinopExp.class);

	/**
	 * Creates the lattice to use for the control flow graph {@code cfg}. The
	 * lattice is an {@link IntersectionSetLattice} holdning {@link ABinopExp}
	 * nodes using the {@link ExpressionEqualitor} for determining equality
	 * between nodes. The lattice is initialize to contain all binary
	 * expressions occuring in the associated nodes of the control flow graph
	 * {@code cfg}.
	 * 
	 * @param cfg
	 *            the control flow graph for which the lattice is created
	 * @return the lattice to use for this analysis
	 */
	@Override
	protected Lattice<EqualitySet<ABinopExp>> createLattice(ICFG cfg) {
		// create the lattice
		return new IntersectionSetLattice<ABinopExp>(
		// retrieve the binary expressions in this control flow graph
				cfg.getNodeDescendants(filter),
				// use the ExpressionEqualitor for equality on expressions
				new ExpressionEqualitor());
	}

	/**
	 * Computes the variable for the entry point {@code point}.
	 * 
	 * @param node
	 *            the associated node
	 * @param point
	 *            the entry point in question
	 * @return {@code true} if the variable for {@code point} changed
	 */
	public @Override
	boolean caseEntryPoint(Node node, EntryPoint point) {
		EqualitySet<ABinopExp> top = getLattice().getSmallestElement();
		top.clear();
		return setVariable(point, top);
	}

	/**
	 * Computes the variable the program associated with the assignment
	 * statement {@code node}.
	 * 
	 * @param node
	 *            the assignment statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getExp().getDescendants(join, filter);
		EqualitySet<ABinopExp> downarrow = join.clone();
		for (ABinopExp exp : join) {
			if (DeclCollector.collectDecls(exp).contains(
					node.getIdentifier().getDecl())) {
				downarrow.remove(exp);
			}
		}
		return setVariable(point, downarrow);
	}

	/**
	 * Computes the variable the program associated with the if-then statement
	 * {@code node}.
	 * 
	 * @param node
	 *            the if-then statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCIfThenStmPoint(AIfThenStm node, CIfThenStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getCondition().getDescendants(join, filter);
		return setVariable(point, join);
	}

	/**
	 * Computes the variable the program associated with the if-then-else
	 * statement {@code node}.
	 * 
	 * @param node
	 *            the if-then-else statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCIfThenElseStmPoint(AIfThenElseStm node,
			CIfThenElseStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getCondition().getDescendants(join, filter);
		return setVariable(point, join);
	}

	/**
	 * Computes the variable the program associated with the while statement
	 * {@code node}.
	 * 
	 * @param node
	 *            the while statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCWhileStmPoint(AWhileStm node, CWhileStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getCondition().getDescendants(join, filter);
		return setVariable(point, join);
	}

	/**
	 * Computes the variable the program associated with the output statement
	 * {@code node}.
	 * 
	 * @param node
	 *            the output statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCOutputStmPoint(AOutputStm node, COutputStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getExp().getDescendants(join, filter);
		return setVariable(point, join);
	}

	/**
	 * Computes the variable the program associated with the return statement
	 * {@code node}.
	 * 
	 * @param node
	 *            the return statement in question
	 * @param point
	 *            the point for which the variable is computed
	 * @return {@code true} if the variable for {@code point} has changed
	 */
	public @Override
	boolean caseCReturnStmPoint(AReturnStm node, CReturnStmPoint point) {
		EqualitySet<ABinopExp> join = join(point);
		node.getExp().getDescendants(join, filter);
		return setVariable(point, join);
	}

	/**
	 * Returns a textual representation of {@code variable}. {@link NodeText} is
	 * passed to {@link EqualitySet#toString(Stringifier)} in order to let
	 * {@link EqualitySet} print each element as defined by {@link NodeText}.
	 * 
	 * @param variable
	 *            the variable for which a textual representation is returned
	 * @return the textual representation of {@code variable}
	 */
	public String getVariableString(EqualitySet<ABinopExp> variable) {
		return variable.toString(new NodeText<ABinopExp>());
	}
}
