package tip.cfg.dfa;

import java.util.*;

import tip.*;
import tip.display.*;
import tip.cfg.*;
import tip.node.*;
import tip.util.*;
import tip.analysis.*;

public class ConstantPropagationAnalysis
		extends
		FlowAnalysis<EqualityMap<ALocalDecl, ConstantPropagationAnalysis.Const>> {
	public ConstantPropagationAnalysis() {
		super(new ForwardsStrategy(), "Constant Propagation");
	}

	public enum Type {
		UNKNOWN, CONSTANT, VARIABLE
	}

	public static class Const {
		private Type type;

		private int value;

		private Const(Type type) {
			this.type = type;
			this.value = 0;
		}

		public Const(int value) {
			this.type = Type.CONSTANT;
			this.value = value;
		}

		public int getValue() {
			return value;
		}

		public Type getType() {
			return type;
		}

		public boolean isUnknown() {
			return type == Type.UNKNOWN;
		}

		public boolean isConstant() {
			return type == Type.CONSTANT;
		}

		public boolean isVariable() {
			return type == Type.VARIABLE;
		}

		public @Override
		String toString() {
			switch (type) {
			case VARIABLE:
				return "?";
			case CONSTANT:
				return "" + value;
			case UNKNOWN:
				return "-";
			}
			throw new IllegalArgumentException();
		}

		public static final Const UNKNOWN = new Const(Type.UNKNOWN);

		public static final Const VARIABLE = new Const(Type.VARIABLE);
	}

	public static class ConstLattice implements Lattice<Const> {
		public Const getSmallestElement() {
			return Const.UNKNOWN;
		}

		public Const getLeastUpperBound(Const a, Const b) {
			if (a == null && b == null) {
				return Const.UNKNOWN;
			} else if (a == null) {
				return b;
			} else if (b == null) {
				return a;
			}

			switch (a.getType()) {
			case VARIABLE:
				return a;
			case CONSTANT:
				switch (b.getType()) {
				case UNKNOWN:
					return a;
				case CONSTANT:
					if (a.getValue() == b.getValue()) {
						return a;
					} else {
						return Const.VARIABLE;
					}
				default:
				}
				return Const.VARIABLE;
			case UNKNOWN:
				return b;
			}
			throw new IllegalArgumentException();
		}

		public boolean equals(Const a, Const b) {
			return (a == null || b == null) ? a == b
					: (a.getType() == b.getType())
							&& (a.getType() != Type.CONSTANT || a.getValue() == b.getValue());
		}

		public int hashCode(Const a) {
			return a == null ? 0 : (a.getType() == Type.CONSTANT ? a.getValue()
					: a.getType().hashCode());
		}
	}

	private Evaluator evaluator = new Evaluator();

	@Override
	protected Lattice<EqualityMap<ALocalDecl, Const>> createLattice(ICFG cfg) {
		return new MapLattice<ALocalDecl, Const>(
				LocalDeclCollector.collectLocalDecls(cfg), new ConstLattice());
	}

	public static class Evaluator extends
			QuestionAnswerAdapter<Map<ALocalDecl, Const>, Const> {
		@Override @SuppressWarnings("unused")
		public Const defaultNode(Node node, Map<ALocalDecl, Const> env) {
			return Const.VARIABLE;
		}

		@Override
		public Const caseABinopExp(ABinopExp e, Map<ALocalDecl, Const> env) {
			Const left = e.getLeft().apply(this, env);
			Const right = e.getRight().apply(this, env);
			boolean oneVariable = left.isVariable() || right.isVariable();
			boolean bothVariable = left.isVariable() && right.isVariable();
			boolean oneUnknown = left.isUnknown() || right.isUnknown();
			switch (e.getBinop().kindPBinop()) {
			case PLUS:
				if (oneVariable) {
					return Const.VARIABLE;
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					return new Const(left.getValue() + right.getValue());
				}
			case MINUS:
				if (oneVariable) {
					return Const.VARIABLE;
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					return new Const(left.getValue() - right.getValue());
				}
			case TIMES:
				if (bothVariable) {
					return Const.VARIABLE;
				} else if (oneVariable) {
					if (oneUnknown) {
						return Const.UNKNOWN;
					} else if (left.isConstant() && left.getValue() == 0) {
						return new Const(0);
					} else if (right.isConstant() && right.getValue() == 0) {
						return new Const(0);
					} else { // no constant == 0
						return Const.VARIABLE;
					}
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					return new Const(left.getValue() * right.getValue());
				}
			case DIVIDE:
				if (bothVariable) {
					return Const.VARIABLE;
				} else if (oneVariable) {
					if (oneUnknown) {
						return Const.UNKNOWN;
					} else if (left.isConstant() && left.getValue() == 0) {
						return new Const(0);
					} else if (right.isConstant() && right.getValue() == 0) {
						return Const.VARIABLE; // known division by zero
					} else { // no constant == 0
						return Const.VARIABLE;
					}
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					if (right.getValue() == 0) {
						return Const.VARIABLE; // known division by zero
					} else { // safe to evaluate
						return new Const(left.getValue() / right.getValue());
					}
				}
			case EQ:
				if (oneVariable) {
					return Const.VARIABLE;
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					return new Const(left.getValue() == right.getValue() ? 1
							: 0);
				}
			case GT:
				if (oneVariable) {
					return Const.VARIABLE;
				} else if (oneUnknown) {
					return Const.UNKNOWN;
				} else { // both constant
					return new Const(left.getValue() > right.getValue() ? 1 : 0);
				}
			}
			throw new InternalCompilerError("Unknown binop type : "
					+ e.getBinop().kindPBinop());
		}

		@Override @SuppressWarnings("unused")
		public Const caseAIntConstExp(AIntConstExp e, Map<ALocalDecl, Const> env) {
			return new Const(e.getValue());
		}

		public @Override
		Const caseAVarExp(AVarExp e, Map<ALocalDecl, Const> env) {
			Const c = Const.UNKNOWN;
			PDecl decl = e.getIdentifier().getDecl();
			if (decl instanceof ALocalDecl) {
				c = env.get(decl);

			}
			if (c == null) {
				c = Const.UNKNOWN;
			}
			return c;
		}
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualityMap<ALocalDecl, Const> join = join(point);
		ALocalDecl local = (ALocalDecl) (node.getIdentifier().getDecl());
		Const result = node.getExp().apply(evaluator, join);
		join.put(local, result);
		return setVariable(point, join);
	}

	public @Override
	boolean caseCDeclStmPoint(ADeclStm node, CDeclStmPoint point) {
		EqualityMap<ALocalDecl, Const> join = join(point);
		for (PDecl decl : node.getDecl()) {
			if (decl instanceof ALocalDecl) {
				join.put((ALocalDecl) decl, Const.VARIABLE);
			}
		}
		return setVariable(point, join);
	}

	@Override @SuppressWarnings("unused")
	public boolean caseEntryPoint(Node node, EntryPoint point) {
		EqualityMap<ALocalDecl, Const> join = join(point);
		for (ALocalDecl decl : join.keySet()) {
			join.put(decl, Const.VARIABLE);
		}
		return setVariable(point, join);
	}

	public String getVariableString(EqualityMap<ALocalDecl, Const> variable) {
		return variable.toString(new NodeText<ALocalDecl>(),
				new DefaultStringifier<Const>());
	}

	public @Override
	String getVariableStringForPoint(IProgramPoint point, int distance) {
		EqualityMap<ALocalDecl, Const> variable = getVariable(point);
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		boolean first = true;
		for (ALocalDecl decl : LocalDeclCollector.collectLocalDecls(point.getVicinity(distance))) {
			Const c = variable.get(decl);
			if (c == null) {
				c = Const.UNKNOWN;
			}
			if (!first) {
				sb.append(",");
			}
			sb.append(decl.getName().getText());
			sb.append("=");
			sb.append(c);
			first = false;
		}
		sb.append("]");
		return sb.toString();
	}
}
