package tip.cfg.dfa;

import java.util.*;

import tip.node.*;
import tip.util.*;
import tip.cfg.*;
import tip.analysis.*;
import tip.display.*;

public class IntervalAnalysis extends
		FlowAnalysis<EqualityMap<ALocalDecl, IntervalAnalysis.Interval>>
		implements
		WideningNarrowingAnalysis<EqualityMap<ALocalDecl, IntervalAnalysis.Interval>> {
	public IntervalAnalysis() {
		super(new ForwardsStrategy(), "Interval");
	}

	public static class Interval implements Lattice<Interval> {
		// left == null => left = -\infty
		// right == null => right = \infty
		// bot is represented as Interval = null

		public Integer l;

		public Integer h;

		public Interval() {
			l = h = null;
		}

		public Interval(Integer _left, Integer _right) {
			l = _left;
			h = _right;
		}

		@Override
		public boolean equals(Object o) {
			if (o instanceof Interval) {
				Interval i = (Interval) o;
				boolean result = true;
				if (l != null) {
					result = l.equals(i.l);
				} else {
					result = l == i.l;
				}
				if (h != null) {
					result = h.equals(i.h);
				} else {
					result = h == i.h;
				}
				return result;
			}
			return false;
		}

		@Override
		public int hashCode() {
			int hash_left = 11;
			int hash_right = 17;
			if (l != null) {
				hash_left = l.hashCode();
			}
			if (h != null) {
				hash_right = h.hashCode();
			}
			return hash_left * 23 + hash_right * 29;
		}

		@Override
		public String toString() {
			StringBuffer sb = new StringBuffer();
			sb.append("[");
			if (l == null) {
				sb.append("-\u221e");
			} else {
				sb.append(l);
			}
			sb.append(",");
			if (h == null) {
				sb.append("\u221e");
			} else {
				sb.append(h);
			}
			sb.append("]");
			return sb.toString();
		}

		public Interval getSmallestElement() {
			return null;
		}

		public Interval getLeastUpperBound(Interval a, Interval b) {
			if (a == null) {
				return b;
			}
			if (b == null) {
				return a;
			}
			Integer left = a.l;
			Integer right = a.h;
			if (left != null) { // a.left != -\infty => left =
								// min(a.left,b.left)
				if (b.l == null) {
					left = null;
				} else {
					left = Math.min(left, b.l);
				}
			}
			if (right != null) { // a.right != \infty => right =
									// max(a.right,b.right)
				if (b.h == null) {
					right = null;
				} else {
					right = Math.max(right, b.h);
				}
			}
			return new Interval(left, right);
		}

		public boolean equals(Interval a, Interval b) {
			return a == null ? b == null : a.equals(b);
		}

		public int hashCode(Interval a) {
			return a == null ? 0 : a.hashCode();
		}
	}

	protected boolean widening = false;

	protected int counter = 0;

	protected Integer[] finiteSet;

	protected int MAX_NARROWING_FACTOR = 10;

	protected int MAX_NARROWING = 1;

	@Override
	protected Lattice<EqualityMap<ALocalDecl, Interval>> createLattice(ICFG cfg) {
		MAX_NARROWING = MAX_NARROWING_FACTOR * cfg.getProgramPoints().size();
		Collection<AIntConstExp> constExps = cfg.getEntry().getNode().getDescendants(
				AIntConstExp.class);
		LinkedList<Integer> integerList = new LinkedList<Integer>();
		for (AIntConstExp constExp : constExps) {
			integerList.add(constExp.getValue());
		}
		Collections.sort(integerList);
		finiteSet = integerList.toArray(new Integer[0]);
		return new MapLattice<ALocalDecl, Interval>(
				LocalDeclCollector.collectLocalDecls(cfg), new Interval());
	}

	public boolean isWidening() {
		return widening;
	}

	public void setWidening() {
		this.widening = true;
	}

	public boolean isNarrowing() {
		return !widening;
	}

	public void setNarrowing() {
		if (this.widening) {
			counter = 0;
		}
		this.widening = false;
	}

	public @Override @SuppressWarnings("unused")
	boolean defaultNode(Node node, IProgramPoint point) {
		counter++;
		EqualityMap<ALocalDecl, Interval> join = join(point);
		if (widening) {
			for (ALocalDecl decl : join.keySet()) {
				join.put(decl, widen(join.get(decl)));
			}
			return setVariable(point, join);
		} else {
			/* return false; */
			return setVariable(point, join) && counter <= MAX_NARROWING;
		}
	}

	public @Override
	boolean caseCDeclStmPoint(ADeclStm node, CDeclStmPoint point) {
		counter++;
		EqualityMap<ALocalDecl, Interval> join = join(point);
		for (PDecl decl : node.getDecl()) {
			join.put((ALocalDecl) decl, new Interval(null, null));
		}
		if (widening) {
			for (ALocalDecl decl : join.keySet()) {
				join.put(decl, widen(join.get(decl)));
			}
			return setVariable(point, join);
		} else {
			/* return false; */
			return setVariable(point, join) && counter <= MAX_NARROWING;
		}
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		counter++;
		EqualityMap<ALocalDecl, Interval> join = join(point);
		join.put((ALocalDecl) node.getIdentifier().getDecl(), eval(join,
				node.getExp()));
		if (widening) {
			for (ALocalDecl decl : join.keySet()) {
				join.put(decl, widen(join.get(decl)));
			}
			return setVariable(point, join);
		} else {
			/* return false; */
			return setVariable(point, join) && counter <= MAX_NARROWING;
		}
	}

	public @Override @SuppressWarnings("unused")
	boolean caseEntryPoint(Node node, EntryPoint point) {
		EqualityMap<ALocalDecl, Interval> join = getLattice().getSmallestElement();
		for (ALocalDecl decl : join.keySet()) {
			join.put(decl, new Interval(null, null));
		}
		return setVariable(point, join);
	}

	public Interval widen(Interval interval) {
		if (interval == null) {
			return null;
		}
		Integer left = null;
		Integer right = null;
		if (interval.l != null) { // interval.left != -\infty
			int i = 0;
			while (i < finiteSet.length && finiteSet[i] <= interval.l) {
				left = finiteSet[i];
				i++;
			}
		}
		if (interval.h != null) { // interval.right != \infty
			int i = finiteSet.length - 1;
			while (i >= 0 && finiteSet[i] >= interval.h) {
				right = finiteSet[i];
				i--;
			}
		}
		return new Interval(left, right);
	}

	public String getVariableString(EqualityMap<ALocalDecl, Interval> variable) {
		return variable == null ? "" : variable.toString(
				new NodeText<ALocalDecl>(), new DefaultStringifier<Interval>());
	}

	public @Override
	String getVariableStringForPoint(IProgramPoint point, int distance) {
		EqualityMap<ALocalDecl, Interval> variable = getVariable(point);
		return variable.toString(
				LocalDeclCollector.collectLocalDecls(point.getVicinity(distance)),
				new NodeText<ALocalDecl>(), new DefaultStringifier<Interval>());
	}

	public static class Evaluator extends
			QuestionAnswerAdapter<EqualityMap<ALocalDecl, Interval>, Interval> {
		@Override @SuppressWarnings("unused")
		public Interval defaultNode(Node node, EqualityMap<ALocalDecl, Interval> sigma) {
			return new Interval(null, null);
		}

		@Override @SuppressWarnings("unused")
		public Interval caseAIntConstExp(AIntConstExp node,
				EqualityMap<ALocalDecl, Interval> sigma) {
			Integer intconst = node.getValue();
			return new Interval(intconst, intconst);
		}

		public @Override
		Interval caseAVarExp(AVarExp varExp,
				EqualityMap<ALocalDecl, Interval> sigma) {
			Interval interval = null;
			if (varExp.getIdentifier().getDecl() instanceof ALocalDecl) {
				interval = sigma.get(varExp.getIdentifier().getDecl());
			}
			if (interval == null) {
				interval = new Interval(null, null);
			}
			return interval;
		}

		public @Override
		Interval caseABinopExp(ABinopExp binopExp,
				EqualityMap<ALocalDecl, Interval> sigma) {
			Interval left = binopExp.getLeft().apply(this, sigma);
			Interval right = binopExp.getRight().apply(this, sigma);
			Interval interval = new Interval(null, null);
			switch (binopExp.getBinop().kindPBinop()) {
			case PLUS: // [l,h] = [l1+l2,h1+h2]
				if (left.l != null && right.l != null) {
					interval.l = left.l + right.l;
				}
				if (left.h != null && right.h != null) {
					interval.h = left.h + right.h;
				}
				break;
			case MINUS: // [l,h] = [l1-h2,h1-l2]
				if (left.l != null && right.h != null) {
					interval.l = left.l - right.h;
				}
				if (left.h != null && right.l != null) {
					interval.h = left.h - right.l;
				}
				break;
			case TIMES: // [l,h] = [min(h1*l2,l1*h2),max(l1*l2,h1*h2)]
				Integer h1l2 = null;
				Integer l1h2 = null;
				Integer l1l2 = null;
				Integer h1h2 = null;
				if (left.h != null && right.l != null) {
					h1l2 = left.h * right.l;
				}
				if (left.l != null && right.h != null) {
					l1h2 = left.l * right.h;
				}
				if (left.l != null && right.l != null) {
					h1l2 = left.l * right.l;
				}
				if (left.h != null && right.h != null) {
					h1l2 = left.h * right.h;
				}
				if (h1l2 != null && l1h2 != null) {
					interval.l = Math.min(h1l2, l1h2);
				}
				if (l1l2 != null && h1h2 != null) {
					interval.h = Math.max(l1l2, h1h2);
				}
				break;
			case DIVIDE: // [l,h] = [min(l1/l2,l1/h2),max(h1/l2,h1/h2)]
				Interval below = null;
				Interval above = null;
				boolean zero = false;
				if (right.l == null && right.h == null) { // [l2,h2] =
															// [-\infty,\infty]
					below = new Interval(null, -1);
					above = new Interval(1, null);
					zero = true;
				} else if (right.l == null) { // [l2,h2] = [-\infty,a]
					if (right.h > 0) {
						below = new Interval(null, -1);
						above = new Interval(1, right.h);
						zero = true;
					} else if (right.h == 0) {
						below = new Interval(null, -1);
						zero = true;
					} else {
						below = new Interval(null, right.h);
					}
				} else if (right.h == null) { // [l2,h2] = [a,\infty]
					if (right.l < 0) {
						below = new Interval(right.l, -1);
						above = new Interval(1, null);
						zero = true;
					} else if (right.l == 0) {
						above = new Interval(1, null);
						zero = true;
					} else {
						above = new Interval(right.l, null);
					}
				} else {
					if (right.l > 0) {
						above = right;
					} else if (right.h < 0) {
						below = right;
					} else if (right.l == 0 && right.h == 0) {
						zero = true;
					} else if (right.l == 0) {
						above = new Interval(1, right.h);
						zero = true;
					} else if (right.h == 0) {
						below = new Interval(right.l, -1);
						zero = true;
					} else {
						above = new Interval(1, right.h);
						below = new Interval(right.l, -1);
						zero = true;
					}
				}
				if (below != null) {
					Integer min_a = null;
					Integer min_b = null;
					Integer max_a = null;
					Integer max_b = null;
					// min a:
					if (left.h != null && below.l != null) {
						min_a = left.h / below.l;
					}
					// min b:
					if (left.h != null) {
						min_b = left.h / below.h;
					}
					// max a:
					if (left.l != null && below.l != null) {
						max_a = left.l / below.l;
					}
					// max b:
					if (left.l != null) {
						max_b = left.l / below.h;
					}

					if (min_a != null && min_b != null) {
						below.l = Math.min(min_a, min_b);
					} else {
						below.l = null;
					}
					if (max_a != null && max_b != null) {
						below.h = Math.max(max_a, max_b);
					} else {
						below.h = null;
					}
				}
				if (above != null) {
					Integer min_a = null;
					Integer min_b = null;
					Integer max_a = null;
					Integer max_b = null;
					// min a:
					if (left.l != null) {
						min_a = left.l / above.l;
					}
					// min b:
					if (left.l != null && above.h != null) {
						min_b = left.l / above.h;
					}
					// max a:
					if (left.h != null) {
						max_a = left.h / above.l;
					}
					// max b:
					if (left.h != null && above.h != null) {
						max_b = left.h / above.h;
					}

					if (min_a != null && min_b != null) {
						above.l = Math.min(min_a, min_b);
					} else {
						above.l = null;
					}
					if (max_a != null && max_b != null) {
						above.h = Math.max(max_a, max_b);
					} else {
						above.h = null;
					}
				}
				if (below != null && above != null) {
					if (below.l != null && above.l != null) {
						interval.l = Math.min(below.l, above.l);
					}
					if (below.h != null && above.h != null) {
						interval.h = Math.max(below.h, above.h);
					}
				} else if (below != null) {
					interval.l = below.l;
					interval.h = below.h;
				} else if (above != null) {
					interval.l = above.l;
					interval.h = above.h;
				} else {
					interval.l = 0;
					interval.h = 0;
				}
				if (zero) {
					if (left.l == null || left.l < 0) {
						interval.l = null;
					}
					if (left.h == null || left.h > 0) {
						interval.h = null;
					}
				}
				break;
			case EQ:
				if (left.l != null && left.h != null && right.l != null
						&& right.h != null && left.l == left.h
						&& left.l == right.l && left.l == right.h) {
					interval.l = 1;
					interval.h = 1;
				} else if ((left.h != null && right.l != null && left.h < right.l)
						|| (left.l != null && right.h != null && left.l > right.h)) {
					interval.l = 0;
					interval.h = 0;
				} else {
					interval.l = 0;
					interval.h = 1;
				}
				break;
			case GT:
				if (left.l != null && right.h != null && left.l > right.h) {
					interval.l = 1;
					interval.h = 1;
				} else if (left.h != null && right.h != null
						&& left.h >= right.l) {
					interval.l = 0;
					interval.h = 0;
				} else {
					interval.l = 0;
					interval.h = 1;
				}
				break;
			}
			return interval;
		}
	}

	public Interval eval(EqualityMap<ALocalDecl, Interval> sigma, PExp exp) {
		if (exp instanceof AIntConstExp) {
			Integer intconst = ((AIntConstExp) exp).getValue();
			return new Interval(intconst, intconst);
		} else if (exp instanceof AVarExp) {
			AVarExp varExp = (AVarExp) exp;
			Interval interval = null;
			if (varExp.getIdentifier().getDecl() instanceof ALocalDecl) {
				interval = sigma.get(varExp.getIdentifier().getDecl());
			}
			if (interval == null) {
				interval = new Interval(null, null);
			}
			return interval;
		} else if (exp instanceof ABinopExp) {
			ABinopExp binopExp = (ABinopExp) exp;
			Interval left = eval(sigma, binopExp.getLeft());
			Interval right = eval(sigma, binopExp.getRight());
			Interval interval = new Interval(null, null);
			switch (binopExp.getBinop().kindPBinop()) {
			case PLUS: // [l,h] = [l1+l2,h1+h2]
				if (left.l != null && right.l != null) {
					interval.l = left.l + right.l;
				}
				if (left.h != null && right.h != null) {
					interval.h = left.h + right.h;
				}
				break;
			case MINUS: // [l,h] = [l1-h2,h1-l2]
				if (left.l != null && right.h != null) {
					interval.l = left.l - right.h;
				}
				if (left.h != null && right.l != null) {
					interval.h = left.h - right.l;
				}
				break;
			case TIMES: // [l,h] = [min(h1*l2,l1*h2),max(l1*l2,h1*h2)]
				Integer h1l2 = null;
				Integer l1h2 = null;
				Integer l1l2 = null;
				Integer h1h2 = null;
				if (left.h != null && right.l != null) {
					h1l2 = left.h * right.l;
				}
				if (left.l != null && right.h != null) {
					l1h2 = left.l * right.h;
				}
				if (left.l != null && right.l != null) {
					h1l2 = left.l * right.l;
				}
				if (left.h != null && right.h != null) {
					h1l2 = left.h * right.h;
				}
				if (h1l2 != null && l1h2 != null) {
					interval.l = Math.min(h1l2, l1h2);
				}
				if (l1l2 != null && h1h2 != null) {
					interval.h = Math.max(l1l2, h1h2);
				}
				break;
			case DIVIDE: // [l,h] = [min(l1/l2,l1/h2),max(h1/l2,h1/h2)]
				Interval below = null;
				Interval above = null;
				boolean zero = false;
				if (right.l == null && right.h == null) { // [l2,h2] =
															// [-\infty,\infty]
					below = new Interval(null, -1);
					above = new Interval(1, null);
					zero = true;
				} else if (right.l == null) { // [l2,h2] = [-\infty,a]
					if (right.h > 0) {
						below = new Interval(null, -1);
						above = new Interval(1, right.h);
						zero = true;
					} else if (right.h == 0) {
						below = new Interval(null, -1);
						zero = true;
					} else {
						below = new Interval(null, right.h);
					}
				} else if (right.h == null) { // [l2,h2] = [a,\infty]
					if (right.l < 0) {
						below = new Interval(right.l, -1);
						above = new Interval(1, null);
						zero = true;
					} else if (right.l == 0) {
						above = new Interval(1, null);
						zero = true;
					} else {
						above = new Interval(right.l, null);
					}
				} else {
					if (right.l > 0) {
						above = right;
					} else if (right.h < 0) {
						below = right;
					} else if (right.l == 0 && right.h == 0) {
						zero = true;
					} else if (right.l == 0) {
						above = new Interval(1, right.h);
						zero = true;
					} else if (right.h == 0) {
						below = new Interval(right.l, -1);
						zero = true;
					} else {
						above = new Interval(1, right.h);
						below = new Interval(right.l, -1);
						zero = true;
					}
				}
				if (below != null) {
					Integer min_a = null;
					Integer min_b = null;
					Integer max_a = null;
					Integer max_b = null;
					// min a:
					if (left.h != null && below.l != null) {
						min_a = left.h / below.l;
					}
					// min b:
					if (left.h != null) {
						min_b = left.h / below.h;
					}
					// max a:
					if (left.l != null && below.l != null) {
						max_a = left.l / below.l;
					}
					// max b:
					if (left.l != null) {
						max_b = left.l / below.h;
					}

					if (min_a != null && min_b != null) {
						below.l = Math.min(min_a, min_b);
					} else {
						below.l = null;
					}
					if (max_a != null && max_b != null) {
						below.h = Math.max(max_a, max_b);
					} else {
						below.h = null;
					}
				}
				if (above != null) {
					Integer min_a = null;
					Integer min_b = null;
					Integer max_a = null;
					Integer max_b = null;
					// min a:
					if (left.l != null) {
						min_a = left.l / above.l;
					}
					// min b:
					if (left.l != null && above.h != null) {
						min_b = left.l / above.h;
					}
					// max a:
					if (left.h != null) {
						max_a = left.h / above.l;
					}
					// max b:
					if (left.h != null && above.h != null) {
						max_b = left.h / above.h;
					}

					if (min_a != null && min_b != null) {
						above.l = Math.min(min_a, min_b);
					} else {
						above.l = null;
					}
					if (max_a != null && max_b != null) {
						above.h = Math.max(max_a, max_b);
					} else {
						above.h = null;
					}
				}
				if (below != null && above != null) {
					if (below.l != null && above.l != null) {
						interval.l = Math.min(below.l, above.l);
					}
					if (below.h != null && above.h != null) {
						interval.h = Math.max(below.h, above.h);
					}
				} else if (below != null) {
					interval.l = below.l;
					interval.h = below.h;
				} else if (above != null) {
					interval.l = above.l;
					interval.h = above.h;
				} else {
					interval.l = 0;
					interval.h = 0;
				}
				if (zero) {
					if (left.l == null || left.l < 0) {
						interval.l = null;
					}
					if (left.h == null || left.h > 0) {
						interval.h = null;
					}
				}
				break;
			case EQ:
				if (left.l != null && left.h != null && right.l != null
						&& right.h != null && left.l == left.h
						&& left.l == right.l && left.l == right.h) {
					interval.l = 1;
					interval.h = 1;
				} else if ((left.h != null && right.l != null && left.h < right.l)
						|| (left.l != null && right.h != null && left.l > right.h)) {
					interval.l = 0;
					interval.h = 0;
				} else {
					interval.l = 0;
					interval.h = 1;
				}
				break;
			case GT:
				if (left.l != null && right.h != null && left.l > right.h) {
					interval.l = 1;
					interval.h = 1;
				} else if (left.h != null && right.h != null
						&& left.h >= right.l) {
					interval.l = 0;
					interval.h = 0;
				} else {
					interval.l = 0;
					interval.h = 1;
				}
				break;
			}
			return interval;
		}
		return new Interval(null, null);
	}
}
