package tip.cfg.dfa;

import tip.display.*;
import tip.cfg.*;
import tip.node.*;
import tip.util.*;

/**
 * {@code LivenessAnalysis} implements the Liveness <em>backwards-may</em>
 * analysis. The implementation does not handle pointer expression. It uses
 * {@link UnionSetLattice} containing {@link ALocalDecl} elements as the
 * variables for the program points.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class LivenessAnalysis extends FlowAnalysis<EqualitySet<ALocalDecl>> {
	/**
	 * Creates a new {@code LivenessAnalysis} analysis.
	 */
	public LivenessAnalysis() {
		// set the dependency strategy to the backwards strategy as this
		// analysis is a backwards analysis
		super(new BackwardsStrategy(), "Liveness");
	}

	/**
	 * Creates the lattice to use for the control flow graph {@code cfg}. The
	 * lattice is a {@link UnionSetLattice} holdning {@link ALocalDecl} nodes
	 * using the {@link IdentityEqualitor} for determining equality between
	 * nodes.
	 * 
	 * @param cfg
	 *            the control flow graph for which the lattice is created
	 * @return the lattice to use for this analysis
	 */
	@Override
	protected Lattice<EqualitySet<ALocalDecl>> createLattice(ICFG cfg) {
		return new UnionSetLattice<ALocalDecl>();
	}

	/**
	 * Utility method for adding all referred {@link ALocalDecl} nodes in the
	 * subtree of {@code node} to {@code set}.
	 * 
	 * @param set
	 *            the set to which the {@link ALocalDecl} nodes are add
	 * @param node
	 *            the node whos subtree is searched
	 * @return the resulting set
	 */
	protected EqualitySet<ALocalDecl> addVarUses(EqualitySet<ALocalDecl> set,
			Node node) {
		set.addAll(LocalDeclCollector.collectLocalDecls(node) // retrieve
																// referred
																// local decls
		);
		return set;
	}

	/**
	 * Handle assignment statements.
	 * 
	 * @param node
	 *            the current assignment statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		// fetch the join value of this point
		EqualitySet<ALocalDecl> join = join(point);
		// remove the local decl of left hand side
		join.remove(node.getIdentifier().getDecl());
		// add referred local decls of the right hand side
		// and update the variable for this point
		return setVariable(point, addVarUses(join, node.getExp()));
	}

	/**
	 * Handle declaration statements.
	 * 
	 * @param node
	 *            the current declaration statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCDeclStmPoint(ADeclStm node, CDeclStmPoint point) {
		// fetch the join value of this point
		EqualitySet<ALocalDecl> join = join(point);
		// remove the declared local decls
		join.removeAll(node.getDescendants(ALocalDecl.class));
		// update the variable for this point
		return setVariable(point, join);
	}

	/**
	 * Handle if-then statements.
	 * 
	 * @param node
	 *            the current if-then statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCIfThenStmPoint(AIfThenStm node, CIfThenStmPoint point) {
		// fetch the join value of this point,
		// add referred local decls of the condition,
		// and update the variable for this point
		return setVariable(point, addVarUses(join(point), node.getCondition()));
	}

	/**
	 * Handle if-then-else statements.
	 * 
	 * @param node
	 *            the current if-then-else statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCIfThenElseStmPoint(AIfThenElseStm node,
			CIfThenElseStmPoint point) {
		// fetch the join value of this point,
		// add referred local decls of the condition,
		// and update the variable for this point
		return setVariable(point, addVarUses(join(point), node.getCondition()));
	}

	/**
	 * Handle while statements.
	 * 
	 * @param node
	 *            the current while statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCWhileStmPoint(AWhileStm node, CWhileStmPoint point) {
		// fetch the join value of this point,
		// add referred local decls of the condition,
		// and update the variable for this point
		return setVariable(point, addVarUses(join(point), node.getCondition()));
	}

	/**
	 * Handle output statements.
	 * 
	 * @param node
	 *            the current output statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCOutputStmPoint(AOutputStm node, COutputStmPoint point) {
		// fetch the join value of this point,
		// add referred local decls of the expression,
		// and update the variable for this point
		return setVariable(point, addVarUses(join(point), node.getExp()));
	}

	/**
	 * Handle return statements.
	 * 
	 * @param node
	 *            the current return statement
	 * @param point
	 *            the current program point
	 * @return {@code true} if the variable for {@code point} was changed
	 */
	public @Override
	boolean caseCReturnStmPoint(AReturnStm node, CReturnStmPoint point) {
		// fetch the join value of this point,
		// add referred local decls of the expression,
		// and update the variable for this point
		return setVariable(point, addVarUses(join(point), node.getExp()));
	}

	/**
	 * Returns a textual representation of {@code variable}. {@link NodeText} is
	 * passed to {@link EqualitySet#toString(Stringifier)} in order to let
	 * {@link EqualitySet} print each element as defined by {@link NodeText}.
	 * 
	 * @param variable
	 *            the variable for which a textual representation is returned
	 * @return the textual representation of {@code variable}
	 */
	public String getVariableString(EqualitySet<ALocalDecl> variable) {
		return variable.toString(new NodeText<ALocalDecl>());
	}
}
