package tip.cfg.dfa;

import tip.*;
import tip.cfg.*;
import tip.node.*;
import tip.util.*;
import tip.display.*;

public class NullPointerAnalysis extends
		FlowAnalysis<EqualityMap<ALocalDecl, NullPointerAnalysis.Null>> {
	public enum Null {
		BOT("_"), NN("NN"), IN("IN"), TOP("?");

		private String text;

		Null(String text) {
			this.text = text;
		}

		@Override
		public String toString() {
			return text;
		}
	}

	public static class NullLattice implements Lattice<Null> {
		public Null getSmallestElement() {
			return Null.BOT;
		}

		public Null getLeastUpperBound(Null a, Null b) {
			if (a == null && b == null) {
				return Null.BOT;
			}
			if (a == null) {
				return b;
			}
			if (b == null) {
				return a;
			}
			switch (a) {
			case BOT:
				return b;
			case NN:
				switch (b) {
				case IN:
					return Null.IN;
				case TOP:
					return Null.TOP;
				default:
					return Null.NN;
				}
			case IN:
				switch (b) {
				case TOP:
					return Null.TOP;
				default:
					return Null.IN;
				}
			case TOP:
				return a;
			}
			return Null.BOT;
		}

		public boolean equals(Null a, Null b) {
			return a == b;
		}

		public int hashCode(Null a) {
			return a == null ? 0 : a.hashCode();
		}

		@Override
		public boolean equals(Object o) {
			return o == null ? false : getClass().equals(o.getClass());
		}

	}

	private NullLattice nullLattice = new NullLattice();

	public NullPointerAnalysis() {
		super(new ForwardsStrategy(), "Null Pointer");
	}

	@Override
	public MapLattice<ALocalDecl, Null> createLattice(ICFG cfg) {
		return new MapLattice<ALocalDecl, Null>(
				LocalDeclCollector.collectLocalDecls(cfg), new NullLattice());
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualityMap<ALocalDecl, Null> join = join(point);
		PDecl decl1 = node.getIdentifier().getDecl();
		if (decl1 instanceof ALocalDecl) {
			ALocalDecl id1 = (ALocalDecl) decl1;
			PExp exp = node.getExp();
			if (exp instanceof AMallocExp) {
				join.put(id1, Null.NN);
			} else if (exp instanceof APointerExp) {
				join.put(id1, Null.NN);
			} else if (exp instanceof AVarExp) {
				PDecl decl2 = ((AVarExp) exp).getIdentifier().getDecl();
				Null b = Null.NN; // if decl2 is a AFuncDecl
				if (decl2 instanceof ALocalDecl) {
					b = join.get(decl2);
				}
				if (b == null) {
					b = Null.BOT;
				}
				join.put(id1, b);
			} else if (exp instanceof ADerefExp) {
				ALocalDecl id2 = (ALocalDecl) ((AVarExp) ((ADerefExp) exp).getExp()).getIdentifier().getDecl();
				Null a = join.get(id2);
				if (a == null) {
					a = Null.BOT;
				}
				for (Target d : Main.pointsToMap.pt(point, id2)) {
					if (d.getTarget() instanceof ALocalDecl) {
						ALocalDecl p = (ALocalDecl) d.getTarget();
						Null b = join.get(p);
						if (b != null) {
							a = nullLattice.getLeastUpperBound(a, b);
						}
					}
				}
				join.put(id1, a);
			} else if (exp instanceof ANullExp) {
				join.put(id1, Null.IN);
			}
		}

		return setVariable(point, join);
	}

	public @Override
	boolean caseCAssignPtrStmPoint(AAssignPtrStm node, CAssignPtrStmPoint point) {
		EqualityMap<ALocalDecl, Null> join = join(point);
		// assumes normalization of pointer expressions
		PDecl decl1 = ((AVarExp) node.getLeft()).getIdentifier().getDecl();
		PDecl decl2 = ((AVarExp) node.getRight()).getIdentifier().getDecl();
		if (decl1 instanceof ALocalDecl) {
			ALocalDecl id1 = (ALocalDecl) decl1;
			Null a = Null.NN; // if decl2 is a AFuncDecl
			if (decl2 instanceof ALocalDecl) {
				a = join.get(decl2);
			}
			if (a == null) {
				a = Null.BOT;
			}
			for (Target d : Main.pointsToMap.pt(point, id1)) {
				if (d.getTarget() instanceof ALocalDecl) {
					ALocalDecl p = (ALocalDecl) d.getTarget();
					Null b = join.get(p);
					if (b != null) {
						a = nullLattice.getLeastUpperBound(a, b);
					}
					join.put(p, a);
				}
			}
		}
		return setVariable(point, join);
	}

	public @Override
	boolean caseCDeclStmPoint(ADeclStm node, CDeclStmPoint point) {
		EqualityMap<ALocalDecl, Null> join = join(point);
		for (PDecl decl : node.getDecl()) {
			if (decl instanceof ALocalDecl) {
				join.put((ALocalDecl) decl, Null.TOP);
			}
		}
		return setVariable(point, join);
	}

	@Override @SuppressWarnings("unused")
	public boolean caseEntryPoint(Node node, EntryPoint point) {
		EqualityMap<ALocalDecl, Null> join = getLattice().getSmallestElement();
		for (ALocalDecl decl : join.keySet()) {
			join.put(decl, Null.TOP);
		}
		return setVariable(point, join);
	}

	public String getVariableString(EqualityMap<ALocalDecl, Null> variable) {
		return variable.toString(new NodeText<ALocalDecl>(),
				new DefaultStringifier<Null>());
	}
	/*
	 * public String variableString(Object variable) { StringBuffer sb = new
	 * StringBuffer(); sb.append("["); Iterator<Map.Entry<ALocalDecl,Null>> it =
	 * ((Map<ALocalDecl,Null>)variable).entrySet().iterator(); while
	 * (it.hasNext()) { Map.Entry<ALocalDecl,Null> entry = it.next();
	 * sb.append(entry.getKey() + " = " + entry.getValue()); if (it.hasNext()) {
	 * sb.append(","); } } sb.append("]"); return sb.toString(); }
	 */
}
