package tip.cfg.dfa;

import java.util.*;
import tip.node.*;
import tip.util.EqualitySet;
import tip.util.Stringifier;
import tip.cfg.*;
import tip.display.*;
import dk.brics.dot.*;
import dk.brics.dot.html.*;
import dk.brics.dot.view.*;

public class ReachingDefinitionsAnalysis extends
		FlowAnalysis<EqualitySet<CAssignVarStmPoint>> {
	public ReachingDefinitionsAnalysis() {
		super(new ForwardsStrategy(), "Reaching Definitions");
	}

	@Override @SuppressWarnings("unused")
	protected Lattice<EqualitySet<CAssignVarStmPoint>> createLattice(ICFG cfg) {
		return new UnionSetLattice<CAssignVarStmPoint>();
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualitySet<CAssignVarStmPoint> assignments = join(point);
		EqualitySet<CAssignVarStmPoint> prevDefs = getLattice().getSmallestElement();

		for (CAssignVarStmPoint n : assignments) {
			if ((n.getNode()).getIdentifier().getDecl() == node.getIdentifier().getDecl()) {
				prevDefs.add(n);
			}
		}

		// Remove all statements where 'id' is assigned to
		assignments.removeAll(prevDefs);
		assignments.add(point);
		return setVariable(point, assignments);
	}

	public class AssignVarStmPointStringifier implements
			Stringifier<CAssignVarStmPoint> {
		public static final int MAX_LENGTH = 10;

		public String toString(CAssignVarStmPoint point) {
			String result = point.getNodeText(new NodeText<Node>());
			if (result.length() > MAX_LENGTH) {
				return result.substring(0, MAX_LENGTH);
			}
			return result;
		}
	}

	public String getVariableString(EqualitySet<CAssignVarStmPoint> variable) {
		return variable.toString(new AssignVarStmPointStringifier());
	}

	public void showDefUseGraph(ICFGCollection cfgs) {
		DotViewer dv = new DotViewer();

		NodeText<Node> nodeText = new NodeText<Node>();

		FontAttributes fontStyle = new FontAttributes();
		fontStyle.setFontName("Courier");
		fontStyle.setFontSize(10.0);
		NodeAttributes nodeStyle = new NodeAttributes();
		nodeStyle.setFontStyle(fontStyle);
		nodeStyle.setShape(DotShape.plaintext);

		EdgeAttributes visibleEdge = new EdgeAttributes();
		visibleEdge.setStyle(EdgeStyle.solid);
		visibleEdge.setDir(Dir.forward);

		EdgeAttributes invisibleEdge = new EdgeAttributes();
		invisibleEdge.setStyle(EdgeStyle.dotted);
		invisibleEdge.setDir(Dir.none);

		for (ICFG cfg : cfgs.getCFGs()) {
			DotGraph graph = new DotGraph(cfg.getName(), true);
			graph.createNodeStatement(nodeStyle);

			Map<IProgramPoint, dk.brics.dot.DotNode> pointToNodeMap = new HashMap<IProgramPoint, dk.brics.dot.DotNode>();
			for (IProgramPoint point : cfg.getProgramPoints()) {
				dk.brics.dot.DotNode node = graph.createNode();
				pointToNodeMap.put(point, node);
				HTMLTableLabel label = new HTMLTableLabel();
				label.setBorder(0);
				label.setCellBorder(1);
				label.setCellSpacing(0);
				label.setCellPadding(1);
				HTMLRow row = label.createRow();
				HTMLLabelCell cell = new HTMLLabelCell(
						point.getNodeText(nodeText));
				cell.setPort("point");
				row.addCell(cell);
				node.setLabel(label);
			}

			graph.createEdgeStatement(invisibleEdge);

			for (IProgramPoint point : cfg.getProgramPoints()) {
				dk.brics.dot.DotNode from = pointToNodeMap.get(point);
				for (IProgramPoint successor : point.getSuccessors()) {
					dk.brics.dot.DotNode to = pointToNodeMap.get(successor);
					graph.createEdge(from, "point", to, "point");
				}
			}

			graph.createEdgeStatement(visibleEdge);

			for (IProgramPoint point : cfg.getProgramPoints()) {
				if (point instanceof EntryPoint || point instanceof ExitPoint) {
					continue;
				}

				dk.brics.dot.DotNode to = pointToNodeMap.get(point);
				Node stmNode = point.getNode();
				Node expNode = stmNode;
				if (stmNode instanceof AIfThenStm) {
					expNode = ((AIfThenStm) stmNode).getCondition();
				} else if (stmNode instanceof AIfThenElseStm) {
					expNode = ((AIfThenElseStm) stmNode).getCondition();
				} else if (stmNode instanceof AWhileStm) {
					expNode = ((AWhileStm) stmNode).getCondition();
				} else if (stmNode instanceof AAssignVarStm) {
					expNode = ((AAssignVarStm) stmNode).getExp();
				}
				Collection<TIdentifier> vars = expNode.getDescendants(TIdentifier.class);
				for (CAssignVarStmPoint assignStmPoint : getVariable(point)) {
					boolean found = false;
					for (TIdentifier id : vars) {
						if (id.getDecl() == assignStmPoint.getNode().getIdentifier().getDecl()) {
							found = true;
							break;
						}
					}
					if (found) {
						dk.brics.dot.DotNode from = pointToNodeMap.get(assignStmPoint);
						if (from != to) {
							graph.createEdge(from, "point", to, "point");
						}
					}
				}
			}
			dv.addGraph(graph);
		}
		dv.showViewer(true);
	}
}
