package tip.cfg.dfa;

import tip.cfg.*;
import tip.node.*;
import tip.util.*;
import java.util.*;

/**
 * {@code ShapeAnalysis} implements the Shape analysis for computing points-to
 * maps.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class ShapeAnalysis extends
		FlowAnalysis<EqualitySet<ShapeAnalysis.TargetPair>> implements
		PointsToMap {
	/**
	 * Creates a new {@code ShapeAnalysis}.
	 */
	public ShapeAnalysis() {
		super(new ForwardsStrategy(), "Shape");
	}

	protected static class TargetPair {
		private Target source;

		private Target destination;

		public TargetPair(Target source, Target destination) {
			if (source == null || destination == null) {
				throw new IllegalArgumentException(
						"null is not allowed as target");
			}
			this.source = source;
			this.destination = destination;
		}

		public Target getSource() {
			return source;
		}

		public Target getDestination() {
			return destination;
		}

		@Override
		public boolean equals(Object o) {
			if (!(o instanceof TargetPair)) {
				return false;
			}
			TargetPair targetPair = (TargetPair) o;
			return source.equals(targetPair.source)
					&& destination.equals(targetPair.destination);
		}

		@Override
		public int hashCode() {
			return source.hashCode() ^ destination.hashCode();
		}

		@Override
		public String toString() {
			return source + "->" + destination;
		}
	}

	@Override @SuppressWarnings("unused")
	protected Lattice<EqualitySet<TargetPair>> createLattice(ICFG cfg) {
		return new UnionSetLattice<TargetPair>();
	}

	public String getVariableString(EqualitySet<TargetPair> variable) {
		return variable.toString();
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualitySet<TargetPair> join = join(point);
		PDecl decl1 = node.getIdentifier().getDecl();
		if (decl1 instanceof ALocalDecl) {
			ALocalDecl id1 = (ALocalDecl) decl1;
			PExp exp = (node).getExp();
			if (exp instanceof AMallocExp) {
				join = downarrow(join, id1);
				join.add(new TargetPair(new DeclarationTarget(id1),
						new MallocTarget((AMallocExp) exp)));
			} else if (exp instanceof APointerExp) {
				join = downarrow(join, id1);
				join.add(new TargetPair(new DeclarationTarget(id1),
						new DeclarationTarget(
								((APointerExp) exp).getIdentifier().getDecl())));
			} else if (exp instanceof AVarExp) {
				join = downarrow(join, id1);
				Target x = new DeclarationTarget(id1);
				Target y = new DeclarationTarget(
						((AVarExp) exp).getIdentifier().getDecl());
				EqualitySet<TargetPair> newEdges = new EqualitySet<TargetPair>();
				for (TargetPair pair : join) {
					if (pair.getSource().equals(y)) {
						newEdges.add(new TargetPair(x, pair.getDestination()));
					}
				}
				join.addAll(newEdges);
			} else if (exp instanceof ADerefExp) {
				join = downarrow(join, id1);
				Target x = new DeclarationTarget(id1);
				Target y = new DeclarationTarget(
						((AVarExp) ((ADerefExp) exp).getExp()).getIdentifier().getDecl());
				EqualitySet<TargetPair> newEdges = new EqualitySet<TargetPair>();
				for (TargetPair pair_ys : join) {
					if (pair_ys.getSource().equals(y)) {
						for (TargetPair pair_st : join) {
							if (pair_st.getSource().equals(
									pair_ys.getDestination())) {
								newEdges.add(new TargetPair(x,
										pair_st.getDestination()));
							}
						}
					}
				}
				join.addAll(newEdges);
			} else if (exp instanceof ANullExp) {
				join = downarrow(join, id1);
			}
		}
		return setVariable(point, join);
	}

	public @Override
	boolean caseCAssignPtrStmPoint(AAssignPtrStm node, CAssignPtrStmPoint point) {
		EqualitySet<TargetPair> join = join(point);
		Target x = new DeclarationTarget(
				((AVarExp) node.getLeft()).getIdentifier().getDecl());
		Target y = new DeclarationTarget(
				((AVarExp) node.getRight()).getIdentifier().getDecl());
		EqualitySet<TargetPair> newEdges = new EqualitySet<TargetPair>();
		for (TargetPair pair_xs : join) {
			if (pair_xs.getSource().equals(x)) {
				EqualitySet<TargetPair> downarrow_s = downarrow(join,
						pair_xs.getDestination());
				for (TargetPair pair_yt : join) {
					if (pair_yt.getSource().equals(y)) {
						downarrow_s.add(new TargetPair(
								pair_xs.getDestination(),
								pair_yt.getDestination()));
						newEdges.addAll(downarrow_s);
					}
				}
			}
		}
		if (newEdges.size() != 0) {
			join = newEdges;
		}
		return setVariable(point, join);
	}

	protected EqualitySet<TargetPair> downarrow(EqualitySet<TargetPair> join,
			ALocalDecl decl) {
		return downarrow(join, new DeclarationTarget(decl));
	}

	protected EqualitySet<TargetPair> downarrow(EqualitySet<TargetPair> join,
			Target t) {
		EqualitySet<TargetPair> result = new EqualitySet<TargetPair>();
		for (TargetPair pair : join) {
			if (!pair.getSource().equals(t)) {
				result.add(pair);
			}
		}
		return result;
	}

	public Set<Target> pt(IProgramPoint point, ALocalDecl decl) {
		EqualitySet<Target> targetSet = new EqualitySet<Target>();
		EqualitySet<TargetPair> variable = getVariable(point);
		if (variable != null) {
			for (TargetPair pair : variable) {
				Target source = pair.getSource();
				if (source instanceof DeclarationTarget) {
					DeclarationTarget declSource = (DeclarationTarget) source;
					if (declSource.getTarget() == decl) {
						targetSet.add(pair.getDestination());
					}
				}
			}
		}
		return targetSet;
	}

	public boolean isFlowSensitive() {
		return true;
	}

	private Map<IProgramPoint, Map<Target, Set<Target>>> ptMap = new HashMap<IProgramPoint, Map<Target, Set<Target>>>();

	public Map<Target, Set<Target>> pt(IProgramPoint point) {
		Map<Target, Set<Target>> map = ptMap.get(point);
		if (map == null) {
			map = new HashMap<Target, Set<Target>>();
			ptMap.put(point, map);
			EqualitySet<TargetPair> variable = getVariable(point);
			for (TargetPair pair : variable) {
				Set<Target> targets = map.get(pair.getSource());
				if (targets == null) {
					targets = new HashSet<Target>();
					map.put(pair.getSource(), targets);
				}
				targets.add(pair.getDestination());
			}
		}
		return map;
	}
}
