package tip.cfg.dfa;

import tip.*;
import tip.cfg.*;
import tip.node.*;
import tip.util.*;
import tip.analysis.*;
import tip.display.*;

public class SignAnalysis extends
		FlowAnalysis<EqualityMap<ALocalDecl, SignAnalysis.Sign>> {
	public SignAnalysis() {
		super(new ForwardsStrategy(), "Sign");
	}

	public enum Sign {
		BOT("_"), ZERO("0"), MINUS("-"), PLUS("+"), UNKNOWN("?");

		private String name;

		private Sign(String name) {
			this.name = name;
		}

		@Override
		public String toString() {
			return name;
		}
	}

	public class SignLattice implements Lattice<Sign> {

		public Sign getSmallestElement() {
			return Sign.BOT;
		}

		public Sign getLeastUpperBound(Sign a, Sign b) {
			if (a == null && b == null) {
				return Sign.BOT;
			}
			if (a == null) {
				return b;
			}
			if (b == null) {
				return a;
			}
			switch (a) {
			case BOT:
				return b;
			case UNKNOWN:
				return Sign.UNKNOWN;
			default:
				return a == b ? a : Sign.UNKNOWN;
			}
		}

		public boolean equals(Sign a, Sign b) {
			return a == null ? b == null : a == b;
		}

		public int hashCode(Sign a) {
			return a == null ? 0 : a.hashCode();
		}

		@Override
		public boolean equals(Object o) {
			return o == null ? false : getClass().equals(o.getClass());
		}

		@Override
		public int hashCode() {
			return getClass().hashCode();
		}
	}

	public static final Sign[][] PLUS_op = {
			{ Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.BOT, Sign.ZERO, Sign.MINUS, Sign.PLUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.MINUS, Sign.MINUS, Sign.UNKNOWN, Sign.UNKNOWN },
			{ Sign.BOT, Sign.PLUS, Sign.UNKNOWN, Sign.PLUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static final Sign[][] MINUS_op = {
			{ Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.BOT, Sign.ZERO, Sign.PLUS, Sign.MINUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.MINUS, Sign.UNKNOWN, Sign.MINUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.PLUS, Sign.PLUS, Sign.UNKNOWN, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static final Sign[][] TIMES_op = {
			{ Sign.BOT, Sign.ZERO, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.ZERO, Sign.ZERO, Sign.ZERO, Sign.ZERO, Sign.ZERO },
			{ Sign.BOT, Sign.ZERO, Sign.PLUS, Sign.MINUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.ZERO, Sign.MINUS, Sign.PLUS, Sign.UNKNOWN },
			{ Sign.BOT, Sign.ZERO, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static final Sign[][] DIVIDE_op = {
			{ Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.BOT, Sign.UNKNOWN, Sign.ZERO, Sign.ZERO, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static final Sign[][] GT_op = {
			{ Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.BOT, Sign.ZERO, Sign.PLUS, Sign.ZERO, Sign.UNKNOWN },
			{ Sign.BOT, Sign.ZERO, Sign.UNKNOWN, Sign.ZERO, Sign.UNKNOWN },
			{ Sign.BOT, Sign.PLUS, Sign.PLUS, Sign.UNKNOWN, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static final Sign[][] EQ_op = {
			{ Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT, Sign.BOT },
			{ Sign.BOT, Sign.PLUS, Sign.ZERO, Sign.ZERO, Sign.UNKNOWN },
			{ Sign.BOT, Sign.ZERO, Sign.UNKNOWN, Sign.ZERO, Sign.UNKNOWN },
			{ Sign.BOT, Sign.ZERO, Sign.ZERO, Sign.UNKNOWN, Sign.UNKNOWN },
			{ Sign.BOT, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN, Sign.UNKNOWN } };

	public static class Evaluator extends
			QuestionAnswerAdapter<EqualityMap<ALocalDecl, Sign>, Sign> {
		@Override @SuppressWarnings("unused")
		public Sign defaultNode(Node node, EqualityMap<ALocalDecl, Sign> map) {
			return Sign.UNKNOWN;
		}

		public @Override
		Sign caseAVarExp(AVarExp node, EqualityMap<ALocalDecl, Sign> map) {
			return map.get(node.getIdentifier().getDecl());
		}

		@Override @SuppressWarnings("unused")
		public Sign caseAIntConstExp(AIntConstExp node,
				EqualityMap<ALocalDecl, Sign> map) {
			int value = node.getValue();
			if (value == 0) {
				return Sign.ZERO;
			} else if (value > 0) {
				return Sign.PLUS;
			} else {
				return Sign.MINUS;
			}
		}

		public @Override
		Sign caseABinopExp(ABinopExp node, EqualityMap<ALocalDecl, Sign> map) {
			Sign a = node.getLeft().apply(this, map);
			Sign b = node.getRight().apply(this, map);
			switch (node.getBinop().kindPBinop()) {
			case PLUS:
				return PLUS_op[a.ordinal()][b.ordinal()];
			case MINUS:
				return MINUS_op[a.ordinal()][b.ordinal()];
			case TIMES:
				return TIMES_op[a.ordinal()][b.ordinal()];
			case DIVIDE:
				return DIVIDE_op[a.ordinal()][b.ordinal()];
			case EQ:
				return EQ_op[a.ordinal()][b.ordinal()];
			case GT:
				return GT_op[a.ordinal()][b.ordinal()];
			}
			throw new InternalCompilerError("Unknown binop kind: "
					+ node.getBinop().kindPBinop());
		}
	}

	private Evaluator evaluator = new Evaluator();

	@Override
	protected Lattice<EqualityMap<ALocalDecl, Sign>> createLattice(ICFG cfg) {
		return new MapLattice<ALocalDecl, Sign>(
				LocalDeclCollector.collectLocalDecls(cfg), new SignLattice());
	}

	public String getVariableString(EqualityMap<ALocalDecl, Sign> variable) {
		return variable.toString(new NodeText<ALocalDecl>(),
				new DefaultStringifier<Sign>());
	}

	public @Override
	boolean caseCAssignVarStmPoint(AAssignVarStm node, CAssignVarStmPoint point) {
		EqualityMap<ALocalDecl, Sign> join = join(point);
		join.put((ALocalDecl) node.getIdentifier().getDecl(),
				node.getExp().apply(evaluator, join));
		return setVariable(point, join);
	}

	public @Override
	boolean caseCDeclStmPoint(ADeclStm node, CDeclStmPoint point) {
		EqualityMap<ALocalDecl, Sign> join = join(point);
		for (PDecl decl : node.getDecl()) {
			join.put((ALocalDecl) decl, Sign.UNKNOWN);
		}
		return setVariable(point, join);
	}

	public @Override @SuppressWarnings("unused")
	boolean caseEntryPoint(Node node, EntryPoint point) {
		EqualityMap<ALocalDecl, Sign> join = getLattice().getSmallestElement();
		for (ALocalDecl decl : join.keySet()) {
			join.put(decl, Sign.UNKNOWN);
		}
		return setVariable(point, join);
	}
}
