/*
 * NOTES
 * AST Walker which creates graphical tree
 */

package tip.display;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.awt.Dialog.ModalityType;
import javax.swing.*;
import javax.swing.tree.*;
import tip.analysis.*;
import tip.node.*;

public class ASTDisplay extends DepthFirstAdapter {
	/**
	 * The dialog title
	 */
	private String title;
	
	/**
	 * The parents stack used to construct the tree data during traversal.
	 */
    private Stack<DefaultMutableTreeNode> parents 
    	= new Stack<DefaultMutableTreeNode>();
 
    /**
     * Creates a new ASTDisplay with the given dialog title.
     * @param title
     * 		the displayed dialog title
     */
    public ASTDisplay(String title) {
    	this.title = title;
    }
    
    /**
     * Adds a field to the node on the top of parents stack.
     * @param label
     * 		the field label
     * @param value
     * 		the displayed value (may be null) 
     */
    protected void addField(String label, Object value) {
    	if (!parents.isEmpty()) {
    		String valueText = value != null ? value.toString() : "null";
    		parents.peek().add(new DefaultMutableTreeNode(
    				label + "=" + valueText));
    	}
    }
    
    /**
     * Adds a map to the node on the top of parents stack.
     * @param label
     * 		the map label
     * @param map
     * 		the displayed map (must be nonnull) 
     */
    protected void addMap(String label, Map<?,?> map) {
    	if (!parents.isEmpty()) {
    		DefaultMutableTreeNode mapNode = new DefaultMutableTreeNode(label);
    		for (Map.Entry<?,?> entry : map.entrySet()) {
        		String keyText = entry.getKey() != null ? entry.getKey().toString() : "null";
        		String valueText = entry.getValue() != null ? entry.getValue().toString() : "null";
        		mapNode.add(new DefaultMutableTreeNode(keyText + "=" + valueText));
    		}
    		parents.peek().add(mapNode);
    	}
    }
    
    protected void pushTreeNode(String label) {
		DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(label);
		parents.push(thisNode);
    }
    
    protected void popTreeNode() {
		DefaultMutableTreeNode thisNode = parents.pop();
		parents.peek().add(thisNode);
    }
    
    protected DefaultMutableTreeNode peekTreeNode() {
    	return parents.peek();
    }
 
    @Override @SuppressWarnings("unused")
	public void outAProgram(AProgram node) {
    	createAndShowDisplay();
    }

	/**
	 * As we come across nonterminals, push them onto the stack
	 * @param node
	 * 		the handled node
	 */
    @Override
	public void defaultIn(Node node) {
    	String label = node.getClass().getSimpleName();
    	pushTreeNode(label);
    }

	/**
	 * As we leave a non terminal, it's parent is the node before it
	 * on the stack, so we pop it off and add it to that node
	 */
    @Override @SuppressWarnings("unused")
	public void defaultOut(Node node) {
    	popTreeNode();
    }

	/**
	 *	Terminals - our parent is always on the top of the stack, so we
	 *	add ourselves to it
	 *	@param token
	 *		the added token
	 */
    @Override
	public void defaultToken(Token token) {
    	String label = token.getClass().getSimpleName()
    		+ "(\"" + token.getText() + "\")";
		DefaultMutableTreeNode thisNode 
			= new DefaultMutableTreeNode(label);
		parents.peek().add(thisNode);
    }
    
    @Override @SuppressWarnings("unused")
    public final void caseStart(Start node) {
    	// avoid traversal since Start nodes are not part of the program
    }    
    @Override @SuppressWarnings("unused")
	public final void caseEOF(EOF node) {
    	// avoid traversal since EOF tokens are not part of the program
    }
    
    /**
     * Creates and shows a display with the collected tree data.
     */
    private final void createAndShowDisplay() {
    	// Create a modal dialog with the given title
    	// Use DOCUMENT_MODAL to allow multiple modal dialogs
        JDialog display = new JDialog(null, title, ModalityType.DOCUMENT_MODAL);
        // Create a tree view of the collected tree data
		JTree tree = new JTree(parents.pop());
		JScrollPane pane = new JScrollPane(tree);
		// Expand all nodes in the tree view
		expandAll(tree, true);

		display.setSize(400, 500);
		display.getContentPane().add(pane);
		// show the display
		showDisplay(display);
    }
    
    ////////////////////////////////////////////////////////////////////////////
    //	static implementation
    ////////////////////////////////////////////////////////////////////////////
	
    /**
     * Enum used for designating for which phases the AST should be displayed.
     */
	public enum Phase {
		PARSER("-Parser"), 
		NORMALIZATION("-Normalization"),
		ENVIRONMENTS("-Environments"),
		NAME_LINKING("-NameLinking"),
		TYPE_CHECKING("-TypeChecking"),
		RESOURCES("-Resources"),
		CODE_GENERATION("-CodeGeneration"),
		;
		
		public final String commandline_option;
		Phase(String option) {
			commandline_option = option;
		}
	}
	
    /**
     * The set of phases for which the AST is displayed.
     */
	private static EnumSet<Phase> displayPhases = EnumSet.noneOf(Phase.class);
    
    /**
     * Adds the given phase to the set of phases for which the AST is displayed.
     * @param phase
     * 		the added phase
     */
    public static void addDisplayPhase(Phase phase) {
    	displayPhases.add(phase);
    }
    
    /**
     * Clears the set of phases for which the AST is displayed.
     */
    public static void clearDisplayPhases() {
    	displayPhases.clear();
    }
    
    /**
     * Shows the ASTDisplay for the given phase. The AST is only display if
     * an implementation has been provided for the phase and its display has 
     * been enabled through its respective commandline option. 
     * @param program
     * 		the program AST
     * @param phase
     * 		the phase for which to display the AST
     */
    public static void showPhaseDisplay(AProgram program, Phase phase) {
   		if (displayPhases.contains(phase)) {
    		switch (phase) {
    		case PARSER:
    			// display the AST after Parser
				program.apply(new ParserDisplay());
				break;
    		case NORMALIZATION:
				// display the AST after Normalization
				program.apply(new NormalizationDisplay());
				break;
    		case ENVIRONMENTS:
				program.apply(new EnvironmentsDisplay());
				break;
    		case NAME_LINKING:
				program.apply(new NameLinkingDisplay());
    			break;
    		case TYPE_CHECKING:
				program.apply(new TypeCheckingDisplay());
    			break;
    		case RESOURCES:
				program.apply(new ResourcesDisplay());
    			break;
    		case CODE_GENERATION:
				program.apply(new CodeGenerationDisplay());
    			break;
    		}
    	}
    }

    /**
     * Wait for all open displays to be closed before continuing.
     */
    public static void awaitDisplays() {
    	try {
    		executorService.shutdown();
        	// Wait for displays to be closed but at most 
    		// for the given time period (just change if needed)
    		executorService.awaitTermination(5, TimeUnit.MINUTES);
    	} catch (InterruptedException e) {
    		// Ignore (and terminate compilation)
    	}
    }


    /**
     * ExecutorService used to synchronize termination.
     */
    private static final ExecutorService executorService = Executors.newCachedThreadPool();
 
    /**
     * Shows the display, delaying compiler termination until 
     * the display is closed.
     * @param display
     * 		the shown display
     */
    protected static void showDisplay(final JDialog display) {
    	executorService.execute(new Runnable() {
    		public void run() {
    			// show dialog (and wait if dialog is modal)
   				display.setVisible(true);
   			}
   		});
	}
    
    /**
     * If expand is {@code true}, expands all nodes in the tree. Otherwise, 
     * collapses all nodes in the tree.
     * @param tree
     * 		the expanded/collapsed tree
     * @param expand
     * 		expand/collapse switch
     */
    // 
    public static void expandAll(JTree tree, boolean expand) {
        TreeNode root = (TreeNode)tree.getModel().getRoot();
    
        // Traverse tree from root
        expandAll(tree, new TreePath(root), expand);
    }

    private static void expandAll(JTree tree, TreePath parent, boolean expand) {
        // Traverse children
        TreeNode node = (TreeNode)parent.getLastPathComponent();
        if (node.getChildCount() >= 0) {
            for (Enumeration<?> e=node.children(); e.hasMoreElements(); ) {
                TreeNode n = (TreeNode)e.nextElement();
                TreePath path = parent.pathByAddingChild(n);
                expandAll(tree, path, expand);
            }
        }
    
        // Expansion or collapse must be done bottom-up
        if (expand) {
            tree.expandPath(parent);
        } else {
            tree.collapsePath(parent);
        }
    }
}
