package tip.display;

import java.util.*;
import tip.node.*;
import tip.util.Stringifier;
import tip.analysis.*;

/**
 * {@code NodeText} implements the {@link Answer} and {@link Stringifier}
 * interfaces to support a custom textual representation of nodes.
 * 
 * @author Johnni Winther
 * @param <T>
 *            type of node for which the {@code NodeText} applies
 */
public class NodeText<T extends Node> extends AnswerAdapter<String> implements
		Stringifier<T> {

	public String toString(T node) {
		return node.apply(this);
	}

	public @Override
	String defaultNode(Node node) {
		return node.toString();
	}

	public @Override
	String caseABinopExp(ABinopExp node) {
		StringBuffer sb = new StringBuffer();
		if (node.getLeft().apply(new Precedence()) > node.apply(new Precedence())) {
			sb.append("(");
			sb.append(node.getLeft().apply(this));
			sb.append(")");
		} else {
			sb.append(node.getLeft().apply(this));
		}
		sb.append(node.getBinop().apply(this));
		if (node.getRight().apply(new Precedence()) > node.apply(new Precedence())) {
			sb.append("(");
			sb.append(node.getRight().apply(this));
			sb.append(")");
		} else {
			sb.append(node.getRight().apply(this));
		}
		return sb.toString();
	}

	public @Override
	String caseAIfThenElseStm(AIfThenElseStm node) {
		return "if (" + node.getCondition().apply(this) + ")";
	}

	public @Override
	String caseAIfThenStm(AIfThenStm node) {
		return "if (" + node.getCondition().apply(this) + ")";
	}

	public @Override
	String caseAWhileStm(AWhileStm node) {
		return "while (" + node.getCondition().apply(this) + ")";
	}

	public @Override
	String caseAAssignVarStm(AAssignVarStm node) {
		return node.getIdentifier().getText() + " = "
				+ node.getExp().apply(this);
	}

	public @Override
	String caseAAssignPtrStm(AAssignPtrStm node) {
		return "*" + node.getLeft().apply(this) + " = "
				+ node.getRight().apply(this);
	}

	public @Override
	String caseAMallocExp(AMallocExp node) {
		return "malloc-" + node.getMallocIndex();
	}

	public @Override @SuppressWarnings("unused")
	String caseAInputExp(AInputExp node) {
		return "input";
	}

	public @Override @SuppressWarnings("unused")
	String caseANullExp(ANullExp node) {
		return "null";
	}

	public @Override
	String caseAPointerExp(APointerExp node) {
		return "&" + node.getIdentifier().apply(this);
	}

	public @Override
	String caseADerefExp(ADerefExp node) {
		return "*" + node.getExp().apply(this);
	}

	public @Override
	String caseAFunctionCallExp(AFunctionCallExp node) {
		StringBuffer sb = new StringBuffer();
		sb.append(node.getIdentifier().apply(this));
		sb.append("(");
		for (Iterator<PExp> i = node.getArgs().iterator(); i.hasNext();) {
			sb.append(i.next().apply(this));
			if (i.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append(")");
		return sb.toString();
	}

	public @Override
	String caseAFunctionPtrCallExp(AFunctionPtrCallExp node) {
		StringBuffer sb = new StringBuffer();
		sb.append("(");
		sb.append(node.getExp().apply(this));
		sb.append(")(");
		for (Iterator<PExp> i = node.getArgs().iterator(); i.hasNext();) {
			sb.append(i.next().apply(this));
			if (i.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append(")");
		return sb.toString();
	}

	public @Override
	String caseAReturnStm(AReturnStm node) {
		return "return " + node.getExp().apply(this);
	}

	public @Override
	String caseAOutputStm(AOutputStm node) {
		return "output " + node.getExp().apply(this);
	}

	public @Override
	String caseADeclStm(ADeclStm node) {
		StringBuffer sb = new StringBuffer();
		sb.append("var ");
		for (Iterator<ALocalDecl> i = node.getDecl().iterator(); i.hasNext();) {
			sb.append(i.next().apply(this));
			if (i.hasNext()) {
				sb.append(",");
			}
		}
		return sb.toString();
	}

	public @Override
	String caseAVarExp(AVarExp node) {
		return node.getIdentifier().getText();
	}

	public @Override
	String caseAIntConstExp(AIntConstExp node) {
		return node.getIntegerConst().getText();
	}

	public @Override
	String caseALocalDecl(ALocalDecl node) {
		return node.getName().getText();
	}

	public @Override
	String defaultPBinop(PBinop binop) {
		return binop.toString().trim();
	}
}
