package tip.phases;

import tip.instructions.*;
import tip.analysis.*;
import tip.node.*;
import java.util.*;

public class CodeGeneration extends DepthFirstAdapter {
	/** The list of instructions for the body currently being generated */
	private List<Instruction> current_instrs;

	/** Puts an instruction into the current list */
	private void put(Instruction inst) {
		current_instrs.add(inst);
	}

	public @Override
	void inABody(ABody body) {
		// Make list of instructions for function
		body.setInstructions(new LinkedList<Instruction>());
		current_instrs = body.getInstructions();
	}

	public @Override
	void caseAIntConstExp(AIntConstExp node) {
		//
		// 'intconst' ==>
		// loadc 'value'
		//
		int value = Integer.parseInt(node.getIntegerConst().getText());
		put(new Iloadc(value));
	}

	public @Override @SuppressWarnings("unused")
	void caseAMallocExp(AMallocExp node) {
		//
		// 'malloc' ==>
		// malloc
		//
		put(new Imalloc());
	}

	public @Override @SuppressWarnings("unused")
	void caseANullExp(ANullExp node) {
		//
		// 'null' ==>
		// loadc 0
		//
		put(new Iloadc(0));
	}

	public @Override
	void outAFunctionCallExp(AFunctionCallExp node) {
		//
		// id(E1,...,En) ==>
		//   
		// E1
		// E2
		// ...
		// En
		// call id
		put(new Icall((AFuncDecl) node.getIdentifier().getDecl()));
	}

	public @Override @SuppressWarnings("unused")
	void caseAInputExp(AInputExp node) {
		//
		// 'input' ==>
		//
		// input
		//
		put(new Iinput());
	}

	public @Override
	void caseAOutputStm(AOutputStm node) {
		//
		// 'output E' ==>
		//
		// E
		// output
		//
		node.getExp().apply(this);
		put(new Ioutput());
	}

	public @Override
	void caseAFunctionPtrCallExp(AFunctionPtrCallExp node) {
		//
		// (E)(E1,...,En) ==>
		//
		// E1
		// E2
		// ...
		// En
		// E
		// call_indirect
		// 
		//
		for (PExp exp : node.getArgs()) {
			exp.apply(this);
		}
		node.getExp().apply(this);
		put(new Icall_indirect());
	}

	public @Override
	void caseAIfThenStm(AIfThenStm node) {
		//
		// if(E) S ==>
		//   
		// E
		// ifeq false
		// S
		// false:
		//

		Ilabel labelFalse = new Ilabel("false");

		node.getCondition().apply(this);
		put(new Iifeq(labelFalse));
		node.getBody().apply(this);
		put(labelFalse);
	}

	public @Override
	void caseAIfThenElseStm(AIfThenElseStm node) {
		//
		// if(E) S1 else S2 ==>
		//   
		// E
		// ifeq false
		// S1
		// goto endif
		// false:
		// S2
		// endif:
		//
		Ilabel labelFalse = new Ilabel("false");
		Ilabel labelEndIf = new Ilabel("endif");

		node.getCondition().apply(this);
		put(new Iifeq(labelFalse));
		node.getThenBody().apply(this);
		put(new Igoto(labelEndIf));
		put(labelFalse);
		node.getElseBody().apply(this);
		put(labelEndIf);
	}

	public @Override
	void caseAWhileStm(AWhileStm node) {
		//
		// while(E) S ==>
		//  
		// goto cond
		// loop:
		// S
		// cond:
		// E
		// ifne loop
		//
		Ilabel labelCond = new Ilabel("cond");
		Ilabel labelLoop = new Ilabel("loop");

		put(new Igoto(labelCond));
		put(labelLoop);
		node.getBody().apply(this);
		put(labelCond);
		node.getCondition().apply(this);
		put(new Iifne(labelLoop));
	}

	public @Override
	void caseAReturnStm(AReturnStm node) {
		//
		// return E; ==>
		//
		// E
		// return
		//
		node.getExp().apply(this);
		put(new Ireturn());
	}

	public @Override
	void caseAVarExp(AVarExp node) {
		//
		// id ==>
		// load offset(id)
		//
		// f ==>
		// mload offset(f)
		//
		if (node.getIdentifier().getDecl() instanceof ALocalDecl) {
			put(new Iload(
					((ALocalDecl) node.getIdentifier().getDecl()).getIndex()));
		} else {
			put(new Imload((AFuncDecl) node.getIdentifier().getDecl()));
		}
	}

	public @Override
	void caseADerefExp(ADerefExp node) {
		//
		// *E ==>
		// E
		// deref
		//
		node.getExp().apply(this);
		put(new Ideref());
	}

	public @Override
	void caseAPointerExp(APointerExp node) {
		//
		// &id ==>
		// aload offset(id)
		//
		put(new Iaload(((ALocalDecl) node.getIdentifier().getDecl()).getIndex()));
	}

	public @Override
	void caseABinopExp(ABinopExp node) {
		//
		// E1 op E2 ==>
		// E1
		// E2
		// add / sub / mul / div
		//
		// where op \in {+,-,*,/}
		//
		//
		// E1 <op> E2 ==>
		//
		// E1
		// E2
		// ifcmp_eq / ifcmp_gt true
		// loadc 0
		// goto end
		// true:
		// loadc 1
		// end:
		//
		// where op \in {==, >}
		//
		node.getLeft().apply(this);
		node.getRight().apply(this);

		switch (node.getBinop().kindPBinop()) {
		case PLUS:
			put(new Iadd());
			break;
		case MINUS:
			put(new Isub());
			break;
		case TIMES:
			put(new Imul());
			break;
		case DIVIDE:
			put(new Idiv());
			break;
		case EQ:
		case GT:
			Ilabel labelTrue = new Ilabel("true");
			Ilabel labelEnd = new Ilabel("end");
			if (node.getBinop().kindPBinop() == EBinop.EQ) {
				put(new Iifcmp_eq(labelTrue));
			} else {
				put(new Iifcmp_gt(labelTrue));
			}
			put(new Iloadc(0));
			put(new Igoto(labelEnd));
			put(labelTrue);
			put(new Iloadc(1));
			put(labelEnd);
			break;
		}
	}

	public @Override
	void caseAAssignVarStm(AAssignVarStm node) {
		//
		// id = E ==>
		// E
		// store offset(id)
		//
		node.getExp().apply(this);
		put(new Istore(((ALocalDecl) node.getIdentifier().getDecl()).getIndex()));
	}

	/*
	 * replaced by *E1 = E2 public @Override void caseAAssignPtrStm(
	 * AAssignPtrStm node ){ // // *id = E ==> // E // load offset(id) [ local
	 * id contains address ] // mstore // node.getExp().apply(this); put(new
	 * Iload( ((ALocalDecl)node.getIdentifier().decl).index)); put(new
	 * Imstore()); }
	 */
	public @Override
	void caseAAssignPtrStm(AAssignPtrStm node) {
		//
		// *E1 = E2 ==>
		// E2
		// E1
		// mstore
		//
		node.getRight().apply(this);
		node.getLeft().apply(this);
		put(new Imstore());
	}
}