package tip.phases;

import tip.*;
import tip.types.*;
import tip.analysis.*;
import tip.node.*;
import java.util.*;

/*
 *  Compiler phase to generate & solve type constraints using
 *  the unification algorithm.
 */
public class TypeChecking extends DepthFirstAdapter {
	// public Type PExp.type = null;
	// public Type TIdentifier.type = null;

	/*
	 * after(PExp p) returning(Object o): target(p) && call(Object clone()) {
	 * ((PExp)o).type = p.type; } after(TIdentifier i) returning(Object o):
	 * target(i) && call(Object clone()) { ((TIdentifier)o).type = i.type; }
	 */
	public @Override
	void caseTIdentifier(TIdentifier name) {
		// Make the type field of all identifier point to the type of the
		// declaration
		if (!(name.parent() instanceof ALocalDecl || name.parent() instanceof AFuncDecl)) {
			name.setType(name.getDecl().getName().getType());
		}
	}

	public @Override
	void outAVarExp(AVarExp node) {
		// Propagate type variable to expression node
		node.setType(node.getIdentifier().getType());
	}

	public @Override
	void outADerefExp(ADerefExp node) {
		//
		// *E generates the constraint
		// [[E]] = &[[*E]]
		//
		PExp exp = node.getExp();
		node.setType(new Variable());
		new Pointer(node.getType()).unify(exp.getType());
	}

	public @Override
	void outAPointerExp(APointerExp node) {
		//
		// &id generates the constraint
		// [[&id]] = &[[id]]
		//
		node.setType(new Variable());
		new Pointer(node.getIdentifier().getType()).unify(node.getType());
	}

	public @Override
	void outABinopExp(ABinopExp node) {
		//
		// E1 op E2 generates the constraints
		// [[E1]] = [[E2]] = [[E1 op E1]] = int
		//
		// where op \in {+,-,*,/,>}
		//
		// E1 == E2 generates the constraints
		// [[E1 == E2]] = int
		// [[E1]]=[[E2]]
		//
		node.setType(new Variable());
		node.getType().unify(new Int());
		if (node.getBinop().kindPBinop() != EBinop.EQ) {
			node.getLeft().getType().unify(new Int());
			node.getRight().getType().unify(new Int());
		} else {
			// Note: not part of TIP cf. slides but makes sense
			node.getLeft().getType().unify(node.getRight().getType());
		}
	}

	public @Override
	void outAAssignVarStm(AAssignVarStm node) {
		//
		// id = E generates the constraints
		// [[id]] = [[E]]
		//

		// Check if 'id' is a local variable
		if (!(node.getIdentifier().getDecl() instanceof ALocalDecl)) {
			Errors.error(node.getToken(), "Expected lvalue");
		}

		node.getExp().getType().unify(node.getIdentifier().getType());
	}

	/*
	 * replaced by *E1 = E2 public @Override void outAAssignPtrStm(
	 * AAssignPtrStm node ){ // // *id = E generates the constraints // [[id]] =
	 * &[[E]] // // Check if 'id' is a local variable if(
	 * !(node.getIdentifier().decl instanceof ALocalDecl) ){
	 * Errors.error(node.getToken(),"Expected lvalue"); } new
	 * Pointer(node.getExp().type).unify(node.getIdentifier().type); }
	 */
	public @Override
	void outAAssignPtrStm(AAssignPtrStm node) {
		//
		// *E1 = E2 generates the constraints
		// [[E1]] = &[[E2]]
		//

		// Check if 'id' is a local variable
		// if( !(node.getIdentifier().decl instanceof ALocalDecl) ){
		// Errors.error(node.getToken(),"Expected lvalue");
		// }

		new Pointer(node.getRight().getType()).unify(node.getLeft().getType());
	}

	public @Override
	void outAIntConstExp(AIntConstExp node) {
		//
		// 'intconst' generates the constraints
		// [[intconst]] = int
		//
		node.setType(new Variable());
		node.getType().unify(new Int());
	}

	public @Override
	void outAMallocExp(AMallocExp node) {
		//
		// 'malloc' generates the constraints
		// [[malloc]] = &\alpha
		//
		node.setType(new Variable());
		node.getType().unify(new Pointer(new Variable()));
	}

	public @Override
	void outANullExp(ANullExp node) {
		//
		// 'null' generates the constraints
		// [[null]] = &\alpha
		//
		node.setType(new Variable());
		node.getType().unify(new Pointer(new Variable()));
	}

	public @Override
	void outAInputExp(AInputExp node) {
		//
		// 'input' generates the constraints
		// [[input]] = int
		//
		node.setType(new Variable());
		node.getType().unify(new Int());
	}

	public @Override
	void outAFunctionCallExp(AFunctionCallExp node) {
		//
		// 'id(E1,...,En)' generates the constraints
		// [[id]] = ([[E1]],..,[[En]])->[[id(E1,...,En)]]
		//
		node.setType(new Variable());

		LinkedList<Type> argTypes = new LinkedList<Type>();
		for (PExp exp : node.getArgs()) {
			argTypes.add(exp.getType());
		}

		if (!(node.getIdentifier().getDecl() instanceof AFuncDecl)) {
			Errors.error(node.getIdentifier(), "Illegal use of identifier '"
					+ node.getIdentifier() + "'");
		}

		Function f = new Function(argTypes, node.getType());
		f.unify(node.getIdentifier().getType());
	}

	public @Override
	void outAFunctionPtrCallExp(AFunctionPtrCallExp node) {
		//
		// '(E)(E1,...,En)' generates the constraints
		// [[E]] = ([[E1]],..,[[En]])->[[(E)(E1,...,En)]]
		//
		node.setType(new Variable());

		LinkedList<Type> argTypes = new LinkedList<Type>();
		for (PExp exp : node.getArgs()) {
			argTypes.add(exp.getType());
		}

		Function f = new Function(argTypes, node.getType());
		f.unify(node.getExp().getType());
	}

	public @Override
	void outAFuncDecl(AFuncDecl node) {
		//
		// 'id(id1,...,idn){...return E; }' generates the constraints
		// [[id]] = ([[id1]],..,[[idn]])->[[E]]
		//
		LinkedList<Type> argTypes = new LinkedList<Type>();
		// TIdentifier name = node.getName();.decl.getName();
		ABody body = node.getBody();
		ABlock block = body.getBlock();

		PExp retexp = ((AReturnStm) block.getStatements().getLast()).getExp();

		for (PDecl formal : node.getFormals()) {
			argTypes.add(((ALocalDecl) formal).getName().getType());
		}

		Function f = new Function(argTypes, retexp.getType());
		// f.unify(name.type);
		f.unify(node.getName().getType());
	}

	public @Override
	void outAIfThenStm(AIfThenStm node) {
		//
		// 'if(E) S' generates the constraints
		// [[E]] = int
		//
		node.getCondition().getType().unify(new Int());
	}

	public @Override
	void outAIfThenElseStm(AIfThenElseStm node) {
		//
		// 'if(E) S1 else S2' generates the constraints
		// [[E]] = int
		//
		node.getCondition().getType().unify(new Int());
	}

	public @Override
	void outAWhileStm(AWhileStm node) {
		//
		// 'while(E) S' generates the constraints
		// [[E]] = int
		//
		node.getCondition().getType().unify(new Int());
	}

	public @Override
	void outAOutputStm(AOutputStm node) {
		//
		// 'output E' generates the constraints
		// [[E]] = int
		//
		node.getExp().getType().unify(new Int());
	}

	public @Override
	void inAProgram(AProgram node) {
		node.apply(new InitTypes());
	}

	public @Override @SuppressWarnings("unused")
	void outAProgram(AProgram node) {
		Type.renameRootVariables();
	}

	private class InitTypes extends DepthFirstAdapter {
		public @Override
		void inALocalDecl(ALocalDecl node) {
			// Create type variables for local declarations
			node.getName().setType(new Variable());
		}

		public @Override
		void inAFuncDecl(AFuncDecl node) {
			// Create type variables for function declarations
			node.getName().setType(new Variable());
		}
	}

}