package tip.types;

import java.util.*;

public abstract class Type {
	private Type parent;

	// All allocated type objects are stored in this list so that we can call
	// clearMarks()
	// before calling toString(). The mark bits are simply used to avoid
	// infinite recursion
	// when printing infinite types.
	private static List<Type> typeNodes = new LinkedList<Type>();

	private boolean marked;

	protected Type() {
		parent = this;
		marked = false;
		typeNodes.add(this);
	}

	/**
	 * Clears all 'marked' in all type objects. Must be called prior to toString
	 * for each object. This is obviously inefficient but simple.
	 */
	public static void clearMarks() {
		for (Type t : typeNodes) {
			t.marked = false;
		}
	}

	/**
	 * Returns true if this object has previously been visited during toString
	 */
	protected boolean isMarked() {
		return marked;
	}

	/**
	 * Marks this object as visited
	 */
	protected void setMarked() {
		marked = true;
	}

	/**
	 * Unify this type and the given type
	 */
	public abstract void unify(Type type);

	/**
	 * Computes and returns the union of this set and the given set. The root is
	 * always this set.
	 */
	public Type union(Type other) {
		Type r1 = find();
		Type r2 = other.find();

		if (r1 == r2) {
			return r1;
		}

		// r1 is always root
		r2.parent = r1;
		return r1;
	}

	/**
	 * Returns the canonical element of the set that this element is in.
	 */
	public Type find() {
		if (parent != this) {
			parent = parent.find();
		}

		return parent;
	}

	/**
	 * Returns true if this element is in the same equivalence class as 'other',
	 * false otherwise.
	 */
	public boolean equiv(Type other) {
		return find() == other.find();
	}

	/**
	 * Returns true if this is a root, false otherwise.
	 */
	public boolean isRoot() {
		return this == find();
	}

	/**
	 * Assigns names to type variables ( a,b,c,...) ML style
	 */
	public static void renameRootVariables() {
		char c = 'a';
		for (Type t : typeNodes) {
			if (!t.isRoot() || !(t instanceof Variable)) {
				continue;
			}

			Variable var = (Variable) t;
			var.setName(c);
			c++;
		}
	}
}
