package tip.util;

import java.util.*;
import tip.cfg.*;
import tip.node.*;
import tip.analysis.*;

/**
 * {@code DeclCollector} defines methods for collecting all referred
 * {@link PDecl} nodes in the subtree of an AST node or in the subtrees of nodes
 * associated with points in a control flow graph.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class DeclCollector extends DepthFirstAdapter {
	// the collection of declarations
	private Set<PDecl> collection = new HashSet<PDecl>();

	/**
	 * Adds {@link PDecl} instances of {@code node.decl} to the collection.
	 * 
	 * @param node
	 *            the visited {@link TIdentifier}
	 */
	public @Override
	void caseTIdentifier(TIdentifier node) {
		if (node.getDecl() != null) {
			collection.add(node.getDecl());
		}
	}

	/**
	 * Adds the {@link PDecl} {@code node} to the collection.
	 * 
	 * @param node
	 *            the visited {@link PDecl}
	 */
	public @Override
	void defaultInPDecl(PDecl node) {
		collection.add(node);
	}

	/**
	 * Collects all {@link PDecl} nodes referred to in the subtree of {@code
	 * node}.
	 * 
	 * @param node
	 *            the node where the collection starts
	 * @return the collection referred {@link PDecl} nodes
	 */
	public static Collection<PDecl> collectDecls(Node node) {
		DeclCollector collector = new DeclCollector();
		if (node != null) {
			node.apply(collector);
		}
		return collector.collection;
	}

	/**
	 * Collects all {@link PDecl} nodes referred to in the subtrees of nodes
	 * associated with the program points in the set of {@code points}.
	 * 
	 * @param points
	 *            the set of program points investigated
	 * @return the collection referred {@link PDecl} nodes
	 */
	public static Collection<PDecl> collectDecls(
			Collection<IProgramPoint> points) {
		DeclCollector collector = new DeclCollector();
		if (points != null) {
			for (IProgramPoint point : points) {
				if (point.getNode() != null) {
					point.getNode().apply(collector);
				}
			}
		}
		return collector.collection;
	}

	/**
	 * Collects all {@link PDecl} nodes referred to in the subtrees of nodes
	 * associated with the program points in the control flow graph {@code cfg}.
	 * 
	 * @param cfg
	 *            the investigated control flow graph
	 * @return the collection referred {@link PDecl} nodes
	 */
	public static Collection<PDecl> collectDecls(ICFG cfg) {
		DeclCollector collector = new DeclCollector();
		if (cfg != null) {
			cfg.apply(collector);
		}
		return collector.collection;
	}
}
