package tip.util;

import tip.InternalCompilerError;
import tip.analysis.*;
import tip.node.*;

/**
 * {@code ExpressionEqualitor} define an equality on binary expressions, such
 * that for instance two different occurences of the binary expression {@code
 * a-b} are regarded the same expression.
 * <p>
 * The class uses two inner classes {@code ExpressionEquality}, an instance of
 * the {@link tip.analysis.QuestionAnswer} interface, and {@code
 * ExpressionHashcode}, an instance of the {@link tip.analysis.Answer}
 * interface, to compute the {@link #equals(ABinopExp, ABinopExp)} and
 * {@link #hashCode(ABinopExp)} methods.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class ExpressionEqualitor implements Equalitor<ABinopExp> {
	/** the local ExpressionEquality instance */
	private static ExpressionEquality equality = new ExpressionEquality();

	/** the local ExpressionHashcode instance */
	private static ExpressionHashcode hashcode = new ExpressionHashcode();

	/**
	 * Returns {@code true} if the two expressions are regard as the same binary
	 * expression.
	 * 
	 * @param a
	 *            the first binary expression
	 * @param b
	 *            the second binary expression
	 * @return {@code true} if the two expressions are regard as the same binary
	 *         expression
	 */
	public boolean equals(ABinopExp a, ABinopExp b) {
		return a == null ? b == null : a.apply(equality, b);
	}

	/**
	 * Returns a hashcode of the binary expression {@code a} consistent with the
	 * {@link #equals(ABinopExp, ABinopExp)} method.
	 * 
	 * @param a
	 *            the binary expression in question
	 * @return the hashcode of {@code a} according to this equalitor
	 */
	public int hashCode(ABinopExp a) {
		return a == null ? 0 : a.apply(hashcode);
	}

	/**
	 * Returns {@code true} if the object {@code o} is an instance
	 * {@link ExpressionEqualitor}. This is an approiate equality on this class
	 * as it has no state.
	 * 
	 * @param o
	 *            the object in question
	 * @return {@code true} if {@code o} is an instance of this class
	 */
	@Override
	public boolean equals(Object o) {
		if (o == null) {
			return false;
		}
		return getClass().equals(o.getClass());
	}

	/**
	 * Returns the hashcode for this class.
	 * 
	 * @return the hashcode for this class
	 */
	@Override
	public int hashCode() {
		return getClass().hashCode();
	}

	/**
	 * {@code ExpressionEquality} is an instance of the
	 * {@link tip.analysis.QuestionAnswer} interface taking question of the type
	 * {@link Node} and returning answers of the type {@link Boolean}. The class
	 * determines equality of expressions to use for the
	 * {@link ExpressionEqualitor}.
	 * <p>
	 * {@link AFunctionCallExp} nodes are regarded as equal on a node basis
	 * since the result of the expression may vary if sideeffects occur.
	 * <p>
	 * {@link AFunctionPtrCallExp} and {@link ADerefExp} nodes are regarded as
	 * equal on a node basis since the target value may change through (other)
	 * pointer manipulating operations.
	 * <p>
	 * {@link AMallocExp} and {@link AInputExp} nodes as equal on a node basis
	 * since the values of the expressions are (potentially) different for each
	 * occurence.
	 * <p>
	 * 
	 * @author Johnni Winther, jw@brics.dk
	 */
	static class ExpressionEquality extends
			QuestionAnswerAdapter<Node, Boolean> {
		/**
		 * Returns {@code false} as the default result of equality of the two
		 * arbitrary nodes {@code a} and {@code b}.
		 * 
		 * @param a
		 *            the first node
		 * @param b
		 *            the second node
		 * @return {@code false} as the default result
		 */
		public @Override
		Boolean defaultNode(Node a, Node b) {
			return false;
		}

		/**
		 * Returns {@code true} if {@code b} is an {@link ABinopExp} which is
		 * equal to {@code a} according to this equalitor. Two binary
		 * expressions are equal if the arguments and operand are equal taking
		 * the commutativity of the operand into account.
		 * 
		 * @param a
		 *            the binary expression
		 * @param b
		 *            the node tested for equality against {@code a}
		 * @return {@code true} if {@code b} represents the same binary
		 *         expression
		 */
		public @Override
		Boolean caseABinopExp(ABinopExp a, Node b) {
			if (b instanceof ABinopExp) {
				ABinopExp otherBinop = (ABinopExp) b;
				if (otherBinop.getBinop().kindPBinop() == a.getBinop().kindPBinop()) {
					switch (a.getBinop().kindPBinop()) {
					// commutative operands
					case TIMES:
					case PLUS:
					case EQ:
						return (a.getLeft().apply(this, otherBinop.getLeft()) && a.getRight().apply(
								this, otherBinop.getRight()))
								|| (a.getLeft().apply(this,
										otherBinop.getRight()) && a.getRight().apply(
										this, otherBinop.getLeft()));
						// not commutative operands
					case DIVIDE:
					case MINUS:
					case GT:
						return a.getLeft().apply(this, otherBinop.getLeft())
								&& a.getRight().apply(this,
										otherBinop.getRight());
					default:
						throw new InternalCompilerError("Unknown binop kind : "
								+ a.getBinop().kindPBinop());
					}
				}
			}
			return false;
		}

		/**
		 * Returns {@code true} if {@code b} is an {@link APointerExp} which is
		 * equal to {@code a} according to this equalitor. Two pointer
		 * expressions are equal if the variables are the same. This is only
		 * safe as long as the addresses of the variables are static, and is
		 * thus not an appropriate equality for analyses computing flow through
		 * pointer variables.
		 * 
		 * @param a
		 *            the pointer expression
		 * @param b
		 *            the node tested for equality against {@code a}
		 * @return {@code true} if {@code b} represents the same pointer
		 *         expression
		 */
		public @Override
		Boolean caseAPointerExp(APointerExp a, Node b) {
			if (b instanceof APointerExp) {
				APointerExp otherPointer = (APointerExp) b;
				return a.getIdentifier().getDecl() == otherPointer.getIdentifier().getDecl();
			}
			return false;
		}

		/**
		 * Returns {@code true} if {@code b} is an {@link AIntConstExp} which is
		 * equal to {@code a} according to this equalitor. Two constant
		 * expressions are equal if the constant values are equal.
		 * 
		 * @param a
		 *            the constant expression
		 * @param b
		 *            the node tested for equality against {@code a}
		 * @return {@code true} if {@code b} represents the same contanst
		 *         expression
		 */
		public @Override
		Boolean caseAIntConstExp(AIntConstExp a, Node b) {
			if (b instanceof AIntConstExp) {
				AIntConstExp otherIntConst = (AIntConstExp) b;
				return a.getIntegerConst().equals(
						otherIntConst.getIntegerConst());
			}
			return false;
		}

		/**
		 * Returns {@code true} if {@code b} is an {@link AVarExp} which is
		 * equal to {@code a} according to this equalitor. Two variable
		 * expressions are equal if the variables are the same.
		 * 
		 * @param a
		 *            the variable expression
		 * @param b
		 *            the node tested for equality against {@code a}
		 * @return {@code true} if {@code b} represents the same variable
		 *         expression
		 */
		public @Override
		Boolean caseAVarExp(AVarExp a, Node b) {
			if (b instanceof AVarExp) {
				AVarExp otherVar = (AVarExp) b;
				return a.getIdentifier().getDecl() == otherVar.getIdentifier().getDecl();
			}
			return false;
		}

		/**
		 * Returns {@code true} if {@code b} is an {@link ANullExp} which is
		 * equal to {@code a} according to this equalitor. Two null expressions
		 * are always equal.
		 * 
		 * @param a
		 *            the binary expression
		 * @param b
		 *            the node tested for equality against {@code a}
		 * @return {@code true} if {@code b} represents the same binary
		 *         expression
		 */
		public @Override
		Boolean caseANullExp(ANullExp a, Node b) {
			return b instanceof ANullExp;
		}
	}

	/**
	 * {@code ExpressionHashcode} is an instance of the
	 * {@link tip.analysis.Answer} interface returning answers of the type
	 * {@link Integer}. The class computes the hashcode for expressions to use
	 * for the {@link ExpressionEqualitor}.
	 * 
	 * @author Johnni Winther, jw@brics.dk
	 */
	static class ExpressionHashcode extends AnswerAdapter<Integer> {
		/**
		 * Returns the hashcode of {@code node} as the default hashcode for
		 * nodes to be consistent with
		 * {@link ExpressionEquality#defaultNode(Node, Node)}.
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer defaultNode(Node node) {
			return node.hashCode();
		}

		/**
		 * Returns a hashcode for the binary expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseABinopExp(ABinopExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseABinopExp(ABinopExp node) {
			return node.getLeft().apply(this)
					^ node.getBinop().kindPBinop().hashCode()
					^ node.getRight().apply(this);
		}

		/**
		 * Returns a hashcode for the pointer expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseAPointerExp(APointerExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseAPointerExp(APointerExp node) {
			return node.getClass().hashCode()
					^ node.getIdentifier().getDecl().hashCode();
		}

		/**
		 * Returns a hashcode for the pointer expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseAPointerExp(APointerExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseADerefExp(ADerefExp node) {
			return node.getClass().hashCode() ^ node.getExp().apply(this);
		}

		/**
		 * Returns a hashcode for the pointer expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseAIntConstExp(AIntConstExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseAIntConstExp(AIntConstExp node) {
			return node.getIntegerConst().hashCode();
		}

		/**
		 * Returns a hashcode for the pointer expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseAIntConstExp(AIntConstExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseAVarExp(AVarExp node) {
			return node.getClass().hashCode()
					^ node.getIdentifier().getDecl().hashCode();
		}

		/**
		 * Returns a hashcode for the pointer expression {@code node} consistent
		 * with the result of
		 * {@link ExpressionEquality#caseANullExp(ANullExp, Node)}
		 * 
		 * @return the hashcode of {@code node}
		 */
		public @Override
		Integer caseANullExp(ANullExp node) {
			return node.getClass().hashCode();
		}
	}
}
