package tip.util;

import java.util.*;
import tip.cfg.*;
import tip.node.*;
import tip.analysis.*;

/**
 * {@code LocalDeclCollector} defines methods for collecting all referred
 * {@link ALocalDecl} nodes in the subtree of an AST node or in the subtrees of
 * nodes associated with points in a control flow graph.
 * 
 * @author Johnni Winther, jw@brics.dk
 */
public class LocalDeclCollector extends DepthFirstAdapter {
	// the collection of declarations
	private Set<ALocalDecl> collection = new HashSet<ALocalDecl>();

	/**
	 * Adds {@link ALocalDecl} instances of {@code node.decl} to the collection.
	 * 
	 * @param node
	 *            the visited {@link TIdentifier}
	 */
	public @Override
	void caseTIdentifier(TIdentifier node) {
		if (node.getDecl() instanceof ALocalDecl) {
			collection.add((ALocalDecl) node.getDecl());
		}
	}

	/**
	 * Adds the {@link ALocalDecl} {@code node} to the collection.
	 * 
	 * @param node
	 *            the visited {@link ALocalDecl}
	 */
	public @Override
	void caseALocalDecl(ALocalDecl node) {
		collection.add(node);
	}

	/**
	 * Collects all {@link ALocalDecl} nodes referred to in the subtree of
	 * {@code node}.
	 * 
	 * @param node
	 *            the node where the collection starts
	 * @return the collection referred {@link ALocalDecl} nodes
	 */
	public static Collection<ALocalDecl> collectLocalDecls(Node node) {
		LocalDeclCollector collector = new LocalDeclCollector();
		if (node != null) {
			node.apply(collector);
		}
		return collector.collection;
	}

	/**
	 * Collects all {@link ALocalDecl} nodes referred to in the subtrees of
	 * nodes associated with the program points in the set of {@code points}.
	 * 
	 * @param points
	 *            the set of program points investigated
	 * @return the collection referred {@link ALocalDecl} nodes
	 */
	public static Collection<ALocalDecl> collectLocalDecls(
			Collection<IProgramPoint> points) {
		LocalDeclCollector collector = new LocalDeclCollector();
		if (points != null) {
			for (IProgramPoint point : points) {
				if (point.getNode() != null) {
					point.getNode().apply(collector);
				}
			}
		}
		return collector.collection;
	}

	/**
	 * Collects all {@link ALocalDecl} nodes referred to in the subtrees of
	 * nodes associated with the program points in the control flow graph
	 * {@code cfg}.
	 * 
	 * @param cfg
	 *            the investigated control flow graph
	 * @return the collection referred {@link ALocalDecl} nodes
	 */
	public static Collection<ALocalDecl> collectLocalDecls(ICFG cfg) {
		LocalDeclCollector collector = new LocalDeclCollector();
		if (cfg != null) {
			cfg.apply(collector);
		}
		return collector.collection;
	}
}
