package tip.util;

import tip.node.*;
import tip.cfg.*;
import java.util.*;

/**
 * {@code SubExpressions} implements the {@link CFGAnswerAdapter} to retrieve
 * child nodes accepted by a {@link NodeFilter} in subexpressions of nodes
 * associated with program points in a control flow graph.
 * 
 * @author Johnni Winther, jw@brics.dk
 * @param <T>
 *            the type of the collected child nodes
 */
public class SubExpressions<T extends Node> extends
		CFGAnswerAdapter<Collection<T>> {
	private NodeFilter<T> filter;

	/**
	 * Creates a new {@code SubExpressions} instance using the
	 * {@link NodeFilter} {@code filter} for filtering nodes.
	 * 
	 * @param filter
	 *            the filter used for filtering nodes
	 */
	public SubExpressions(NodeFilter<T> filter) {
		this.filter = filter;
	}

	/**
	 * Returns all child nodes of {@code node} accepted by the filter.
	 * 
	 * @return all child nodes of {@code node} accepted by the filter
	 */
	@SuppressWarnings("unused")
	public @Override
	Collection<T> defaultNode(Node node, IProgramPoint point) {
		return node.getDescendants(filter);
	}

	/**
	 * Returns an empty set of nodes.
	 * 
	 * @return an empty set of nodes
	 */
	@SuppressWarnings("unused")
	public @Override
	Collection<T> caseEntryPoint(Node node, EntryPoint point) {
		return Collections.emptySet();
	}

	/**
	 * Returns an empty set of nodes.
	 * 
	 * @return an empty set of nodes
	 */
	@SuppressWarnings("unused")
	public @Override
	Collection<T> caseExitPoint(Node node, ExitPoint point) {
		return Collections.emptySet();
	}

	/**
	 * Returns all child nodes of the condition expression of {@code node}
	 * accepted by the filter and {@code node} if it is accepted by the filter.
	 * 
	 * @return child nodes of {@code node} accepted by the filter
	 */
	@SuppressWarnings( { "unchecked", "unused" })
	public @Override
	Collection<T> caseCIfThenStmPoint(AIfThenStm node, CIfThenStmPoint point) {
		Collection<T> result = node.getCondition().getDescendants(filter);
		if (filter.accept(node)) {
			result.add((T) node);
		}
		return result;
	}

	/**
	 * Returns all child nodes of the condition expression of {@code node}
	 * accepted by the filter and {@code node} if it is accepted by the filter.
	 * 
	 * @return child nodes of {@code node} accepted by the filter
	 */
	@SuppressWarnings( { "unchecked", "unused" })
	public @Override
	Collection<T> caseCIfThenElseStmPoint(AIfThenElseStm node,
			CIfThenElseStmPoint point) {
		Collection<T> result = node.getCondition().getDescendants(filter);
		if (filter.accept(node)) {
			result.add((T) node);
		}
		return result;
	}

	/**
	 * Returns all child nodes of the condition expression of {@code node}
	 * accepted by the filter and {@code node} if it is accepted by the filter.
	 * 
	 * @return child nodes of {@code node} accepted by the filter
	 */
	@SuppressWarnings( { "unchecked", "unused" })
	public @Override
	Collection<T> caseCWhileStmPoint(AWhileStm node, CWhileStmPoint point) {
		Collection<T> result = node.getCondition().getDescendants(filter);
		if (filter.accept(node)) {
			result.add((T) node);
		}
		return result;
	}
}
