package tip.vm;

import java.io.IOException;

/*
 *  Memory layout
 * 
 *  |---------------|     <-- 0
 *  | size of code  |
 *  |---------------|
 *  | index of main |
 *  |---------------|
 *  |               |
 *  |     F1        |     <-- codePtr
 *  |               |
 *  |---------------|
 *  |      .        |
 *  |      .        |
 *  |      .        |
 *  |---------------|
 *  |               |
 *  |     Fn        |
 *  |               |
 *  |---------------|     <-- HEAP_BASE
 *  |               |
 *  |     HEAP      |     <-- heapPtr
 *  |               |
 *  |---------------|     <-- STACK_BASE
 *  |               |
 *  |     STACK     |     <-- framePtr
 *  |               |     <-- stackPtr
 *  |---------------|
 * 
 * 
 */

public class TipVm {
	private int codePtr;

	private int stackPtr;

	private int framePtr;

	private int heapPtr;

	private int[] mem;

	public static final int MEM_SIZE = 16384;

	public static final int HEAP_BASE = 1024;

	public static final int STACK_BASE = 2048;

	// Instructions
	public static final int OP_ADD = 1;

	public static final int OP_SUB = 2;

	public static final int OP_MUL = 3;

	public static final int OP_DIV = 4;

	public static final int OP_IFEQ = 10;

	public static final int OP_IFNE = 11;

	public static final int OP_IFCMP_EQ = 12;

	public static final int OP_IFCMP_GT = 13;

	public static final int OP_LOADC = 20; // loadc 'const' - push constant on
											// stack

	public static final int OP_LOAD = 21; // load offset(local) - push contents
											// of local on stack

	public static final int OP_ALOAD = 22; // aload offset(local) - push
											// absolute address of local on
											// stack

	public static final int OP_STORE = 23;

	public static final int OP_MLOAD = 24; // mload mem - push memory location
											// on stack

	public static final int OP_MSTORE = 25;

	public static final int OP_CALL = 30;

	public static final int OP_CALL_INDIRECT = 31;

	public static final int OP_RETURN = 32;

	public static final int OP_GOTO = 33;

	public static final int OP_MALLOC = 40;

	public static final int OP_INPUT = 41;

	public static final int OP_OUTPUT = 42;

	public static final int OP_DEREF = 50; // deref

	public TipVm(int[] mem) {
		this.mem = mem;
	}

	public void runProgram() {
		try {
			exec();
		} catch (Exception e) {
			System.out.println("VM crash");
		}
	}

	public void push(int value) {
		mem[stackPtr++] = value;
	}

	public int pop() {
		return mem[--stackPtr];
	}

	public int malloc() {
		return heapPtr++;
	}

	public void callFunction(int function, int returnAddress) {
		int numLocals, numArgs;

		int prevFramePtr = framePtr;

		// index = mem[codePtr+1]; // location of function header
		numLocals = mem[function];
		numArgs = mem[function + 1]; // args already on stack

		// Set new framePtr
		framePtr = stackPtr - numArgs;

		// Make room for locals
		stackPtr += numLocals;

		// Save return address
		// push(codePtr+2);
		push(returnAddress);

		// Save framePtr
		push(prevFramePtr);

		// Set codePtr to location of function to call
		codePtr = function + 2;
	}

	public int getLocal(int index) {
		return mem[framePtr + index];
	}

	public void setLocal(int index, int value) {
		mem[framePtr + index] = value;
	}

	public int deref(int address) {
		return mem[address];
	}

	public void mstore(int address, int value) {
		// mem[mem[address]] = value;
		mem[address] = value;
	}

	public int readInput() {
		// TODO: fix this method
		int input = 0;
		try {
			if (System.in.available() == 0) {
				System.out.print("input: ");

				// wait for input
			}
			byte[] b = new byte[100];

			int length = 0;
			System.in.read(b);

			// Determine length
			for (int i = 0; i < 100; i++) {
				if (b[i] == '\n' || b[i] == '\r') {
					break;
				}
				length++;
			}
			byte[] num = new byte[length];
			for (int i = 0; i < length; i++) {
				num[i] = b[i];
			}

			input = Integer.parseInt(new String(num));

		} catch (IOException e) {
		}
		return input;
	}

	protected void exec() {
		int currentInst;
		boolean terminate = false;

		codePtr = 0;
		stackPtr = STACK_BASE;
		framePtr = STACK_BASE;
		heapPtr = HEAP_BASE;

		// Set up initial stack frame

		// TODO: For now only main without args works
		// callFunction(mem[0],0);
		callFunction(mem[1], 0);

		int v1, v2, index;

		while (!terminate) {
			currentInst = mem[codePtr];
			switch (currentInst) {
			case OP_ADD:
				v1 = pop();
				v2 = pop();
				push(v1 + v2);
				break;
			case OP_SUB:
				v1 = pop();
				v2 = pop();
				push(v2 - v1);
				break;
			case OP_MUL:
				v1 = pop();
				v2 = pop();
				push(v1 * v2);
				break;
			case OP_DIV:
				v1 = pop();
				v2 = pop();
				push(v2 / v1);
				break;

			case OP_IFEQ:
				index = mem[codePtr + 1];
				v1 = pop();
				if (v1 == 0) {
					// true
					codePtr = index;
					continue;
				} else {
					// false
					codePtr++;
				}
				break;

			case OP_IFNE:
				index = mem[codePtr + 1];
				v1 = pop();
				if (v1 != 0) {
					// true
					codePtr = index;
					continue;
				} else {
					// false
					codePtr++;
				}
				break;

			case OP_IFCMP_EQ:
				index = mem[codePtr + 1];
				v1 = pop();
				v2 = pop();
				if (v1 == v2) {
					// true
					codePtr = index;
					continue;
				} else {
					// false
					codePtr++;
				}
				break;

			case OP_IFCMP_GT:
				index = mem[codePtr + 1];
				v1 = pop();
				v2 = pop();
				if (v2 > v1) {
					// true
					codePtr = index;
					continue;
				} else {
					// false
					codePtr++;
				}
				break;

			case OP_LOADC:
				// Load constant
				push(mem[codePtr + 1]);
				codePtr++;
				break;

			case OP_LOAD:
				// Load contents of local
				index = mem[codePtr + 1];
				push(getLocal(index));
				codePtr++;
				break;

			case OP_ALOAD:
				// Load local offset
				index = mem[codePtr + 1];
				// Push absolute address of local
				push(index + framePtr);
				codePtr++;
				break;

			case OP_MLOAD:
				// Load absolute address
				index = mem[codePtr + 1];
				push(index);
				codePtr++;
				break;

			case OP_STORE:
				// Store in local
				v1 = pop();
				index = mem[codePtr + 1];
				setLocal(index, v1);
				codePtr++;
				break;

			case OP_MSTORE:
				// Store in memory cell (absolute)
				index = pop();
				v1 = pop();
				mstore(index, v1);
				break;

			case OP_DEREF:
				// Deref top of stack
				push(deref(pop()));
				break;

			case OP_GOTO:
				codePtr = mem[codePtr + 1];
				continue;

			case OP_CALL:
				callFunction(mem[codePtr + 1], codePtr + 2);
				continue;

			case OP_CALL_INDIRECT:
				index = pop();
				callFunction(index, codePtr + 1);
				continue;

			case OP_RETURN:
				// Save result
				v1 = pop();

				// Save framePtr
				v2 = pop();

				// Get return address
				codePtr = pop();

				// Restore stackPtr
				stackPtr = framePtr;

				// Restore previous framePtr
				framePtr = v2;

				if (stackPtr == STACK_BASE) {
					terminate = true;

					// Output result
					System.out.println(v1);

					break;
				}

				// Put the result on the top of stack
				push(v1);

				continue;

			case OP_MALLOC:
				push(malloc());
				break;

			case OP_INPUT:
				push(readInput());
				break;

			case OP_OUTPUT:
				v1 = pop();
				System.out.println(v1);
				break;

			}
			codePtr++;
		}
	}
}
